/**
 * TrueReality - N64/compile_static.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 * Version information:
 *
 *
**/










#include <stdio.h>
#include <type_sizes.h>
#include <file_loader_extern.h>
#include <compile_static_extern.h>
#include "../io.h"
#include "interrupts.h"
#include "memory.h"
#include "registers.h"
#include "rsp_registers.h"
#include "compile_static.h"
#include "compile_static_tab.h"
#include "compile_static_mnemonic.h"










/* this could be possible sped up by using `& 0x80000000' instead
   of `>> 31' ibut i am not sure - pls email! - niki */

#define CHECK_64BIT_OVERFLOW(x, y, z, dest) \
        if( ((x>>63)==(y>>63)) && ((x>>63)!=(z>>63)) ) \
        { \
                reg.do_or_check_sthg = RS4300I_DO_INTEGER_OVERFLOW_EXCEPTION; \
        } \
        else \
        { \
                reg.r[dest] = z; \
        }

#define CHECK_32BIT_OVERFLOW(x, y, z, dest) \
        if( ((x>>31)==(y>>31)) && ((x>>31)!=(z>>31)) ) \
        { \
                reg.do_or_check_sthg = RS4300I_DO_INTEGER_OVERFLOW_EXCEPTION; \
        } \
        else \
        { \
                reg.r[dest] = z; \
        }

#ifdef TR_64BIT_LONG
        #define TR_MASK_ERL      0xfffffffffffffffbUL
        #define TR_MASK_EXL      0xfffffffffffffffdUL
#else
        #define TR_MASK_ERL      0xfffffffffffffffbULL
        #define TR_MASK_EXL      0xfffffffffffffffdULL
#endif










void tr_compile_rom()
{
        long i;

        printf("Compiling rom image ($%lx bytes) ...", rom.length);
        fflush(stdout);

        tr_i = rom.image_instr;

        for(i=0; i<(rom.length>>2); i++)
        {
                /*
                printf("%08"PR32"x\n", ((tr_u32 *)rom.image)[i]);
                */
                tr_compile_tab[TR_COMPILE_OPCODE(((tr_u32 *)rom.image)[i])](
                                ((tr_u32 *)rom.image)[i],
                                &tr_i[i]
                                );
        }

        printf(" OK\n"
        "-------------------------------------------------------------------------------\n");
        fflush(stdout);

} /* void tr_compile_rom() */










static void tr_reserved()
{
        TR_PRINT_DEBUGGING_INFO("reserved", NOTHING);

        reg.do_or_check_sthg = RS4300I_DO_RESERVED_INSTRUCTION_EXCEPTION;

        reg.halt = 1;

} /* static void tr_reserved() */





static void tr_invalid()
{
        TR_PRINT_DEBUGGING_INFO("invalid", NOTHING);

        reg.halt = 1;

} /* static void tr_invalid() */





static void tr_unimplemented()
{
        TR_PRINT_DEBUGGING_INFO("unimplemented", NOTHING);

        reg.halt = 1;

} /* static void tr_unimplemented() */





/* special functions */




/* Shift Left Logical */

static void tr_SLL()
{
        TR_PRINT_DEBUGGING_INFO("SLL", RD_RT_SA);

        reg.r[C_RD] = (tr_s64)(tr_s32)((((tr_u32)reg.r[C_RT]) << C_SA) & 0xffffffff);
        reg.r[0] = 0;

} /* static void tr_SLL() */





#if 0 /* not used often */
static void tr_SLL_opt()
{
        TR_PRINT_DEBUGGING_INFO("_SLL_", RD_RT_SA);

        reg.r[C_RD] = (tr_s64)(tr_s32)(reg.r[C_RT]);
        reg.r[0] = 0;

} /* static void tr_SLL_opt() */
#endif





static void tr_CLR_opt()
{
        TR_PRINT_DEBUGGING_INFO("_CLR_", RD);

        reg.r[C_RD] = 0;

} /* static void tr_CLR_opt() */





static void tr_NOP_opt()
{
        TR_PRINT_DEBUGGING_INFO("_NOP_", NOTHING);

} /* static void tr_NOP_opt() */





static void tr_SRL()
{
        TR_PRINT_DEBUGGING_INFO("SRL", RD_RT_SA);

        reg.r[C_RD] = (tr_s64)(tr_s32)((tr_u32)reg.r[C_RT] >> C_SA);
        reg.r[0] = 0;

} /* static void tr_SRL() */





static void tr_SRA()
{
        TR_PRINT_DEBUGGING_INFO("SRA", RD_RT_SA);

        reg.r[C_RD] = (tr_s64)(tr_s32)((tr_s32)reg.r[C_RT] >> C_SA);
        reg.r[0] = 0;

} /* static void tr_SRA() */





static void tr_SLLV()
{
        TR_PRINT_DEBUGGING_INFO("SLLV", RD_RT_RS);

        reg.r[C_RD] = (tr_s64)(tr_s32)(((tr_u32)(reg.r[C_RT]) << (reg.r[C_RS] & 0x1f)) & 0xffffffff);
        reg.r[0] = 0;

} /* static void tr_SLLV() */





static void tr_SRLV()
{
        TR_PRINT_DEBUGGING_INFO("SRLV", RD_RT_RS);

        reg.r[C_RD] = (tr_s64)(tr_s32)((tr_u32)reg.r[C_RT] >> (reg.r[C_RS] & 0x1f));
        reg.r[0] = 0;

} /* static void tr_SRLV() */





static void tr_SRAV()
{
        TR_PRINT_DEBUGGING_INFO("SRAV", RD_RT_RS);

        reg.r[C_RD] = (tr_s64)(tr_s32)((tr_s32)reg.r[C_RT] >> (reg.r[C_RS] & 0x1f));
        reg.r[0] = 0;

} /* static void tr_SRAV() */





static void tr_JR()
{
        TR_PRINT_DEBUGGING_INFO("JR", RS);

        reg.delay = DELAY_JUMP;
        reg.delay_jump = reg.r[C_RS];

} /* static void tr_JR() */





static void tr_JALR()
{
        TR_PRINT_DEBUGGING_INFO("JALR", RD_RS);

        reg.delay = DELAY_JUMP;
        reg.delay_jump = reg.r[C_RS];
        reg.r[C_RD] = (tr_s64)(tr_s32)(C_PC() + 8);
        reg.r[0] = 0;

} /* static void tr_JALR() */





static void tr_SYSCALL()
{
        TR_PRINT_DEBUGGING_INFO("SYSCALL", NOTHING);

        reg.do_or_check_sthg |= RS4300I_DO_SYSCALL_EXCEPTION;

} /* static void tr_SYSCALL() */





static void tr_BREAK()
{
        TR_PRINT_DEBUGGING_INFO("BREAK", NOTHING);

        reg.do_or_check_sthg |= RS4300I_DO_BREAK_EXCEPTION;

} /* static void tr_BREAK() */





static void tr_SYNC()
{
        TR_PRINT_DEBUGGING_INFO("SYNC (ignored!)", NOTHING);

} /* static void tr_SYNC() */





static void tr_MFHI()
{
        TR_PRINT_DEBUGGING_INFO("MFHI", RD);

        reg.r[C_RD] = reg.hi;
        reg.r[0] = 0;

} /* static void tr_MFHI() */





static void tr_MTHI()
{
        TR_PRINT_DEBUGGING_INFO("MTHI", RS);

        reg.hi = reg.r[C_RS];

} /* static void tr_MTHI() */





static void tr_MFLO()
{
        TR_PRINT_DEBUGGING_INFO("MFLO", RD);

        reg.r[C_RD] = reg.lo;
        reg.r[0] = 0;

} /* static void tr_MFLO() */





static void tr_MTLO()
{
        TR_PRINT_DEBUGGING_INFO("MTLO", RS);

        reg.lo = reg.r[C_RS];

} /* static void tr_MTLO() */





static void tr_DSLLV()
{
        TR_PRINT_DEBUGGING_INFO("DSLLV", RD_RT_RS);

        reg.r[C_RD] = reg.r[C_RT] << (reg.r[C_RS] & 0x3f);
        reg.r[0] = 0;

} /* static void tr_DSLLV() */





static void tr_DSRLV()
{
        TR_PRINT_DEBUGGING_INFO("DSRLV", RD_RT_RS);

        reg.r[C_RD] = (tr_u64)reg.r[C_RT] >> (reg.r[C_RS] & 0x3f);
        reg.r[0] = 0;

} /* static void tr_DSRLV() */





static void tr_DSRAV()
{
        TR_PRINT_DEBUGGING_INFO("DSRAV", RD_RT_RS);

        reg.r[C_RD] = reg.r[C_RT] >> (reg.r[C_RS] & 0x3f);
        reg.r[0] = 0;

} /* static void tr_DSRAV() */





static void tr_MULT()
{
        tr_u64 temp;

        TR_PRINT_DEBUGGING_INFO("MULT", RS_RT);

#ifdef MACOS
    /* Gil: This is the only way that Spice Rotate will work on the Mac version, so
            I think that this is the right way to do it. */
    /* niki: gil, you must have a fucking stupid compiler! spicerot.v64 works
             the other way on my machine! */
        temp = (tr_s64)(tr_s32)(reg.r[C_RS] * reg.r[C_RT]);
#else
        temp = ((tr_s64)(tr_s32)reg.r[C_RS]) * ((tr_s64)(tr_s32)reg.r[C_RT]);
#endif
        reg.lo = (tr_s64)(tr_s32)(temp);
        reg.hi = (tr_s64)(tr_s32)(temp >> 32);

} /* static void tr_MULT() */





static void tr_MULTU()
{
        tr_u64 temp;

        TR_PRINT_DEBUGGING_INFO("MULTU", RS_RT);

#ifdef MACOS
    /* Gil: Spice Rotate doesn't need this, but I would guess that it should
            be the same as above. */
    /* niki: seems that your compiler handles 64 bit mult not corect. */
        temp = (tr_u64)(tr_u32)(reg.r[C_RS] * reg.r[C_RT]);
#else
        temp = ((tr_u64)(tr_u32)reg.r[C_RS]) * ((tr_u64)(tr_u32)reg.r[C_RT]);
#endif
        reg.lo = (tr_s64)(tr_s32)(temp);
        reg.hi = (tr_s64)(tr_s32)(temp >> 32);


} /* static void tr_MULTU() */





static void tr_DIV()
{
        TR_PRINT_DEBUGGING_INFO("DIV", RS_RT);

        if((tr_s32)reg.r[C_RT] != 0)
        {
                reg.lo = (tr_s64)(tr_s32)((tr_s32)reg.r[C_RS] / (tr_s32)reg.r[C_RT]);
                reg.hi = (tr_s64)(tr_s32)((tr_s32)reg.r[C_RS] % (tr_s32)reg.r[C_RT]);
        }

} /* static void tr_DIV() */





static void tr_DIVU()
{
        TR_PRINT_DEBUGGING_INFO("DIVU", RS_RT);

        if((tr_u32)reg.r[C_RT] != 0)
        {
                reg.lo = (tr_s64)(tr_s32)((tr_u32)reg.r[C_RS] / (tr_u32)reg.r[C_RT]);
                reg.hi = (tr_s64)(tr_s32)((tr_u32)reg.r[C_RS] % (tr_u32)reg.r[C_RT]);
        }

} /* static void tr_DIVU() */





static void tr_DMULT()
{
        TR_PRINT_DEBUGGING_INFO("DMULT", RS_RT);

        reg.lo = reg.r[C_RS] * reg.r[C_RT];
/* BUG! */
        reg.hi = 0;

} /* static void tr_DMULT() */





static void tr_DMULTU()
{
        TR_PRINT_DEBUGGING_INFO("DMULTU", RS_RT);

        reg.lo = (tr_u64)reg.r[C_RS] * (tr_u64)reg.r[C_RT];
/* BUG! */
        reg.hi = 0;

} /* static void tr_DMULTU() */





static void tr_DDIV()
{
        TR_PRINT_DEBUGGING_INFO("DDIV", RS_RT);

        if(reg.r[C_RT] != 0)
        {
                reg.lo = reg.r[C_RS] / reg.r[C_RT];
                reg.hi = reg.r[C_RS] % reg.r[C_RT];
        }

} /* static void tr_DDIV() */





static void tr_DDIVU()
{
        TR_PRINT_DEBUGGING_INFO("DDIVU", RS_RT);

        if((tr_u64)reg.r[C_RT] != 0)
        {
                reg.lo = (tr_u64)reg.r[C_RS] / (tr_u64)reg.r[C_RT];
                reg.hi = (tr_u64)reg.r[C_RS] % (tr_u64)reg.r[C_RT];
        }

} /* static void tr_DDIVU() */





static void tr_ADD()
{
        tr_s32 x, y, z;

        TR_PRINT_DEBUGGING_INFO("ADD", RD_RS_RT);

        x = (tr_s32)reg.r[C_RS];
        y = (tr_s32)reg.r[C_RT];

        z = x + y;

        CHECK_32BIT_OVERFLOW(x, y, z, C_RD);

        reg.r[0] = 0;

} /* static void tr_ADD() */





static void tr_ADDU()
{
        TR_PRINT_DEBUGGING_INFO("ADDU", RD_RS_RT);

        reg.r[C_RD] = (tr_s64)(tr_s32)(reg.r[C_RS] + reg.r[C_RT]);
        reg.r[0] = 0;

} /* static void tr_ADDU() */





static void tr_SUB()
{
        tr_s32 x, y, z;

        TR_PRINT_DEBUGGING_INFO("SUB", RD_RS_RT);

        x = (tr_s32)reg.r[C_RS];
        y = (tr_s32)reg.r[C_RT];

        z = x - y;

        y = ~y;
        CHECK_32BIT_OVERFLOW(x, y, z, C_RD);
        reg.r[0] = 0;

} /* static void tr_SUB() */





static void tr_SUBU()
{
        TR_PRINT_DEBUGGING_INFO("SUBU", RD_RS_RT);

        reg.r[C_RD] = (tr_s64)((tr_s32)reg.r[C_RS] - (tr_s32)reg.r[C_RT]);
        reg.r[0] = 0;

} /* static void tr_SUBU() */





static void tr_AND()
{
        TR_PRINT_DEBUGGING_INFO("AND", RD_RS_RT);

        reg.r[C_RD] = reg.r[C_RS] & reg.r[C_RT];
        reg.r[0] = 0;

} /* static void tr_AND() */





static void tr_OR()
{
        TR_PRINT_DEBUGGING_INFO("OR", RD_RS_RT);

        reg.r[C_RD] = reg.r[C_RS] | reg.r[C_RT];
        reg.r[0] = 0;

} /* static void tr_OR() */





static void tr_XOR()
{
        TR_PRINT_DEBUGGING_INFO("XOR", RD_RS_RT);

        reg.r[C_RD] = reg.r[C_RS] ^ reg.r[C_RT];
        reg.r[0] = 0;

} /* static void tr_XOR() */





static void tr_NOR()
{
        TR_PRINT_DEBUGGING_INFO("NOR", RD_RS_RT);

        reg.r[C_RD] = ~(reg.r[C_RS] | reg.r[C_RT]);
        reg.r[0] = 0;

} /* static void tr_NOR() */





static void tr_SLT()
{
        TR_PRINT_DEBUGGING_INFO("SLT", RD_RS_RT);

        if(reg.r[C_RS] < reg.r[C_RT])
                reg.r[C_RD] = 1;
        else
                reg.r[C_RD] = 0;
        reg.r[0] = 0;

} /* static void tr_SLT() */





static void tr_SLTU()
{
        TR_PRINT_DEBUGGING_INFO("SLTU", RD_RS_RT);

        if((tr_u64)reg.r[C_RS] < (tr_u64)reg.r[C_RT])
                reg.r[C_RD] = 1;
        else
                reg.r[C_RD] = 0;
        reg.r[0] = 0;

} /* static void tr_SLTU() */





static void tr_DADD()
{
        tr_s64 x, y, z;

        TR_PRINT_DEBUGGING_INFO("DADD", RD_RS_RT);

        x = (tr_s64)reg.r[C_RS];
        y = (tr_s64)reg.r[C_RT];

        z = x + y;
        CHECK_64BIT_OVERFLOW(x, y, z, C_RD)
        reg.r[0] = 0;

} /* static void tr_DADD() */





static void tr_DADDU()
{
        TR_PRINT_DEBUGGING_INFO("DADDU", RD_RS_RT);

        reg.r[C_RD] = reg.r[C_RS] + reg.r[C_RT];
        reg.r[0] = 0;

} /* static void tr_DADDU() */





static void tr_DSUB()
{
        tr_s64 x, y, z;

        TR_PRINT_DEBUGGING_INFO("DSUB", RD_RS_RT);

        x = (tr_s64)reg.r[C_RS];
        y = (tr_s64)reg.r[C_RT];

        z = x - y;

        y = ~y;
        CHECK_64BIT_OVERFLOW(x, y, z, C_RD)
        reg.r[0] = 0;

} /* static void tr_DSUB() */





static void tr_DSUBU()
{
        TR_PRINT_DEBUGGING_INFO("DSUBU", RD_RS_RT);

        reg.r[C_RD] = reg.r[C_RS] - reg.r[C_RT];
        reg.r[0] = 0;

} /* static void tr_DSUBU() */





static void tr_TGE()
{
        TR_PRINT_DEBUGGING_INFO("TGE", RS_RT);

        if((tr_s64)reg.r[C_RS] >= (tr_s64)reg.r[C_RT])
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_TGE() */





static void tr_TGEU()
{
        TR_PRINT_DEBUGGING_INFO("TGEU", RS_RT);

        if((tr_u64)reg.r[C_RS] >= (tr_u64)reg.r[C_RT])
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_TGEU() */





static void tr_TLT()
{
        TR_PRINT_DEBUGGING_INFO("TLT", RS_RT);

        if((tr_s64)reg.r[C_RS] < (tr_s64)reg.r[C_RT])
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_TLT() */





static void tr_TLTU()
{
        TR_PRINT_DEBUGGING_INFO("TLTU", RS_RT);

        if((tr_u64)reg.r[C_RS] < (tr_u64)reg.r[C_RT])
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_TLTU() */





static void tr_TEQ()
{
        TR_PRINT_DEBUGGING_INFO("TEQ", RS_RT);

        if(reg.r[C_RS] == reg.r[C_RT])
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_TEQ() */





static void tr_TNE()
{
        TR_PRINT_DEBUGGING_INFO("TNE", RS_RT);

        if(reg.r[C_RS] != reg.r[C_RT])
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_TNE() */





static void tr_DSLL()
{
        TR_PRINT_DEBUGGING_INFO("DSLL", RD_RT_SA);

        reg.r[C_RD] = reg.r[C_RT] << C_SA;
        reg.r[0] = 0;

} /* static void tr_DSLL() */





static void tr_DSRL()
{
        TR_PRINT_DEBUGGING_INFO("DSRL", RD_RT_SA);

        reg.r[C_RD] = (tr_u64)reg.r[C_RT] >> C_SA;
        reg.r[0] = 0;

} /* static void tr_DSRL() */





static void tr_DSRA()
{
        TR_PRINT_DEBUGGING_INFO("DSRA", RD_RT_SA);

        reg.r[C_RD] = reg.r[C_RT] >> C_SA;
        reg.r[0] = 0;

} /* static void tr_DSRA() */





static void tr_DSLL32()
{
        TR_PRINT_DEBUGGING_INFO("DSLL32", RD_RT_SA);

        reg.r[C_RD] = reg.r[C_RT] << (32 + C_SA);
        reg.r[0] = 0;

} /* static void tr_DSLL32() */





static void tr_DSRL32()
{
        TR_PRINT_DEBUGGING_INFO("DSRL32", RD_RT_SA);

        reg.r[C_RD] = (tr_u64)reg.r[C_RT] >> (32 + C_SA);
        reg.r[0] = 0;

} /* static void tr_DSRL32() */





static void tr_DSRA32()
{
        TR_PRINT_DEBUGGING_INFO("DSRA32", RD_RT_SA);

        reg.r[C_RD] = reg.r[C_RT] >> (32 + C_SA);
        reg.r[0] = 0;

} /* static void tr_DSRA32() */






/* core functions */
static void tr_J()
{
        TR_PRINT_DEBUGGING_INFO("J", T);

#ifdef SPEED_HACK
        if(rsp_reg.halt)
        if(C_T == C_PC())   /* if jump to same myself  */
                if((tr_i+1)->r.f == 0x00000000)   /*    if next instr == nop */
                {
                        /*
			*/
			printf("tr_speed_hack: count: %"PR32"x\n", reg.cpr[0][COUNT]);
                        reg.cpr[0][COUNT] = next_interrupt_count - 1;
                        /*
			*/
			printf("tr_speed_hack: count: %"PR32"x\n\n", reg.cpr[0][COUNT]);
                }
#endif

        reg.delay = DELAY_JUMP;
        reg.delay_jump = (C_T << 2) | (reg.delay_jump & 0xf0000000);

} /* static void tr_J() */





static void tr_JAL()
{
        TR_PRINT_DEBUGGING_INFO("JAL", T);

        reg.delay = DELAY_JUMP;
        reg.delay_jump = (C_T << 2) | (reg.delay_jump & 0xf0000000);
        reg.r[31] = (tr_s64)(tr_s32)(C_PC() + 8);

} /* static void tr_JAL() */





static void tr_BEQ()
{
        TR_PRINT_DEBUGGING_INFO("BEQ", RS_RT_I);

        if(reg.r[C_RS] == reg.r[C_RT])
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_BEQ() */





static void tr_BNE()
{
        TR_PRINT_DEBUGGING_INFO("BNE", RS_RT_I);

        if(reg.r[C_RS] != reg.r[C_RT])
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_BNE() */





#ifdef SPEED_HACK
static void tr_BNE_opt()
{
        TR_PRINT_DEBUGGING_INFO("_BNE_ (loop!)", NOTHING);

        reg.delay = DO_LOOP;

} /* static void tr_BNE_opt() */
#endif /* #ifdef SPEED_HACK  */





static void tr_BLEZ()
{
        TR_PRINT_DEBUGGING_INFO("BLEZ", RS_RT_I);

        if((reg.r[C_RS]) <= 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_BLEZ() */





static void tr_BGTZ()
{
        TR_PRINT_DEBUGGING_INFO("BGTZ", RS_RT_I);

        if((reg.r[C_RS]) > 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_BGTZ() */





static void tr_ADDI()
{
        tr_s32 x, y, z;

        TR_PRINT_DEBUGGING_INFO("ADDI", RT_RS_I);

        x = (tr_s32)reg.r[C_RS];
        y = (tr_s32)C_I;

        z = x + y;

        CHECK_32BIT_OVERFLOW(x, y, z, C_RT);

        reg.r[0] = 0;

} /* static void tr_ADDI() */





static void tr_ADDIU()
{
        TR_PRINT_DEBUGGING_INFO("ADDIU", RT_RS_I);

        reg.r[C_RT] = reg.r[C_RS] + C_I;
        reg.r[0] = 0;

} /* static void tr_ADDIU() */





static void tr_SLTI()
{
        TR_PRINT_DEBUGGING_INFO("SLTI", RT_RS_I);

        if(reg.r[C_RS] < (tr_s64)C_I)
                reg.r[C_RT] = 1;
        else
                reg.r[C_RT] = 0;

        reg.r[0] = 0;

} /* static void tr_SLTI() */





static void tr_SLTIU()
{
        TR_PRINT_DEBUGGING_INFO("SLTIU", RT_RS_I);

        if((tr_u64)reg.r[C_RS] < (tr_u64)(tr_s64)C_I)
                reg.r[C_RT] = 1;
        else
                reg.r[C_RT] = 0;

        reg.r[0] = 0;

} /* static void tr_SLTIU() */





static void tr_ANDI()
{
        TR_PRINT_DEBUGGING_INFO("ANDI", RT_RS_I);

        reg.r[C_RT] = reg.r[C_RS] & (tr_u64)(tr_u16)C_I;
        reg.r[0] = 0;

} /* static void tr_ANDI() */





static void tr_ORI()
{
        TR_PRINT_DEBUGGING_INFO("ORI", RT_RS_I);

        reg.r[C_RT] = reg.r[C_RS] | (tr_u64)(tr_u16)C_I;
        reg.r[0] = 0;

} /* static void tr_ORI() */





static void tr_XORI()
{
        TR_PRINT_DEBUGGING_INFO("XORI", RT_RS_I);

        reg.r[C_RT] = reg.r[C_RS] ^ (tr_u64)(tr_u16)C_I;
        reg.r[0] = 0;

} /* static void tr_XORI() */





static void tr_LUI()
{
        TR_PRINT_DEBUGGING_INFO("LUI", RT_I);

        reg.r[C_RT] = (tr_s64)(tr_s32)(C_I << 16);
        reg.r[0] = 0;

} /* static void tr_LUI() */





static void tr_BEQL()
{
        TR_PRINT_DEBUGGING_INFO("BEQL", RS_RT_I);

        if(reg.r[C_RS] == reg.r[C_RT])
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;


} /* static void tr_BEQL() */





static void tr_BNEL()
{
        TR_PRINT_DEBUGGING_INFO("BNEL", RS_RT_I);

        if(reg.r[C_RS] != reg.r[C_RT])
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_BNEL() */





static void tr_BLEZL()
{
        TR_PRINT_DEBUGGING_INFO("BLEZL", RS_RT_I);

        if(((tr_s64)reg.r[C_RS]) <= 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_BLEZL() */





static void tr_BGTZL()
{
        TR_PRINT_DEBUGGING_INFO("BGTZL", RS_RT_I);

        if(((tr_s64)reg.r[C_RS]) > 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_BGTZL() */





static void tr_DADDI()
{
        TR_PRINT_DEBUGGING_INFO("DADDI", RT_RS_I);

        reg.halt = 1;

} /* static void tr_DADDI() */





static void tr_DADDIU()
{
        TR_PRINT_DEBUGGING_INFO("DADDIU", RT_RS_I);

        reg.halt = 1;

} /* static void tr_DADDIU() */





static void tr_LDL()
{
        TR_PRINT_DEBUGGING_INFO("LDL", RT_O_B);

        reg.halt = 1;

} /* static void tr_LDL() */





static void tr_LDR()
{
        TR_PRINT_DEBUGGING_INFO("LDR", RT_O_B);

        reg.halt = 1;

} /* static void tr_LDR() */





static void tr_LB()
{
        TR_PRINT_DEBUGGING_INFO("LB", RT_O_B);

        reg.halt = 1;

} /* static void tr_LB() */





static void tr_LH()
{
        TR_PRINT_DEBUGGING_INFO("LH", RT_O_B);

        reg.halt = 1;

} /* static void tr_LH() */





static void tr_LWL()
{
        TR_PRINT_DEBUGGING_INFO("LWL", RT_O_B);

        reg.halt = 1;

} /* static void tr_LWL() */





static void tr_LW()
{
        TR_PRINT_DEBUGGING_INFO("LW", RT_O_B);

        reg.r[C_RT] = (tr_s64)((tr_s32)doReadMemWord(((tr_s32)reg.r[C_RS]) + C_I));
        /* exception */
        reg.r[0] = 0;

} /* static void tr_LW() */





static void tr_LBU()
{
        TR_PRINT_DEBUGGING_INFO("LBU", RT_O_B);

        reg.r[C_RT] = (tr_s64)((tr_u8)doReadMemByte(((tr_s32)reg.r[C_RS]) + C_I));
        reg.r[0] = 0;
        

} /* static void tr_LBU() */





static void tr_LHU()
{
        TR_PRINT_DEBUGGING_INFO("LHU", RT_O_B);

        reg.r[C_RT] = (tr_s64)((tr_u16)doReadMemHalfWord(((tr_s32)reg.r[C_RS]) + C_I));
        /* exception */
        reg.r[0] = 0;

} /* static void tr_LHU() */





static void tr_LWR()
{
        TR_PRINT_DEBUGGING_INFO("LWR", RT_O_B);

        reg.halt = 1;

} /* static void tr_LWR() */





static void tr_LWU()
{
        TR_PRINT_DEBUGGING_INFO("LWU", RT_O_B);

        reg.halt = 1;

} /* static void tr_LWU() */





static void tr_SB()
{
        TR_PRINT_DEBUGGING_INFO("SB", RT_O_B);

        doWriteMemByte((tr_u8)reg.r[C_RT], ((tr_s32)reg.r[C_RS]) + C_I);

} /* static void tr_SB() */





static void tr_SH()
{
        TR_PRINT_DEBUGGING_INFO("SH", RT_O_B);

        doWriteMemHalfWord((tr_u16)reg.r[C_RT], ((tr_s32)reg.r[C_RS]) + C_I);
        /* exception */

} /* static void tr_SH() */





static void tr_SWL()
{
        TR_PRINT_DEBUGGING_INFO("SWL", RT_O_B);

        reg.halt = 1;

} /* static void tr_SWL() */





static void tr_SW()
{
        tr_u32 addr;
        tr_u32 data;

        TR_PRINT_DEBUGGING_INFO("SW", RT_O_B);

        addr = ((tr_u32)reg.r[C_RS]) + C_I;
        data = (tr_u32)reg.r[C_RT];

        doWriteMemWord(data, addr);

    /* BUG: the next stuff does NOT work with TLB! i hope that the TLB 
       is never used with code! */

        addr &= 0xdfffffff;  /* we mask out the 0xa0000000 stuff
                                and make it look like 0x8000000 */
        if( ((addr >= 0x80000000) && (addr < (0x80000000 + TR_RDMEM_MAX))) )
        {
                addr &= TR_RDMEM_MASK;
                addr >>= 2;
                tr_compile_tab[TR_COMPILE_OPCODE(data)](
                        data,
                        &mem.rd_ram_instr[addr] );
        }

        /* exception */

} /* static void tr_SW() */





static void tr_SDL()
{
        TR_PRINT_DEBUGGING_INFO("SDL", RT_O_B);

        reg.halt = 1;

} /* static void tr_SDL() */





static void tr_SDR()
{
        TR_PRINT_DEBUGGING_INFO("SDR", RT_O_B);

        reg.halt = 1;

} /* static void tr_SDR() */





static void tr_SWR()
{
        TR_PRINT_DEBUGGING_INFO("SWR", RT_O_B);

        reg.halt = 1;

} /* static void tr_SWR() */





static void tr_CACHE()
{
        TR_PRINT_DEBUGGING_INFO("CACHE (ignored!)", RT_O_B);

} /* static void tr_CACHE() */





static void tr_LL()
{
        TR_PRINT_DEBUGGING_INFO("LL", RT_O_B);

        reg.halt = 1;

} /* static void tr_LL() */





static void tr_LWC1()
{
        TR_PRINT_DEBUGGING_INFO("LWC1", RT_O_B);

        reg.cpr[1][C_FT] = (tr_s64)((tr_s32)doReadMemWord(((tr_s32)reg.r[C_RS]) + C_I));
        /* exception */

} /* static void tr_LWC1() */





static void tr_LWC2()
{
        TR_PRINT_DEBUGGING_INFO("LWC2", RT_O_B);

        reg.halt = 1;

} /* static void tr_LWC2() */





static void tr_LLD()
{
        TR_PRINT_DEBUGGING_INFO("LLD", RT_O_B);

        reg.halt = 1;

} /* static void tr_LLD() */





static void tr_LDC1()
{
        TR_PRINT_DEBUGGING_INFO("LDC1", RT_O_B);

        reg.halt = 1;

} /* static void tr_LDC1() */





static void tr_LDC2()
{
        TR_PRINT_DEBUGGING_INFO("LDC2", RT_O_B);

        reg.halt = 1;

} /* static void tr_LDC2() */





static void tr_LD()
{
        TR_PRINT_DEBUGGING_INFO("LD", RT_O_B);

        reg.r[C_RT] = doReadMemDoubleWord(((tr_s32)reg.r[C_RS]) + C_I);
        /* exception */
        reg.r[0] = 0;

} /* static void tr_LD() */





static void tr_SC()
{
        TR_PRINT_DEBUGGING_INFO("SC", RT_O_B);

        reg.halt = 1;

} /* static void tr_SC() */





static void tr_SWC1()
{
        TR_PRINT_DEBUGGING_INFO("SWC1", RT_O_B);

        reg.halt = 1;

} /* static void tr_SWC1() */





static void tr_SWC2()
{
        TR_PRINT_DEBUGGING_INFO("SWC2", RT_O_B);

        reg.halt = 1;

} /* static void tr_SWC2() */





static void tr_SCD()
{
        TR_PRINT_DEBUGGING_INFO("SCD", RT_O_B);

        reg.halt = 1;

} /* static void tr_SCD() */





static void tr_SDC1()
{
        TR_PRINT_DEBUGGING_INFO("SDC1", RT_O_B);

        reg.halt = 1;

} /* static void tr_SDC1() */





static void tr_SDC2()
{
        TR_PRINT_DEBUGGING_INFO("SDC2", RT_O_B);

        reg.halt = 1;

} /* static void tr_SDC2() */





static void tr_SD()
{
        TR_PRINT_DEBUGGING_INFO("SD", RT_O_B);

        doWriteMemDoubleWord(reg.r[C_RT], ((tr_s32)reg.r[C_RS]) + C_I);
        /* exception */

} /* static void tr_SD() */






/* REGister IMMediate functions */
static void tr_regimm_BLTZ()
{
        TR_PRINT_DEBUGGING_INFO("BLTZ", RS_I);

        if(((tr_s64)reg.r[C_RS]) < 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_regimm_BLTZ() */





static void tr_regimm_BGEZ()
{
        TR_PRINT_DEBUGGING_INFO("BGEZ", RS_I);

        if(((tr_s64)reg.r[C_RS]) >= 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_regimm_BGEZ() */





static void tr_regimm_BLTZL()
{
        TR_PRINT_DEBUGGING_INFO("BLTZL", RS_I);

        if(((tr_s64)reg.r[C_RS]) < 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_regimm_BLTZL() */





static void tr_regimm_BGEZL()
{
        TR_PRINT_DEBUGGING_INFO("BGEZL", RS_I);

        if(((tr_s64)reg.r[C_RS]) >= 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_regimm_BGEZL() */





static void tr_regimm_TGEI()
{
        TR_PRINT_DEBUGGING_INFO("TGEI", RS_I);

        if((tr_s64)reg.r[C_RS] >= (tr_s64)(tr_s32)C_I)
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_regimm_TGEI() */





static void tr_regimm_TGEIU()
{
        TR_PRINT_DEBUGGING_INFO("TGEIU", RS_I);

        if((tr_u64)reg.r[C_RS] >= (tr_u64)(tr_s64)(tr_s32)C_I)
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_regimm_TGEIU() */





static void tr_regimm_TLTI()
{
        TR_PRINT_DEBUGGING_INFO("TLTI", RS_I);

        if((tr_s64)reg.r[C_RS] < (tr_s64)(tr_s32)C_I)
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_regimm_TLTI() */





static void tr_regimm_TLTIU()
{
        TR_PRINT_DEBUGGING_INFO("TLTIU", RS_I);

        if((tr_u64)reg.r[C_RS] < (tr_u64)(tr_s64)(tr_s32)C_I)
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_regimm_TLTIU() */





static void tr_regimm_TEQI()
{
        TR_PRINT_DEBUGGING_INFO("TEQI", RS_I);

        if(reg.r[C_RS] == (tr_s64)(tr_s32)C_I)
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_regimm_TEQI() */





static void tr_regimm_TNEI()
{
        TR_PRINT_DEBUGGING_INFO("TNEI", RS_I);

        if(reg.r[C_RS] != (tr_u64)(tr_s64)(tr_s32)C_I)
        {
                reg.do_or_check_sthg |= RS4300I_DO_TRAP_EXCEPTION;
        }

} /* static void tr_regimm_TNEI() */





static void tr_regimm_BLTZAL()
{
        TR_PRINT_DEBUGGING_INFO("BLTZAL", RS_I);

        reg.r[31] = (tr_s64)(tr_s32)(C_PC() + 8);

        if(((tr_s64)reg.r[C_RS]) < 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }


} /* static void tr_regimm_BLTZAL() */





static void tr_regimm_BGEZAL()
{
        TR_PRINT_DEBUGGING_INFO("BGEZAL", RS_I);

        reg.r[31] = (tr_s64)(tr_s32)(C_PC() + 8);

        if(((tr_s64)reg.r[C_RS]) >= 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }

} /* static void tr_regimm_BGEZAL() */





static void tr_regimm_BLTZALL()
{
        TR_PRINT_DEBUGGING_INFO("BLTZALL", RS_I);

        reg.r[31] = (tr_s64)(tr_s32)(C_PC() + 8);

        if(((tr_s64)reg.r[C_RS]) < 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_regimm_BLTZALL() */





static void tr_regimm_BGEZALL()
{
        TR_PRINT_DEBUGGING_INFO("BGEZALL", RS_I);

        reg.r[31] = (tr_s64)(tr_s32)(C_PC() + 8);

        if(((tr_s64)reg.r[C_RS]) >= 0)
        {
                reg.delay = DELAY_BRANCH;
                reg.delay_branch = C_I;
        }
        else
                tr_i++;

} /* static void tr_regimm_BGEZALL() */






/* cop0 (TranslationLookasideBuffer / ExceptionRETurn) */
static void tr_MFC0()
{
        TR_PRINT_DEBUGGING_INFO("MFC0", RT_RD);

        reg.r[C_RT] = (tr_s64)(tr_s32)reg.cpr[0][C_RD];
        /* cop0 unusable exception */
        reg.r[0] = 0;

} /* static void tr_MFC0() */





static void tr_DMFC0()
{
        TR_PRINT_DEBUGGING_INFO("DMFC0", RT_RD);

        reg.halt = 1;

} /* static void tr_DMFC0() */





static void tr_CFC0()
{
        TR_PRINT_DEBUGGING_INFO("CFC0", RT_RD);

        reg.halt = 1;

} /* static void tr_CFC0() */





static void tr_MTC0()
{
        TR_PRINT_DEBUGGING_INFO("MTC0", RT_RD);

        if(C_RD == COUNT)
        {
                rs4300i_VI_or_CP_calc();
        }

        /* if write to CAUSE reg mask out all bits except IP0 & IP1 */
        if(C_RD == CAUSE)
        {
                reg.cpr[0][CAUSE] = reg.r[C_RT] & 0x00000300;
                return;
        }

        reg.cpr[0][C_RD] = reg.r[C_RT];

        /**
         * if
         *    write to STATUS reg &&
         *    write included any interrupt bits
         *
         * check for cpu interrupts
        **/

        if((C_RD == STATUS) && (reg.r[C_RT] & 0xff01))
                reg.do_or_check_sthg |= RS4300I_CHECK_CPU_INTERRUPTS;

        /* if write to COMPARE reg clr IP7 (timer interrupt) in CAUSE reg */
        if(C_RD == COMPARE)
        {
                reg.cpr[0][CAUSE] &= ~0x00008000;
                rs4300i_VI_or_CP_calc();
        }

} /* static void tr_MTC0() */





static void tr_DMTC0()
{
        TR_PRINT_DEBUGGING_INFO("DMTC0", RT_RD);

        reg.halt = 1;

} /* static void tr_DMTC0() */





static void tr_CTC0()
{
        TR_PRINT_DEBUGGING_INFO("CTC0", RT_RD);

        reg.halt = 1;

} /* static void tr_CTC0() */





static void tr_BC0F()
{
        TR_PRINT_DEBUGGING_INFO("BC0F", I);

        reg.halt = 1;

} /* static void tr_BC0F() */





static void tr_BC0T()
{
        TR_PRINT_DEBUGGING_INFO("BC0T", I);

        reg.halt = 1;

} /* static void tr_BC0T() */





static void tr_BC0FL()
{
        TR_PRINT_DEBUGGING_INFO("BC0FL", I);

        reg.halt = 1;

} /* static void tr_BC0FL() */





static void tr_BC0TL()
{
        TR_PRINT_DEBUGGING_INFO("BC0TL", I);

        reg.halt = 1;

} /* static void tr_BC0TL() */





static void tr_cop0_TLBR()
{
        tr_u64 index;

        TR_PRINT_DEBUGGING_INFO("TLBR", NOTHING);

        index = reg.cpr[0][INDEX] & 0x1f;

        reg.cpr[0][PAGEMASK] = reg.tlb[index].hh;
        reg.cpr[0][ENTRYHI]  = reg.tlb[index].hl & (~reg.tlb[index].hh);
        reg.cpr[0][ENTRYLO1] = reg.tlb[index].lh | reg.tlb[index].g;
        reg.cpr[0][ENTRYLO0] = reg.tlb[index].ll | reg.tlb[index].g;

        /* coprozessor unusable exception */

} /* static void tr_cop0_TLBR() */





static void tr_cop0_TLBWI()
{
        struct rs4300i_tlb *tlb;



        TR_PRINT_DEBUGGING_INFO("TLBWI", NOTHING);



        tlb = &(reg.tlb[reg.cpr[0][INDEX]]);

        tlb->hh = reg.cpr[0][PAGEMASK];
        tlb->hl = reg.cpr[0][ENTRYHI];
        tlb->lh = reg.cpr[0][ENTRYLO1];
        tlb->ll = reg.cpr[0][ENTRYLO0];

        tlb->g       = 0x01 & tlb->lh & tlb->ll;
        tlb->mask    = tlb->hh | 0x00001fff;
        tlb->invmask = ~(tlb->mask);

        switch(tlb->hh)
        {
            case 0x00000000: /*   4k pagesize */
                tlb->even_odd_checkbit = 0x00001000;   /* bit 12 */
                break;
            case 0x00006000: /*  16k pagesize */
                tlb->even_odd_checkbit = 0x00004000;   /* bit 14 */
                break;
            case 0x0001e000: /*  64k pagesize */
                tlb->even_odd_checkbit = 0x00010000;   /* bit 16 */
                break;
            case 0x0007e000: /* 256k pagesize */
                tlb->even_odd_checkbit = 0x00040000;   /* bit 18 */
                break;
            case 0x001fe000: /*   1M pagesize */
                tlb->even_odd_checkbit = 0x00100000;   /* bit 20 */
                break;
            case 0x007fe000: /*   4M pagesize */
                tlb->even_odd_checkbit = 0x00400000;   /* bit 22 */
                break;
            case 0x01ffe000: /*  16M pagesize */
                tlb->even_odd_checkbit = 0x01000000;   /* bit 24 */
                break;
            default: /* should not happen! */
                tlb->even_odd_checkbit = 0;
                break;
        }
/*
        printf("   mask: %08"P32"x\n", tlb->hh);
        printf("   vpn2: %08"P32"x\n", tlb->hl);
        printf("   pfn1: %08"P32"x\n", tlb->lh);
        printf("   pfn0: %08"P32"x\n", tlb->ll);
*/
        /* coprozessor unusable exception */

        /* CPU_HALT */

} /* static void tr_cop0_TLBWI() */





static void tr_cop0_TLBWR()
{
        TR_PRINT_DEBUGGING_INFO("TLBWR", NOTHING);

        reg.halt = 1;

} /* static void tr_cop0_TLBWR() */





static void tr_cop0_TLBP()
{
        int i;
        
        TR_PRINT_DEBUGGING_INFO("TLBP", NOTHING);

        reg.cpr[0][INDEX] = 0x80000000;

        for(i=0; i<32; i++)
        {
                if
                (
                        ( (reg.tlb[i].hl & ~0x1fff) == (reg.cpr[0][ENTRYHI] & ~0x1fff) )
                        &&
                        (
                                reg.tlb[i].g
                                ||
                                ( (reg.tlb[i].hl & 0xff) == (reg.cpr[0][ENTRYHI] & 0xff) )
                        )
                )
                {
                        reg.cpr[0][INDEX] = i;
                        break;
                }

        } /* for(i=0; i<32; i++) */

        /* coprozessor unusable exception */

} /* static void tr_cop0_TLBP() */





static void tr_cop0_ERET()
{
        TR_PRINT_DEBUGGING_INFO("ERET", NOTHING);

        if(reg.cpr[0][STATUS] & 0x0004)   /* if(ERL exception) */
        {
                reg.delay_jump = reg.cpr[0][ERROREPC];   /* PC = ErrorEPC */
                reg.cpr[0][STATUS] &= TR_MASK_ERL;   /* STATUS -= ERL */
        }
        else   /* else ("normal" exception) */
        {
                reg.delay_jump = reg.cpr[0][EPC];        /* PC = EPC */
                reg.cpr[0][STATUS] &= TR_MASK_EXL;   /* STATUS -= EXL */
        }

        reg.delay = DELAY_JUMP_EXEC;
        reg.llbit = 0;

	/*
	reg.halt = 1;
	*/

} /* static void tr_cop0_ERET() */






/* cop1 (FloatingPointUnit) */
static void tr_MFC1()
{
        TR_PRINT_DEBUGGING_INFO("MFC1", RT_RD);

        reg.r[C_RT] = (tr_s64)(tr_s32)FS;
        /* cop1 unusable exception */
        reg.r[0] = 0;

} /* static void tr_MFC1() */





static void tr_DMFC1()
{
        TR_PRINT_DEBUGGING_INFO("DMFC1", RT_RD);

        reg.halt = 1;

} /* static void tr_DMFC1() */





static void tr_CFC1()
{
        TR_PRINT_DEBUGGING_INFO("CFC1", RT_RD);

        reg.r[C_RT] = (tr_s64)(tr_s32)reg.ccr[1][C_FS];
        /* cop1 unusable exception */
        reg.r[0] = 0;
                

} /* static void tr_CFC1() */





static void tr_MTC1()
{
        TR_PRINT_DEBUGGING_INFO("MTC1", RT_RD);

        reg.cpr[1][C_FS] = (tr_s64)(tr_s32)reg.r[C_RT];
        /* cop1 unusable exception */

} /* static void tr_MTC1() */





static void tr_DMTC1()
{
        TR_PRINT_DEBUGGING_INFO("DMTC1", RT_RD);

        reg.halt = 1;

} /* static void tr_DMTC1() */





static void tr_CTC1()
{
        TR_PRINT_DEBUGGING_INFO("CTC1", RT_RD);

        reg.ccr[1][C_FS] = reg.r[C_RT];
        /* cop1 unusable exception */

} /* static void tr_CTC1() */





static void tr_BC1F()
{
        TR_PRINT_DEBUGGING_INFO("BC1F", I);

        reg.halt = 1;

} /* static void tr_BC1F() */





static void tr_BC1T()
{
        TR_PRINT_DEBUGGING_INFO("BC1T", I);

        reg.halt = 1;

} /* static void tr_BC1T() */





static void tr_BC1FL()
{
        TR_PRINT_DEBUGGING_INFO("BC1FL", I);

        reg.halt = 1;

} /* static void tr_BC1FL() */





static void tr_BC1TL()
{
        TR_PRINT_DEBUGGING_INFO("BC1TL", I);

        reg.halt = 1;

} /* static void tr_BC1TL() */





static void tr_cop1_ADD_s()
{
        tr_s64 x, y, z;

        TR_PRINT_DEBUGGING_INFO("ADD.s", FD_FS_FT);

        x = FS;
        y = FT;

#ifdef TR_ENDIAN_IS_LITTLE
        *((float *)&z) = *((float *)&x) + *((float *)&y);
        ((tr_s32 *)&z)[1] = 0;
#endif
#ifdef TR_ENDIAN_IS_BIG
        ((float *)&z)[1] = ((float *)&x)[1] + ((float *)&y)[1];
        ((tr_s32 *)&z)[0] = 0;
#endif

        STORE_FPR(z);

        /* cop unusable exception */
        /* floating point exception */

        /* unimplemented operation exception */
        /* invalid operation exception */
        /* inexact exception */
        /* overflow exception */
        /* underflow exception */

} /* static void tr_cop1_ADD.s() */

static void tr_cop1_ADD_d()
{
        tr_s64 x, y, z;

        TR_PRINT_DEBUGGING_INFO("ADD.d", FD_FS_FT);

        x = FS;
        y = FT;

        *((double *)&z) = *((double *)&x) + *((double *)&y);

        STORE_FPR(z);

        /* cop unusable exception */
        /* floating point exception */

        /* unimplemented operation exception */
        /* invalid operation exception */
        /* inexact exception */
        /* overflow exception */
        /* underflow exception */

} /* static void tr_cop1_ADD_d() */





static void tr_cop1_SUB_s()
{
        TR_PRINT_DEBUGGING_INFO("SUB.s", FD_FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_SUB_s() */

static void tr_cop1_SUB_d()
{
        TR_PRINT_DEBUGGING_INFO("SUB.d", FD_FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_SUB_d() */





static void tr_cop1_MUL_s()
{
        tr_s64 x, y, z;

        TR_PRINT_DEBUGGING_INFO("MUL.s", FD_FS_FT);

        x = FS;
        y = FT;

#ifdef TR_ENDIAN_IS_LITTLE
        *((float *)&z) = *((float *)&x) * *((float *)&y);
        ((tr_s32 *)&z)[1] = 0;
#endif
#ifdef TR_ENDIAN_IS_BIG
        ((float *)&z)[1] = ((float *)&x)[1] * ((float *)&y)[1];
        ((tr_s32 *)&z)[0] = 0;
#endif

        STORE_FPR(z);

        /* cop unusable exception */
        /* floating point exception */

        /* unimplemented operation exception */
        /* invalid operation exception */
        /* inexact exception */
        /* overflow exception */
        /* underflow exception */

} /* static void tr_cop1_MUL_s() */

static void tr_cop1_MUL_d()
{
        TR_PRINT_DEBUGGING_INFO("MUL.d", FD_FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_MUL_d() */





static void tr_cop1_DIV_s()
{
        tr_s64 x, y, z;

        TR_PRINT_DEBUGGING_INFO("DIV.s", FD_FS_FT);

        x = FS;
        y = FT;

#ifdef TR_ENDIAN_IS_LITTLE
        if(*((float *)&y) != 0)
        {
                *((float *)&z) = *((float *)&x) / *((float *)&y);
                ((tr_s32 *)&z)[1] = 0;
#endif
#ifdef TR_ENDIAN_IS_BIG
        if(((float *)&y)[1] != 0)
        {

                ((float *)&z)[1] = ((float *)&x)[1] / ((float *)&y)[1];
                ((tr_s32 *)&z)[0] = 0;
#endif
        }
        else
        {
                printf("DIVISION BY ZERO");
                /* division-by-zero exception */
                return;
        }

	STORE_FPR(z);

        /* cop unusable exception */
        /* floating point exception */

        /* unimplemented operation exception */
        /* invalid operation exception */
        /* inexact exception */
        /* overflow exception */
        /* underflow exception */

} /* static void tr_cop1_DIV_s() */

static void tr_cop1_DIV_d()
{
        TR_PRINT_DEBUGGING_INFO("DIV.d", FD_FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_DIV_d() */





static void tr_cop1_SQRT_s()
{
        TR_PRINT_DEBUGGING_INFO("SQRT.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_SQRT_s() */

static void tr_cop1_SQRT_d()
{
        TR_PRINT_DEBUGGING_INFO("SQRT.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_SQRT_d() */





static void tr_cop1_ABS_s()
{
        TR_PRINT_DEBUGGING_INFO("ABS.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_ABS_s() */

static void tr_cop1_ABS_d()
{
        TR_PRINT_DEBUGGING_INFO("ABS.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_ABS_d() */





static void tr_cop1_MOV_s()
{
        TR_PRINT_DEBUGGING_INFO("MOV.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_MOV_s() */

static void tr_cop1_MOV_d()
{
        TR_PRINT_DEBUGGING_INFO("MOV.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_MOV_d() */





static void tr_cop1_NEG_s()
{
        TR_PRINT_DEBUGGING_INFO("NEG.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_NEG_s() */

static void tr_cop1_NEG_d()
{
        TR_PRINT_DEBUGGING_INFO("NEG.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_NEG_d() */





static void tr_cop1_ROUNDl_s()
{
        TR_PRINT_DEBUGGING_INFO("ROUND.L.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_ROUNDl_s() */

static void tr_cop1_ROUNDl_d()
{
        TR_PRINT_DEBUGGING_INFO("ROUND.L.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_ROUNDl_d() */





static void tr_cop1_TRUNCl_s()
{
        TR_PRINT_DEBUGGING_INFO("TRUNC.L.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_TRUNCl_s() */

static void tr_cop1_TRUNCl_d()
{
        TR_PRINT_DEBUGGING_INFO("TRUNC.L.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_TRUNCl_d() */





static void tr_cop1_CEILl_s()
{
        TR_PRINT_DEBUGGING_INFO("CEIL.L.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CEILl_s() */

static void tr_cop1_CEILl_d()
{
        TR_PRINT_DEBUGGING_INFO("CEIL.L.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CEILl_d() */





static void tr_cop1_FLOORl_s()
{
        TR_PRINT_DEBUGGING_INFO("FLOOR.L.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_FLOORl_s() */

static void tr_cop1_FLOORl_d()
{
        TR_PRINT_DEBUGGING_INFO("FLOOR.L.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_FLOORl_d() */





static void tr_cop1_ROUNDw_s()
{
        TR_PRINT_DEBUGGING_INFO("ROUND.W.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_ROUNDw_s() */

static void tr_cop1_ROUNDw_d()
{
        tr_s64 x, z;
        double fract;

        TR_PRINT_DEBUGGING_INFO("ROUND.W.d", FD_FS);

        x = FS;

/* corrected: BUG!!! used modf in a wrong way! - shitt - niki */

#ifdef TR_ENDIAN_IS_LITTLE
        fract = modf(*((double *)&x), ((tr_s32 *)&z));
        if(fract >  0.5)
                *((tr_s32 *)&z) = *((tr_s32 *)&z) + 1;
        else if(fract < -0.5)
                *((tr_s32 *)&z) = *((tr_s32 *)&z) - 1;
        else if(abs(fract) == 0.5)
#ifdef TR_64BIT_LONG
                *((tr_s32 *)&z) = *((tr_s32 *)&z) & ~1L;
#else
                *((tr_s32 *)&z) = *((tr_s32 *)&z) & ~1LL;
#endif
#endif
#ifdef TR_ENDIAN_IS_BIG
        fract = (tr_s32)modf(*((double *)&x), ((tr_s32 *)&z)[1]);
        if(fract >  0.5)
                ((tr_s32 *)&z)[1] = ((tr_s32 *)&z)[1] + 1;
         else if(fract < -0.5)
                ((tr_s32 *)&z)[1] = ((tr_s32 *)&z)[1] - 1;
         else if(abs(fract) == 0.5)
#ifdef TR_64BIT_LONG
                ((tr_s32 *)&z)[1] = ((tr_s32 *)&z)[1] & ~1L;
#else
                ((tr_s32 *)&z)[1] = ((tr_s32 *)&z)[1] & ~1LL;
#endif
#endif

        STORE_FPR((tr_s64)(tr_s32)z);

        /* cop unusable exception */
        /* floating point exception */

        /* unimplemented operation exception */
        /* invalid operation exception */
        /* inexact exception */
        /* overflow exception */

} /* static void tr_cop1_ROUNDw_d() */





static void tr_cop1_TRUNCw_s()
{
        TR_PRINT_DEBUGGING_INFO("TRUNC.W.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_TRUNCw_s() */

static void tr_cop1_TRUNCw_d()
{
        TR_PRINT_DEBUGGING_INFO("TRUNC.W.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_TRUNCw_d() */





static void tr_cop1_CEILw_s()
{
        TR_PRINT_DEBUGGING_INFO("CEIL.W.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CEILw_s() */

static void tr_cop1_CEILw_d()
{
        TR_PRINT_DEBUGGING_INFO("CEIL.W.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CEILw_d() */





static void tr_cop1_FLOORw_s()
{
        TR_PRINT_DEBUGGING_INFO("FLOOR.W.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_FLOORw_s() */

static void tr_cop1_FLOORw_d()
{
        TR_PRINT_DEBUGGING_INFO("FLOOR.W.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_FLOORw_d() */





static void tr_cop1_CVTs_d()
{
        TR_PRINT_DEBUGGING_INFO("CVT.S.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTs_d() */

static void tr_cop1_CVTs_w()
{
        tr_s64 z;

        TR_PRINT_DEBUGGING_INFO("CVT.S.w", FD_FS);

        z = FS;

#ifdef TR_ENDIAN_IS_LITTLE
        *((float *)&z) = (float)(*((tr_s32 *)&z));
#endif
#ifdef TR_ENDIAN_IS_BIG
        ((float *)&z)[1] = (float)(((tr_s32 *)&z)[1]);
#endif

    /* clear the upper 32 bits */
#ifdef TR_ENDIAN_IS_LITTLE
        ((tr_s32 *)&z)[1] = 0;
#endif
#ifdef TR_ENDIAN_IS_BIG
        ((tr_s32 *)&z)[0] = 0;
#endif

        STORE_FPR(z);

        /* cop unusable exception */
        /* cop exception trap */

        /* unimplemented operation exception */
        /* invalid operation exception */

} /* static void tr_cop1_CVTs_w() */

static void tr_cop1_CVTs_l()
{
        TR_PRINT_DEBUGGING_INFO("CVT.S.l", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTs_l() */





static void tr_cop1_CVTd_s()
{
        tr_s64 z;

        TR_PRINT_DEBUGGING_INFO("CVT.D.s", FD_FS);

        z = FS;

#ifdef TR_ENDIAN_IS_LITTLE
        *((double *)&z) = (double)(*((float *)&z));
#endif
#ifdef TR_ENDIAN_IS_BIG
        *((double *)&z) = (double)(((float *)&z)[1]);
#endif

        STORE_FPR(z);

        /* cop unusable exception */
        /* cop exception trap */

        /* unimplemented operation exception */
        /* invalid operation exception */

} /* static void tr_cop1_CVTd_s() */

static void tr_cop1_CVTd_w()
{
        TR_PRINT_DEBUGGING_INFO("CVT.D.w", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTd_w() */

static void tr_cop1_CVTd_l()
{
        TR_PRINT_DEBUGGING_INFO("CVT.D.l", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTd_l() */





static void tr_cop1_CVTw_s()
{
        TR_PRINT_DEBUGGING_INFO("CVT.W.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTw_s() */

static void tr_cop1_CVTw_d()
{
        TR_PRINT_DEBUGGING_INFO("CVT.W.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTw_d() */





static void tr_cop1_CVTl_s()
{
        TR_PRINT_DEBUGGING_INFO("CVT.L.s", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTl_s() */

static void tr_cop1_CVTl_d()
{
        TR_PRINT_DEBUGGING_INFO("CVT.L.d", FD_FS);

        reg.halt = 1;

} /* static void tr_cop1_CVTl_d() */





static void tr_cop1_CF_s()
{
        TR_PRINT_DEBUGGING_INFO("C.F.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CF_s() */

static void tr_cop1_CF_d()
{
        TR_PRINT_DEBUGGING_INFO("C.F.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CF_d() */





static void tr_cop1_CUN_s()
{
        TR_PRINT_DEBUGGING_INFO("C.UN.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CUN_s() */

static void tr_cop1_CUN_d()
{
        TR_PRINT_DEBUGGING_INFO("C.UN.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CUN_d() */





static void tr_cop1_CEQ_s()
{
        TR_PRINT_DEBUGGING_INFO("C.EQ.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CEQ_s() */

static void tr_cop1_CEQ_d()
{
        TR_PRINT_DEBUGGING_INFO("C.EQ.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CEQ_d() */





static void tr_cop1_CUEQ_s()
{
        TR_PRINT_DEBUGGING_INFO("C.UEQ.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CUEQ_s() */

static void tr_cop1_CUEQ_d()
{
        TR_PRINT_DEBUGGING_INFO("C.UEQ.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CUEQ_d() */





static void tr_cop1_COLT_s()
{
        TR_PRINT_DEBUGGING_INFO("C.OLT.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_COLT_s() */

static void tr_cop1_COLT_d()
{
        TR_PRINT_DEBUGGING_INFO("C.OLT.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_COLT_d() */





static void tr_cop1_CULT_s()
{
        TR_PRINT_DEBUGGING_INFO("C.ULT.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CULT_s() */

static void tr_cop1_CULT_d()
{
        TR_PRINT_DEBUGGING_INFO("C.ULT.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CULT_d() */





static void tr_cop1_COLE_s()
{
        TR_PRINT_DEBUGGING_INFO("C.OLE.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_COLE_s() */

static void tr_cop1_COLE_d()
{
        TR_PRINT_DEBUGGING_INFO("C.OLE.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_COLE_d() */





static void tr_cop1_CULE_s()
{
        TR_PRINT_DEBUGGING_INFO("C.ULE.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CULE_s() */

static void tr_cop1_CULE_d()
{
        TR_PRINT_DEBUGGING_INFO("C.ULE.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CULE_d() */





static void tr_cop1_CSF_s()
{
        TR_PRINT_DEBUGGING_INFO("C.SF.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CSF_s() */

static void tr_cop1_CSF_d()
{
        TR_PRINT_DEBUGGING_INFO("C.SF.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CSF_d() */





static void tr_cop1_CNGLE_s()
{
        TR_PRINT_DEBUGGING_INFO("C.NGLE.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGLE_s() */

static void tr_cop1_CNGLE_d()
{
        TR_PRINT_DEBUGGING_INFO("C.NGLE.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGLE_d() */





static void tr_cop1_CSEQ_s()
{
        TR_PRINT_DEBUGGING_INFO("C.SEQ.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CSEQ_s() */

static void tr_cop1_CSEQ_d()
{
        TR_PRINT_DEBUGGING_INFO("C.SEQ.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CSEQ_d() */





static void tr_cop1_CNGL_s()
{
        TR_PRINT_DEBUGGING_INFO("C.NGL.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGL_s() */

static void tr_cop1_CNGL_d()
{
        TR_PRINT_DEBUGGING_INFO("C.NGL.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGL_d() */





static void tr_cop1_CLT_s()
{
        TR_PRINT_DEBUGGING_INFO("C.LT.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CLT_s() */

static void tr_cop1_CLT_d()
{
        TR_PRINT_DEBUGGING_INFO("C.LT.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CLT_d() */





static void tr_cop1_CNGE_s()
{
        TR_PRINT_DEBUGGING_INFO("C.NGE.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGE_s() */

static void tr_cop1_CNGE_d()
{
        TR_PRINT_DEBUGGING_INFO("C.NGE.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGE_d() */





static void tr_cop1_CLE_s()
{
        TR_PRINT_DEBUGGING_INFO("C.LE.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CLE_s() */

static void tr_cop1_CLE_d()
{
        TR_PRINT_DEBUGGING_INFO("C.LE.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CLE_d() */





static void tr_cop1_CNGT_s()
{
        TR_PRINT_DEBUGGING_INFO("C.NGT.s", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGT_s() */

static void tr_cop1_CNGT_d()
{
        TR_PRINT_DEBUGGING_INFO("C.NGT.d", FS_FT);

        reg.halt = 1;

} /* static void tr_cop1_CNGT_d() */




























static void gen_reserved(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "reserved", NOTHING);

        dest->d.f = tr_reserved;
        dest->d.c = code;

} /* static void gen_reserved(tr_u32 code, tr_instruction *dest) */










/* special functions */


/* Shift Left Logical */

static void gen_SLL(tr_u32 code, tr_instruction *dest)
{
        if( (M_RD(code) == 0) )
        {
                /* if we write into r0 -> NOP */

                TR_PRINT_DEBUGGING_GEN_INFO(code, "_NOP_", NOTHING);
                dest->r.f = tr_NOP_opt;
        }
/* i think that the next `else if' is not used often - niki */
#if 0
        else if( (M_SA(code) == 0) )
        {
                /* if we shift 0 bits -> just do the sign extension */

                TR_PRINT_DEBUGGING_GEN_INFO(code, "_SLL_", RD);
                dest->r.f = tr_SLL_opt;
        }
#endif
        else if( (M_RT(code) == 0) )
        {
                /* if the target reg = r0 -> just clear rd */

                TR_PRINT_DEBUGGING_GEN_INFO(code, "_CLR_", RD);
                dest->r.f = tr_CLR_opt;
                dest->r.rd = D_RD(code);
        }
        else
        {
                /* this is the real `Shift Left Logical' */

                TR_PRINT_DEBUGGING_GEN_INFO(code, "SLL", RD_RT_SA);
                dest->r.f  = tr_SLL;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                dest->r.sa = D_SA(code);
        }

} /* static void gen_SLL(tr_u32 code, tr_instruction *dest) */





/* Shift Right Logical */

static void gen_SRL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SRL", RD_RT_SA);

        dest->r.f  = tr_SRL;
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);
        dest->r.sa = D_SA(code);

} /* static void gen_SRL(tr_u32 code, tr_instruction *dest) */





/* Shift Right Arithmetic */

static void gen_SRA(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SRA", RD_RT_SA);

        dest->r.f  = tr_SRA;
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);
        dest->r.sa = D_SA(code);

} /* static void gen_SRA(tr_u32 code, tr_instruction *dest) */





/* Shift Left Logical Variable */

static void gen_SLLV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SLLV", RD_RT_RS);

        dest->r.f  = tr_SLLV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SLLV(tr_u32 code, tr_instruction *dest) */





/* Shift Right Logical Variable */

static void gen_SRLV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SRLV", RD_RT_RS);

        dest->r.f  = tr_SRLV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SRLV(tr_u32 code, tr_instruction *dest) */





/* Shift Right Arithmetic Variable */

static void gen_SRAV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SRAV", RD_RT_RS);

        dest->r.f  = tr_SRAV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SRAV(tr_u32 code, tr_instruction *dest) */





static void gen_JR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "JR", RS);

        dest->r.f  = tr_JR;
        dest->r.rs = D_RS(code);

} /* static void gen_JR(tr_u32 code, tr_instruction *dest) */





static void gen_JALR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "JALR", RD_RS);

        dest->r.f  = tr_JALR;
        dest->r.rs = D_RS(code);
        dest->r.rd = D_RD(code);

} /* static void gen_JALR(tr_u32 code, tr_instruction *dest) */





static void gen_SYSCALL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SYSCALL", NOTHING);

        dest->r.f  = tr_SYSCALL;

} /* static void gen_SYSCALL(tr_u32 code, tr_instruction *dest) */





static void gen_BREAK(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BREAK", NOTHING);

        dest->r.f  = tr_BREAK;

} /* static void gen_BREAK(tr_u32 code, tr_instruction *dest) */





static void gen_SYNC(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SYNC", NOTHING);

        dest->r.f  = tr_SYNC;

} /* static void gen_SYNC(tr_u32 code, tr_instruction *dest) */





static void gen_MFHI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "MFHI", RD);

        dest->r.f  = tr_MFHI;
        dest->r.rd = D_RD(code);

} /* static void gen_MFHI(tr_u32 code, tr_instruction *dest) */





static void gen_MTHI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "MTHI", RS);

        dest->r.f  = tr_MTHI;
        dest->r.rs = D_RS(code);

} /* static void gen_MTHI(tr_u32 code, tr_instruction *dest) */





static void gen_MFLO(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "MFLO", RD);

        dest->r.f  = tr_MFLO;
        dest->r.rd = D_RD(code);

} /* static void gen_MFLO(tr_u32 code, tr_instruction *dest) */





static void gen_MTLO(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "MTLO", RS);

        dest->r.f  = tr_MTLO;
        dest->r.rs = D_RS(code);

} /* static void gen_MTLO(tr_u32 code, tr_instruction *dest) */





static void gen_DSLLV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSLLV", RD_RT_RS);

        dest->r.f  = tr_DSLLV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DSLLV(tr_u32 code, tr_instruction *dest) */





static void gen_DSRLV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSRLV", RD_RT_RS);

        dest->r.f  = tr_DSRLV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DSRLV(tr_u32 code, tr_instruction *dest) */





static void gen_DSRAV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSRAV", RD_RT_RS);

        dest->r.f  = tr_DSRAV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DSRAV(tr_u32 code, tr_instruction *dest) */





static void gen_MULT(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "MULT", RS_RT);

        dest->r.f  = tr_MULT;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_MULT(tr_u32 code, tr_instruction *dest) */





static void gen_MULTU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "MULTU", RS_RT);

        dest->r.f  = tr_MULTU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_MULTU(tr_u32 code, tr_instruction *dest) */





static void gen_DIV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DIV", RS_RT);

        dest->r.f  = tr_DIV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_DIV(tr_u32 code, tr_instruction *dest) */





static void gen_DIVU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DIVU", RS_RT);

        dest->r.f  = tr_DIVU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_DIVU(tr_u32 code, tr_instruction *dest) */





static void gen_DMULT(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DMULT", RS_RT);

        dest->r.f  = tr_DMULT;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_DMULT(tr_u32 code, tr_instruction *dest) */





static void gen_DMULTU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DMULTU", RS_RT);

        dest->r.f  = tr_DMULTU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_DMULTU(tr_u32 code, tr_instruction *dest) */





static void gen_DDIV(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DDIV", RS_RT);

        dest->r.f  = tr_DDIV;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_DDIV(tr_u32 code, tr_instruction *dest) */





static void gen_DDIVU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DDIVU", RS_RT);

        dest->r.f  = tr_DDIVU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_DDIVU(tr_u32 code, tr_instruction *dest) */





static void gen_ADD(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "ADD", RD_RS_RT);

        dest->r.f  = tr_ADD;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_ADD(tr_u32 code, tr_instruction *dest) */





static void gen_ADDU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "ADDU", RD_RS_RT);

        dest->r.f  = tr_ADDU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_ADDU(tr_u32 code, tr_instruction *dest) */





static void gen_SUB(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SUB", RD_RS_RT);

        dest->r.f  = tr_SUB;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SUB(tr_u32 code, tr_instruction *dest) */





static void gen_SUBU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SUBU", RD_RS_RT);

        dest->r.f  = tr_SUBU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SUBU(tr_u32 code, tr_instruction *dest) */





static void gen_AND(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "AND", RD_RS_RT);

        dest->r.f  = tr_AND;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_AND(tr_u32 code, tr_instruction *dest) */





static void gen_OR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "OR", RD_RS_RT);

        dest->r.f  = tr_OR;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_OR(tr_u32 code, tr_instruction *dest) */





static void gen_XOR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "XOR", RD_RS_RT);

        dest->r.f  = tr_XOR;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_XOR(tr_u32 code, tr_instruction *dest) */





static void gen_NOR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "NOR", RD_RS_RT);

        dest->r.f  = tr_NOR;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_NOR(tr_u32 code, tr_instruction *dest) */





static void gen_SLT(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SLT", RD_RS_RT);

        dest->r.f  = tr_SLT;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SLT(tr_u32 code, tr_instruction *dest) */





static void gen_SLTU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SLTU", RD_RS_RT);

        dest->r.f  = tr_SLTU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_SLTU(tr_u32 code, tr_instruction *dest) */





static void gen_DADD(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DADD", RD_RS_RT);

        dest->r.f  = tr_DADD;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DADD(tr_u32 code, tr_instruction *dest) */





static void gen_DADDU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DADDU", RD_RS_RT);

        dest->r.f  = tr_DADDU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DADDU(tr_u32 code, tr_instruction *dest) */





static void gen_DSUB(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSUB", RD_RS_RT);

        dest->r.f  = tr_DSUB;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DSUB(tr_u32 code, tr_instruction *dest) */





static void gen_DSUBU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSUBU", RD_RS_RT);

        dest->r.f  = tr_DSUBU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);

} /* static void gen_DSUBU(tr_u32 code, tr_instruction *dest) */





static void gen_TGE(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "TGE", RS_RT);

        dest->r.f  = tr_TGE;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_TGE(tr_u32 code, tr_instruction *dest) */





static void gen_TGEU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "TGEU", RS_RT);

        dest->r.f  = tr_TGEU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_TGEU(tr_u32 code, tr_instruction *dest) */





static void gen_TLT(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "TLT", RS_RT);

        dest->r.f  = tr_TLT;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_TLT(tr_u32 code, tr_instruction *dest) */





static void gen_TLTU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "TLTU", RS_RT);

        dest->r.f  = tr_TLTU;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_TLTU(tr_u32 code, tr_instruction *dest) */





static void gen_TEQ(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "TEQ", RS_RT);

        dest->r.f  = tr_TEQ;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_TEQ(tr_u32 code, tr_instruction *dest) */





static void gen_TNE(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "TNE", RS_RT);

        dest->r.f  = tr_TNE;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);

} /* static void gen_TNE(tr_u32 code, tr_instruction *dest) */





static void gen_DSLL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSLL", RD_RT_SA);

        dest->r.f  = tr_DSLL;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.sa = D_SA(code);

} /* static void gen_DSLL(tr_u32 code, tr_instruction *dest) */





static void gen_DSRL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSRL", RD_RT_SA);

        dest->r.f  = tr_DSRL;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.sa = D_SA(code);

} /* static void gen_DSRL(tr_u32 code, tr_instruction *dest) */





static void gen_DSRA(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSRA", RD_RT_SA);

        dest->r.f  = tr_DSRA;
        dest->r.rs = D_RS(code);
        dest->r.rt = D_RT(code);
        dest->r.sa = D_SA(code);

} /* static void gen_DSRA(tr_u32 code, tr_instruction *dest) */





static void gen_DSLL32(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSLL32", RD_RT_SA);

        dest->r.f  = tr_DSLL32;
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);
        dest->r.sa = D_SA(code);

} /* static void gen_DSLL32(tr_u32 code, tr_instruction *dest) */





static void gen_DSRL32(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSRL32", RD_RT_SA);

        dest->r.f  = tr_DSRL32;
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);
        dest->r.sa = D_SA(code);

} /* static void gen_DSRL32(tr_u32 code, tr_instruction *dest) */





static void gen_DSRA32(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DSRA32", RD_RT_SA);

        dest->r.f  = tr_DSRA32;
        dest->r.rt = D_RT(code);
        dest->r.rd = D_RD(code);
        dest->r.sa = D_SA(code);

} /* static void gen_DSRA32(tr_u32 code, tr_instruction *dest) */






/* core functions */
static void gen_regimm(tr_u32 code, tr_instruction *dest)
{
        switch(M_RT(code))
        {
            case 0x00000000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BLTZ", RS_I);
                dest->i.f  = tr_regimm_BLTZ;
                break;

            case 0x00010000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BGEZ", RS_I);
                dest->i.f  = tr_regimm_BGEZ;
                break;

            case 0x00020000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BLTZL", RS_I);
                dest->i.f  = tr_regimm_BLTZL;
                break;

            case 0x00030000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BGEZL", RS_I);
                dest->i.f  = tr_regimm_BGEZL;
                break;

            case 0x00040000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00050000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00060000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00070000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00080000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "TGEI", RS_I);
                dest->i.f  = tr_regimm_TGEI;
                break;

            case 0x00090000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "TGEIU", RS_I);
                dest->i.f  = tr_regimm_TGEIU;
                break;

            case 0x000a0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "TLTI", RS_I);
                dest->i.f  = tr_regimm_TLTI;
                break;

            case 0x000b0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "TLTIU", RS_I);
                dest->i.f  = tr_regimm_TLTIU;
                break;

            case 0x000c0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "TEQI", RS_I);
                dest->i.f  = tr_regimm_TEQI;
                break;

            case 0x000d0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x000e0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "TNEI", RS_I);
                dest->i.f  = tr_regimm_TNEI;
                break;

            case 0x000f0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00100000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BLTZAL", RS_I);
                dest->i.f  = tr_regimm_BLTZAL;
                break;

            case 0x00110000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BGEZAL", RS_I);
                dest->i.f  = tr_regimm_BGEZAL;
                break;

            case 0x00120000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BLTZALL", RS_I);
                dest->i.f  = tr_regimm_BLTZALL;
                break;

            case 0x00130000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "BGEZALL", RS_I);
                dest->i.f  = tr_regimm_BGEZALL;
                break;

            case 0x00140000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00150000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00160000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00170000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00180000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x00190000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x001a0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x001b0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x001c0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x001d0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x001e0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

            case 0x001f0000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "regimm_reserved", NOTHING);
                dest->i.f  = tr_reserved;
                break;

        } /* switch(D_RT(code)) */

        dest->i.rs = D_RS(code);
        dest->i.i  = D_I(code);

} /* static void gen_regimm(tr_u32 code, tr_instruction *dest) */





static void gen_J(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "J", T);

        dest->j.f = tr_J;
        dest->j.t = D_T(code);

} /* static void gen_J(tr_u32 code, tr_instruction *dest) */





static void gen_JAL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "JAL", T);

        dest->j.f = tr_JAL;
        dest->j.t = D_T(code);

} /* static void gen_JAL(tr_u32 code, tr_instruction *dest) */





static void gen_BEQ(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BEQ", RS_RT_I);

        dest->i.f  = tr_BEQ;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BEQ(tr_u32 code, tr_instruction *dest) */





static void gen_BNE(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BNE", RS_RT_I);

#ifdef SPEED_HACK
        if(D_I(code) == -1)
        {
            /* loop */
                dest->i.f  = tr_BNE_opt;
        }
        else
#endif
        {
                dest->i.f  = tr_BNE;
                dest->i.rs = D_RS(code);
                dest->i.rt = D_RT(code);
                dest->i.i  = D_I(code);
        }

} /* static void gen_BNE(tr_u32 code, tr_instruction *dest) */





static void gen_BLEZ(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BLEZ", RS_RT_I);

        dest->i.f  = tr_BLEZ;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BLEZ(tr_u32 code, tr_instruction *dest) */





static void gen_BGTZ(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BGTZ", RS_RT_I);

        dest->i.f  = tr_BGTZ;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BGTZ(tr_u32 code, tr_instruction *dest) */





static void gen_ADDI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "ADDI", RT_RS_I);

        dest->i.f  = tr_ADDI;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_ADDI(tr_u32 code, tr_instruction *dest) */





static void gen_ADDIU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "ADDIU", RT_RS_I);

        dest->i.f  = tr_ADDIU;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_ADDIU(tr_u32 code, tr_instruction *dest) */





static void gen_SLTI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SLTI", RT_RS_I);

        dest->i.f  = tr_SLTI;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SLTI(tr_u32 code, tr_instruction *dest) */





static void gen_SLTIU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SLTIU", RT_RS_I);

        dest->i.f  = tr_SLTIU;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SLTIU(tr_u32 code, tr_instruction *dest) */





static void gen_ANDI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "ANDI", RT_RS_I);

        dest->i.f  = tr_ANDI;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_ANDI(tr_u32 code, tr_instruction *dest) */





static void gen_ORI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "ORI", RT_RS_I);

        dest->i.f  = tr_ORI;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_ORI(tr_u32 code, tr_instruction *dest) */





static void gen_XORI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "XORI", RT_RS_I);

        dest->i.f  = tr_XORI;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_XORI(tr_u32 code, tr_instruction *dest) */





static void gen_LUI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LUI", RT_I);

        dest->i.f  = tr_LUI;
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LUI(tr_u32 code, tr_instruction *dest) */





static void gen_cop0(tr_u32 code, tr_instruction *dest)
{
        switch(M_RS(code))
        {
            case 0x00000000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "MFC0", RT_RD);
                dest->r.f  = tr_MFC0;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00200000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "DMFC0", RT_RD);
                dest->r.f  = tr_DMFC0;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00400000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "CFC0", RT_RD);
                dest->r.f  = tr_CFC0;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00800000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "MTC0", RT_RD);
                dest->r.f  = tr_MTC0;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00a00000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "DMTC0", RT_RD);
                dest->r.f  = tr_DMTC0;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00c00000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "CTC0", RT_RD);
                dest->r.f  = tr_CTC0;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x01000000:
                switch(M_RT(code))
                {
                    case 0x000000000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC0F", I);
                        dest->i.f = tr_BC0F;
                        dest->i.i = D_I(code);
                        break;

                    case 0x000010000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC0T", I);
                        dest->i.f = tr_BC0T;
                        dest->i.i = D_I(code);
                        break;

                    case 0x000020000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC0FL", I);
                        dest->i.f = tr_BC0FL;
                        dest->i.i = D_I(code);
                        break;

                    case 0x000030000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC0TL", I);
                        dest->i.f = tr_BC0TL;
                        dest->i.i = D_I(code);
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop0_reserved", NOTHING);
                        dest->d.f  = tr_reserved;
                        dest->d.c  = code;
                        break;

                } /* switch(M_RT(code)) */
                break;

            case 0x02000000:
            case 0x02100000:
            case 0x02200000:
            case 0x02300000:
            case 0x02400000:
            case 0x02500000:
            case 0x02600000:
            case 0x02700000:
            case 0x02800000:
            case 0x02900000:
            case 0x02a00000:
            case 0x02b00000:
            case 0x02c00000:
            case 0x02d00000:
            case 0x02e00000:
            case 0x02f00000:
                switch(M_F(code))
                {
                    case 0x00000001:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TLBR", NOTHING);
                        dest->r.f  = tr_cop0_TLBR;
                        break;

                    case 0x00000002:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TLBWI", NOTHING);
                        dest->r.f  = tr_cop0_TLBWI;
                        break;

                    case 0x00000006:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TLBWR", NOTHING);
                        dest->r.f  = tr_cop0_TLBWR;
                        break;

                    case 0x00000008:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TLBP", NOTHING);
                        dest->r.f  = tr_cop0_TLBP;
                        break;

                    case 0x00000010:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop0_reserved", NOTHING);
                        dest->d.f  = tr_reserved;
                        dest->d.c  = code;
                        break;

                    case 0x00000018:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ERET", NOTHING);
                        dest->r.f  = tr_cop0_ERET;
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop0_invalid", NOTHING);
                        dest->d.f  = tr_invalid;
                        dest->d.c  = code;
                        break;

                } /* switch(M_F(code)) */
                break;

            default:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "cop0_reserved", NOTHING);
                dest->d.f  = tr_reserved;
                dest->d.c  = code;
                break;

        } /* switch(M_RS(code)) */

} /* static void gen_cop0(tr_u32 code, tr_instruction *dest) */





static void gen_cop1(tr_u32 code, tr_instruction *dest)
{
        switch(M_RS(code))
        {
            case 0x00000000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "MFC1", RT_RD);
                dest->r.f  = tr_MFC1;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00200000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "DMFC1", RT_RD);
                dest->r.f  = tr_DMFC1;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00400000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "CFC1", RT_RD);
                dest->r.f  = tr_CFC1;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00800000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "MTC1", RT_RD);
                dest->r.f  = tr_MTC1;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00a00000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "DMTC1", RT_RD);
                dest->r.f  = tr_DMTC1;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x00c00000:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "CTC1", RT_RD);
                dest->r.f  = tr_CTC1;
                dest->r.rt = D_RT(code);
                dest->r.rd = D_RD(code);
                break;

            case 0x01000000:
                switch(M_RT(code))
                {
                    case 0x000000000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC1F", I);
                        dest->i.f = tr_BC1F;
                        dest->i.i = D_I(code);
                        break;

                    case 0x000010000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC1T", I);
                        dest->i.f = tr_BC1T;
                        dest->i.i = D_I(code);
                        break;

                    case 0x000020000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC1FL", I);
                        dest->i.f = tr_BC1FL;
                        dest->i.i = D_I(code);
                        break;

                    case 0x000030000:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "BC1TL", I);
                        dest->i.f = tr_BC1TL;
                        dest->i.i = D_I(code);
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_reserved", NOTHING);
                        dest->d.f  = tr_reserved;
                        dest->d.c  = code;
                        break;

                } /* switch(M_RT(code)) */
                break;

            case 0x02000000:   /* S */
                switch(M_F(code))
                {
                    case 0x00000000:   /* ADD */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ADD.s", FD_FS_FT);
                        dest->f.f  = tr_cop1_ADD_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000001:   /* SUB */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "SUB.s", FD_FS_FT);
                        dest->f.f  = tr_cop1_SUB_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000002:   /* MUL */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "MUL.s", FD_FS_FT);
                        dest->f.f  = tr_cop1_MUL_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000003:   /* DIV */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "DIV.s", FD_FS_FT);
                        dest->f.f  = tr_cop1_DIV_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000004:   /* SQRT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "SQRT.s", FD_FS);
                        dest->f.f  = tr_cop1_SQRT_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000005:   /* ABS */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ABS.s", FD_FS);
                        dest->f.f  = tr_cop1_ABS_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000006:   /* MOV */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "MOV.s", FD_FS);
                        dest->f.f  = tr_cop1_MOV_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000007:   /* NEG */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "NEG.s", FD_FS);
                        dest->f.f  = tr_cop1_NEG_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000008:   /* ROUND.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ROUND.L.s", FD_FS);
                        dest->f.f  = tr_cop1_ROUNDl_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000009:   /* TRUNC.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TRUNC.L.s", FD_FS);
                        dest->f.f  = tr_cop1_TRUNCl_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000a:   /* CEIL.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CEIL.L.s", FD_FS);
                        dest->f.f  = tr_cop1_CEILl_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000b:   /* FLOOR.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "FLOOR.L.s", FD_FS);
                        dest->f.f  = tr_cop1_FLOORl_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000c:   /* ROUND.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ROUND.W.s", FD_FS);
                        dest->f.f  = tr_cop1_ROUNDw_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000d:   /* TRUNC.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TRUNC.W.s", FD_FS);
                        dest->f.f  = tr_cop1_TRUNCw_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000e:   /* CEIL.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CEIL.W.s", FD_FS);
                        dest->f.f  = tr_cop1_CEILw_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000f:   /* FLOOR.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "FLOOR.W.s", FD_FS);
                        dest->f.f  = tr_cop1_FLOORw_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000021:   /* CVT.D */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.D.s", FD_FS);
                        dest->f.f  = tr_cop1_CVTd_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000024:   /* CVT.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.W.s", FD_FS);
                        dest->f.f  = tr_cop1_CVTw_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000025:   /* CVT.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.L.s", FD_FS);
                        dest->f.f  = tr_cop1_CVTl_s;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000030:   /* C.F */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.F.s", FS_FT);
                        dest->f.f  = tr_cop1_CF_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000031:   /* C.UN */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.UN.s", FS_FT);
                        dest->f.f  = tr_cop1_CUN_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000032:   /* C.EQ */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.EQ.s", FS_FT);
                        dest->f.f  = tr_cop1_CEQ_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000033:   /* C.UEQ */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.UEQ.s", FS_FT);
                        dest->f.f  = tr_cop1_CUEQ_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000034:   /* C.OLT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.OLT.s", FS_FT);
                        dest->f.f  = tr_cop1_COLT_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000035:   /* C.ULT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.ULT.s", FS_FT);
                        dest->f.f  = tr_cop1_CULT_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000036:   /* C.OLE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.OLE.s", FS_FT);
                        dest->f.f  = tr_cop1_COLE_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000037:   /* C.ULE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.ULE.s", FS_FT);
                        dest->f.f  = tr_cop1_CULE_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000038:   /* C.SF */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.SF.s", FS_FT);
                        dest->f.f  = tr_cop1_CSF_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000039:   /* C.NGLE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGLE.s", FS_FT);
                        dest->f.f  = tr_cop1_CNGLE_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003a:   /* C.SEQ */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.SEQ.s", FS_FT);
                        dest->f.f  = tr_cop1_CSEQ_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003b:   /* C.NGL */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGL.s", FS_FT);
                        dest->f.f  = tr_cop1_CNGL_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003c:   /* C.LT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.LT.s", FS_FT);
                        dest->f.f  = tr_cop1_CLT_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003d:   /* C.NGE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGE.s", FS_FT);
                        dest->f.f  = tr_cop1_CNGE_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003e:   /* C.LE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.LE.s", FS_FT);
                        dest->f.f  = tr_cop1_CLE_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003f:   /* C.NGT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGT.s", FS_FT);
                        dest->f.f  = tr_cop1_CNGT_s;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000020:   /* CVT.S */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_invalid", NOTHING);
                        dest->d.f  = tr_invalid;
                        dest->d.c  = code;
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_unimplemented", NOTHING);
                        dest->d.f  = tr_unimplemented;
                        dest->d.c  = code;
                        break;

                } /* switch(M_F(code)) */
                break;

            case 0x02200000:   /* D */
                switch(M_F(code))
                {
                    case 0x00000000:   /* ADD */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ADD.d", FD_FS_FT);
                        dest->f.f  = tr_cop1_ADD_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000001:   /* SUB */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "SUB.d", FD_FS_FT);
                        dest->f.f  = tr_cop1_SUB_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000002:   /* MUL */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "MUL.d", FD_FS_FT);
                        dest->f.f  = tr_cop1_MUL_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000003:   /* DIV */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "DIV.d", FD_FS_FT);
                        dest->f.f  = tr_cop1_DIV_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000004:   /* SQRT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "SQRT.d", FD_FS);
                        dest->f.f  = tr_cop1_SQRT_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000005:   /* ABS */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ABS.d", FD_FS);
                        dest->f.f  = tr_cop1_ABS_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000006:   /* MOV */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "MOV.d", FD_FS);
                        dest->f.f  = tr_cop1_MOV_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000007:   /* NEG */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "NEG.d", FD_FS);
                        dest->f.f  = tr_cop1_NEG_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000008:   /* ROUND.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ROUND.L.d", FD_FS);
                        dest->f.f  = tr_cop1_ROUNDl_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000009:   /* TRUNC.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TRUNC.L.d", FD_FS);
                        dest->f.f  = tr_cop1_TRUNCl_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000a:   /* CEIL.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CEIL.L.d", FD_FS);
                        dest->f.f  = tr_cop1_CEILl_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000b:   /* FLOOR.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "FLOOR.L.d", FD_FS);
                        dest->f.f  = tr_cop1_FLOORl_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000c:   /* ROUND.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "ROUND.W.d", FD_FS);
                        dest->f.f  = tr_cop1_ROUNDw_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000d:   /* TRUNC.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "TRUNC.W.d", FD_FS);
                        dest->f.f  = tr_cop1_TRUNCw_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000e:   /* CEIL.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CEIL.W.d", FD_FS);
                        dest->f.f  = tr_cop1_CEILw_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x0000000f:   /* FLOOR.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "FLOOR.W.d", FD_FS);
                        dest->f.f  = tr_cop1_FLOORw_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000020:   /* CVT.S */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.S.d", FD_FS);
                        dest->f.f  = tr_cop1_CVTs_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000024:   /* CVT.W */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.W.d", FD_FS);
                        dest->f.f  = tr_cop1_CVTw_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000025:   /* CVT.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.L.d", FD_FS);
                        dest->f.f  = tr_cop1_CVTl_d;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000030:   /* C.F */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.F.d", FS_FT);
                        dest->f.f  = tr_cop1_CF_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000031:   /* C.UN */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.UN.d", FS_FT);
                        dest->f.f  = tr_cop1_CUN_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000032:   /* C.EQ */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.EQ.d", FS_FT);
                        dest->f.f  = tr_cop1_CEQ_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000033:   /* C.UEQ */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.UEQ.d", FS_FT);
                        dest->f.f  = tr_cop1_CUEQ_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000034:   /* C.OLT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.OLT.d", FS_FT);
                        dest->f.f  = tr_cop1_COLT_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000035:   /* C.ULT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.ULT.d", FS_FT);
                        dest->f.f  = tr_cop1_CULT_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000036:   /* C.OLE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.OLE.d", FS_FT);
                        dest->f.f  = tr_cop1_COLE_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000037:   /* C.ULE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.ULE.d", FS_FT);
                        dest->f.f  = tr_cop1_CULE_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000038:   /* C.SF */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.SF.d", FS_FT);
                        dest->f.f  = tr_cop1_CSF_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000039:   /* C.NGLE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGLE.d", FS_FT);
                        dest->f.f  = tr_cop1_CNGLE_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003a:   /* C.SEQ */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.SEQ.d", FS_FT);
                        dest->f.f  = tr_cop1_CSEQ_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003b:   /* C.NGL */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGL.d", FS_FT);
                        dest->f.f  = tr_cop1_CNGL_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003c:   /* C.LT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.LT.d", FS_FT);
                        dest->f.f  = tr_cop1_CLT_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003d:   /* C.NGE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGE.d", FS_FT);
                        dest->f.f  = tr_cop1_CNGE_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003e:   /* C.LE */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.LE.d", FS_FT);
                        dest->f.f  = tr_cop1_CLE_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x0000003f:   /* C.NGT */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "C.NGT.d", FS_FT);
                        dest->f.f  = tr_cop1_CNGT_d;
                        dest->f.fs = D_FS(code);
                        dest->f.ft = D_FT(code);
                        break;

                    case 0x00000021:   /* CVT.D */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_invalid", NOTHING);
                        dest->d.f  = tr_invalid;
                        dest->d.c  = code;
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_unimplemented", NOTHING);
                        dest->d.f  = tr_unimplemented;
                        dest->d.c  = code;
                        break;

                } /* switch(M_F(code)) */
                break;

            case 0x02800000:   /* W */
                switch(M_F(code))
                {
                    case 0x00000020:   /* CVT.S */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.S.w", FD_FS);
                        dest->f.f  = tr_cop1_CVTs_w;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000021:   /* CVT.D */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.D.w", FD_FS);
                        dest->f.f  = tr_cop1_CVTd_w;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000006:   /* MOV */
                    case 0x00000008:   /* TRUNC.L */
                    case 0x00000009:   /* ROUND.L */
                    case 0x0000000a:   /* CEIL.L */
                    case 0x0000000b:   /* FLOOR.L */
                    case 0x0000000c:   /* TRUNC.W */
                    case 0x0000000d:   /* ROUND.W */
                    case 0x0000000e:   /* CEIL.W */
                    case 0x0000000f:   /* FLOOR.W */
                    case 0x00000024:   /* CVT.W */
                    case 0x00000025:   /* CVT.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_invalid", NOTHING);
                        dest->d.f  = tr_invalid;
                        dest->d.c  = code;
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_unimplemented", NOTHING);
                        dest->d.f  = tr_unimplemented;
                        dest->d.c  = code;
                        break;

                } /* switch(M_F(code)) */
                break;

            case 0x02a00000:   /* L */
                switch(M_F(code))
                {
                    case 0x00000020:   /* CVT.S */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.S.l", FD_FS);
                        dest->f.f  = tr_cop1_CVTs_l;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000021:   /* CVT.D */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "CVT.D.l", FD_FS);
                        dest->f.f  = tr_cop1_CVTd_l;
                        dest->f.fd = D_FD(code);
                        dest->f.fs = D_FS(code);
                        break;

                    case 0x00000006:   /* MOV */
                    case 0x00000008:   /* TRUNC.L */
                    case 0x00000009:   /* ROUND.L */
                    case 0x0000000a:   /* CEIL.L */
                    case 0x0000000b:   /* FLOOR.L */
                    case 0x0000000c:   /* TRUNC.W */
                    case 0x0000000d:   /* ROUND.W */
                    case 0x0000000e:   /* CEIL.W */
                    case 0x0000000f:   /* FLOOR.W */
                    case 0x00000024:   /* CVT.W */
                    case 0x00000025:   /* CVT.L */
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_invalid", NOTHING);
                        dest->d.f  = tr_invalid;
                        dest->d.c  = code;
                        break;

                    default:
                        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_unimplemented", NOTHING);
                        dest->d.f  = tr_unimplemented;
                        dest->d.c  = code;
                        break;

                } /* switch(M_F(code)) */
                break;

            default:
                TR_PRINT_DEBUGGING_GEN_INFO(code, "cop1_unimplemented", NOTHING);
                dest->d.f  = tr_unimplemented;
                dest->d.c  = code;
                break;

        } /* switch(M_RS(code)) */

} /* static void gen_cop1(tr_u32 code, tr_instruction *dest) */





static void gen_cop2(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "cop2", NOTHING);

        dest->d.f = tr_invalid;
        dest->d.c = code;

} /* static void gen_cop2(tr_u32 code, tr_instruction *dest) */





static void gen_BEQL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BEQL", RS_RT_I);

        dest->i.f  = tr_BEQL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BEQL(tr_u32 code, tr_instruction *dest) */





static void gen_BNEL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BNEL", RS_RT_I);

        dest->i.f  = tr_BNEL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BNEL(tr_u32 code, tr_instruction *dest) */





static void gen_BLEZL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BLEZL", RS_RT_I);

        dest->i.f  = tr_BLEZL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BLEZL(tr_u32 code, tr_instruction *dest) */





static void gen_BGTZL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "BGTZL", RS_RT_I);

        dest->i.f  = tr_BGTZL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_BGTZL(tr_u32 code, tr_instruction *dest) */





static void gen_DADDI(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DADDI", RT_RS_I);

        dest->i.f  = tr_DADDI;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_DADDI(tr_u32 code, tr_instruction *dest) */





static void gen_DADDIU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "DADDIU", RT_RS_I);

        dest->i.f  = tr_DADDIU;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_DADDIU(tr_u32 code, tr_instruction *dest) */





static void gen_LDL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LDL", RT_O_B);

        dest->i.f  = tr_LDL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LDL(tr_u32 code, tr_instruction *dest) */





static void gen_LDR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LDR", RT_O_B);

        dest->i.f  = tr_LDR;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LDR(tr_u32 code, tr_instruction *dest) */





static void gen_LB(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LB", RT_O_B);

        dest->i.f  = tr_LB;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LB(tr_u32 code, tr_instruction *dest) */





static void gen_LH(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LH", RT_O_B);

        dest->i.f  = tr_LH;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LH(tr_u32 code, tr_instruction *dest) */





static void gen_LWL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LWL", RT_O_B);

        dest->i.f  = tr_LWL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LWL(tr_u32 code, tr_instruction *dest) */





static void gen_LW(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LW", RT_O_B);

        dest->i.f  = tr_LW;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LW(tr_u32 code, tr_instruction *dest) */





static void gen_LBU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LBU", RT_O_B);

        dest->i.f  = tr_LBU;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LBU(tr_u32 code, tr_instruction *dest) */





static void gen_LHU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LHU", RT_O_B);

        dest->i.f  = tr_LHU;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LHU(tr_u32 code, tr_instruction *dest) */





static void gen_LWR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LWR", RT_O_B);

        dest->i.f  = tr_LWR;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LWR(tr_u32 code, tr_instruction *dest) */





static void gen_LWU(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LWU", RT_O_B);

        dest->i.f  = tr_LWU;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LWU(tr_u32 code, tr_instruction *dest) */





static void gen_SB(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SB", RT_O_B);

        dest->i.f  = tr_SB;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SB(tr_u32 code, tr_instruction *dest) */





static void gen_SH(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SH", RT_O_B);

        dest->i.f  = tr_SH;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SH(tr_u32 code, tr_instruction *dest) */





static void gen_SWL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SWL", RT_O_B);

        dest->i.f  = tr_SWL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SWL(tr_u32 code, tr_instruction *dest) */





static void gen_SW(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SW", RT_O_B);

        dest->i.f  = tr_SW;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SW(tr_u32 code, tr_instruction *dest) */





static void gen_SDL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SDL", RT_O_B);

        dest->i.f  = tr_SDL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SDL(tr_u32 code, tr_instruction *dest) */





static void gen_SDR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SDR", RT_O_B);

        dest->i.f  = tr_SDR;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SDR(tr_u32 code, tr_instruction *dest) */





static void gen_SWR(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SWR", RT_O_B);

        dest->i.f  = tr_SWR;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SWR(tr_u32 code, tr_instruction *dest) */





static void gen_CACHE(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "CACHE", RT_O_B);

        dest->i.f  = tr_CACHE;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_CACHE(tr_u32 code, tr_instruction *dest) */





static void gen_LL(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LL", RT_O_B);

        dest->i.f  = tr_LL;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LL(tr_u32 code, tr_instruction *dest) */





static void gen_LWC1(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LWC1", RT_O_B);

        dest->i.f  = tr_LWC1;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LWC1(tr_u32 code, tr_instruction *dest) */





static void gen_LWC2(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LWC2", RT_O_B);

        dest->i.f  = tr_LWC2;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LWC2(tr_u32 code, tr_instruction *dest) */





static void gen_LLD(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LLD", RT_O_B);

        dest->i.f  = tr_LLD;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LLD(tr_u32 code, tr_instruction *dest) */





static void gen_LDC1(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LDC1", RT_O_B);

        dest->i.f  = tr_LDC1;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LDC1(tr_u32 code, tr_instruction *dest) */





static void gen_LDC2(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LDC2", RT_O_B);

        dest->i.f  = tr_LDC2;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LDC2(tr_u32 code, tr_instruction *dest) */





static void gen_LD(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "LD", RT_O_B);

        dest->i.f  = tr_LD;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_LD(tr_u32 code, tr_instruction *dest) */





static void gen_SC(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SC", RT_O_B);

        dest->i.f  = tr_SC;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SC(tr_u32 code, tr_instruction *dest) */





static void gen_SWC1(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SWC1", RT_O_B);

        dest->i.f  = tr_SWC1;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SWC1(tr_u32 code, tr_instruction *dest) */





static void gen_SWC2(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SWC2", RT_O_B);

        dest->i.f  = tr_SWC2;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SWC2(tr_u32 code, tr_instruction *dest) */





static void gen_SCD(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SCD", RT_O_B);

        dest->i.f  = tr_SCD;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SCD(tr_u32 code, tr_instruction *dest) */





static void gen_SDC1(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SDC1", RT_O_B);

        dest->i.f  = tr_SDC1;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SDC1(tr_u32 code, tr_instruction *dest) */





static void gen_SDC2(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SDC2", RT_O_B);

        dest->i.f  = tr_SDC2;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SDC2(tr_u32 code, tr_instruction *dest) */





static void gen_SD(tr_u32 code, tr_instruction *dest)
{
        TR_PRINT_DEBUGGING_GEN_INFO(code, "SD", RT_O_B);

        dest->i.f  = tr_SD;
        dest->i.rs = D_RS(code);
        dest->i.rt = D_RT(code);
        dest->i.i  = D_I(code);

} /* static void gen_SD(tr_u32 code, tr_instruction *dest) */







