/**
 * TrueReality - N64/compile_static_mnemonic.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#ifndef __TR_COMPILE_STATIC_MNEMONIC_H
#define __TR_COMPILE_STATIC_MNEMONIC_H





#include "cpu.h"
#include "compile_static.h"
#include <parser_extern.h>
#include <memory_extern.h>





/**********\
*          *
* mnemonic *
*          *
\**********/

/* masks */
#define M_OP(x) ((x) & 0xfc000000)
#define M_RS(x) ((x) & 0x03e00000)
#define M_RT(x) ((x) & 0x001f0000)
#define M_RD(x) ((x) & 0x0000f800)
#define M_SA(x) ((x) & 0x000007c0)
#define M_F(x)  ((x) & 0x0000003f)
#define M_I(x)  ((x) & 0x0000ffff)
#define M_T(x)  ((x) & 0x03ffffff)

/* disassemble */
#define D_RS(x)  ( ((x) >> 21) & 0x0000001f )
#define D_RT(x)  ( ((x) >> 16) & 0x0000001f )
#define D_RD(x)  ( ((x) >> 11) & 0x0000001f )
#define D_SA(x)  ( ((x) >>  6) & 0x0000001f )
#define D_I(x)   ( (tr_s16) ((x) & 0x0000ffff) )
#define D_T(x)   (          ((x) & 0x03ffffff) )   /* attension: not multiplied by 4 (`<< 2')! */

#define D_FS(x)  D_RD(x)
#define D_FT(x)  D_RT(x)
#define D_FD(x)  D_SA(x)
#define D_FMT(x) D_RS(x)

/* current values */
#define C_RS (tr_i->r.rs)
#define C_RT (tr_i->r.rt)
#define C_RD (tr_i->r.rd)
#define C_SA (tr_i->r.sa)
#define C_I  (tr_i->i.i)
#define C_T  (tr_i->j.t)

#define C_FS  C_RD
#define C_FT  C_RT
#define C_FD  C_SA
#define C_FMT C_RS

/* the next stuff looks complicated, i know ;-) */

#define C_PC() (                                                              \
        ( !(reg.delay_jump & 0x04000000) )   /* rdram */                      \
        ?                                                                     \
                (                                                             \
                        (((tr_u32)(tr_i - mem.rd_ram_instr)) << 2)            \
                        |                                                     \
                        (reg.delay_jump & ~TR_RDMEM_MASK)                     \
                )                                                             \
        :                                                                     \
        ( reg.delay_jump & 0x00001000 )      /* imem */                       \
        ?                                                                     \
                (                                                             \
                        (((tr_u32)(tr_i - mem.sp_imem_instr)) << 2)           \
                        |                                                     \
                        (reg.delay_jump & 0xfffff000)                         \
                )                                                             \
        :                                    /* dmem */                       \
                (                                                             \
                        (((tr_u32)(tr_i - mem.sp_dmem_instr)) << 2)           \
                        |                                                     \
                        (reg.delay_jump & 0xfffff000)                         \
                )                                                             \
)

#define C_JUMP() (                                                            \
        ( !(reg.delay_jump & 0x04000000) )   /* rdram */                      \
        ?                                                                     \
        ( mem.rd_ram_instr + ((reg.delay_jump & TR_RDMEM_MASK) >> 2) )        \
        :                                                                     \
        ( reg.delay_jump & 0x00001000 )      /* imem */                       \
        ?                                                                     \
        ( mem.sp_imem_instr + ((reg.delay_jump & 0xfff) >> 2) )               \
        :                                    /* dmem */                       \
        ( mem.sp_dmem_instr + ((reg.delay_jump & 0xfff) >> 2) )               \
)

#define FS                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][C_FS]                                        \
        :                                                               \
                ((reg.cpr[1][C_FS+1] << 32) | (reg.cpr[1][C_FS] & 0xffffffff)) \
)

#define FT                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][C_FT]                                        \
        :                                                               \
                ((reg.cpr[1][C_FT+1] << 32) | (reg.cpr[1][C_FT] & 0xffffffff)) \
)


#define STORE_FPR(value)                                                \
{                                                                       \
        if(reg.cpr[0][12] & 0x04000000)                                 \
        {                                                               \
                reg.cpr[1][C_FD] = (value);                             \
        }                                                               \
        else                                                            \
        {                                                               \
                reg.cpr[1][C_FD+0] = (value);                           \
                reg.cpr[1][C_FD+0] = (tr_s64)(tr_s32)reg.cpr[1][C_FD+0]; \
                reg.cpr[1][C_FD+1] = (value) >> 32;                     \
                reg.cpr[1][C_FD+1] = (tr_s64)(tr_s32)reg.cpr[1][C_FD+1]; \
        }                                                               \
}


/* assemble */
#define A_RS(x) (((x) & 0x0000001f) << 21)
#define A_RT(x) (((x) & 0x0000001f) << 16)
#define A_RD(x) (((x) & 0x0000001f) << 11)
#define A_SA(x) (((x) & 0x0000001f) <<  6)
#define A_I(x)  (((x) & 0x0000ffff) <<  0)
#define A_T(x)  (((x) & 0x03ffffff) <<  0)   /* attension: not divided by 4 (`>> 2')!*/

#define A_FS(x)  A_RD(x)
#define A_FT(x)  A_RT(x)
#define A_FD(x)  A_SA(x)
#define A_FMT(x) A_RS(x)



/* macros for generating code (these need the assemble macros above!) */

#define TR_A_SLL(rd,rt,sa)        ( 0x00000000 | A_RT(rt) | A_RD(rd) | A_SA(sa) )
#define TR_A_CLR(rd)              ( TR_A_SLL(rd, 0, 0) )
#define TR_A_NOP()                ( TR_A_CLR(0) )
#define TR_A_SRL(rd,rt,sa)        ( 0x00000002 | A_RT(rt) | A_RD(rd) | A_SA(sa) )
#define TR_A_SRA(rd,rt,sa)        ( 0x00000003 | A_RT(rt) | A_RD(rd) | A_SA(sa) )

#define TR_A_SLLV(rd,rt,rs)       ( 0x00000004 | A_RT(rt) | A_RD(rd) | A_RS(rs) )
#define TR_A_SRLV(rd,rt,rs)       ( 0x00000006 | A_RT(rt) | A_RD(rd) | A_RS(rs) )
#define TR_A_SRAV(rd,rt,rs)       ( 0x00000007 | A_RT(rt) | A_RD(rd) | A_RS(rs) )

#define TR_A_JR(rs)               ( 0x00000008 | A_RS(rs) )
#define TR_A_JALR(rd,rs)          ( 0x00000009 | A_RD(rd) | A_RS(rs) )

#define TR_A_SYSCALL()            ( 0x0000000c )
#define TR_A_BREAK()              ( 0x0000000d )
#define TR_A_SYNC()               ( 0x0000000f ) 

#define TR_A_MFHI(rd)             ( 0x00000010 | A_RD(rd) )
#define TR_A_MTHI(rs)             ( 0x00000011 | A_RS(rs) )
#define TR_A_MFLO(rd)             ( 0x00000012 | A_RD(rd) )
#define TR_A_MTLO(rs)             ( 0x00000013 | A_RS(rs) )

#define TR_A_DSLLV(rd,rt,rs)      ( 0x00000014 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_DSRLV(rd,rt,rs)      ( 0x00000016 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_DSRAV(rd,rt,rs)      ( 0x00000017 | A_RS(rs) | A_RT(rt) | A_RD(rd) )

#define TR_A_MULT(rs,rt)          ( 0x00000018 | A_RS(rs) | A_RT(rt) )
#define TR_A_MULTU(rs,rt)         ( 0x00000019 | A_RS(rs) | A_RT(rt) )
#define TR_A_DIV(rs,rt)           ( 0x0000001a | A_RS(rs) | A_RT(rt) )
#define TR_A_DIVU(rs,rt)          ( 0x0000001b | A_RS(rs) | A_RT(rt) )
#define TR_A_DMULT(rs,rt)         ( 0x0000001c | A_RS(rs) | A_RT(rt) )
#define TR_A_DMULTU(rs,rt)        ( 0x0000001d | A_RS(rs) | A_RT(rt) )
#define TR_A_DDIV(rs,rt)          ( 0x0000001e | A_RS(rs) | A_RT(rt) )
#define TR_A_DDIVU(rs,rt)         ( 0x0000001f | A_RS(rs) | A_RT(rt) )

#define TR_A_ADD(rd,rs,rt)        ( 0x00000020 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_ADDU(rd,rs,rt)       ( 0x00000021 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_SUB(rd,rs,rt)        ( 0x00000022 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_SUBU(rd,rs,rt)       ( 0x00000023 | A_RS(rs) | A_RT(rt) | A_RD(rd) )

#define TR_A_AND(rd,rs,rt)        ( 0x00000024 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_OR(rd,rs,rt)         ( 0x00000025 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_XOR(rd,rs,rt)        ( 0x00000026 | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_NOR(rd,rs,rt)        ( 0x00000027 | A_RS(rs) | A_RT(rt) | A_RD(rd) ) 

#define TR_A_SLT(rd,rs,rt)        ( 0x0000002a | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_SLTU(rd,rs,rt)       ( 0x0000002b | A_RS(rs) | A_RT(rt) | A_RD(rd) )

#define TR_A_DADD(rd,rs,rt)       ( 0x0000002c | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_DADDU(rd,rs,rt)      ( 0x0000002d | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_DSUB(rd,rs,rt)       ( 0x0000002e | A_RS(rs) | A_RT(rt) | A_RD(rd) )
#define TR_A_DSUBU(rd,rs,rt)      ( 0x0000002f | A_RS(rs) | A_RT(rt) | A_RD(rd) )

#define TR_A_TGE(rs,rt)           ( 0x00000030 | A_RS(rs) | A_RT(rt) )
#define TR_A_TGEU(rs,rt)          ( 0x00000031 | A_RS(rs) | A_RT(rt) )
#define TR_A_TLT(rs,rt)           ( 0x00000032 | A_RS(rs) | A_RT(rt) )
#define TR_A_TLTU(rs,rt)          ( 0x00000033 | A_RS(rs) | A_RT(rt) )
#define TR_A_TEQ(rs,rt)           ( 0x00000034 | A_RS(rs) | A_RT(rt) )
#define TR_A_TNE(rs,rt)           ( 0x00000036 | A_RS(rs) | A_RT(rt) )

#define TR_A_DSLL(rd,rt,sa)       ( 0x00000038 | A_RD(rd) | A_RT(rt) | A_SA(sa) )
#define TR_A_DSRL(rd,rt,sa)       ( 0x0000003a | A_RD(rd) | A_RT(rt) | A_SA(sa) )
#define TR_A_DSRA(rd,rt,sa)       ( 0x0000003b | A_RD(rd) | A_RT(rt) | A_SA(sa) )
#define TR_A_DSLL32(rd,rt,sa)     ( 0x0000003c | A_RD(rd) | A_RT(rt) | A_SA(sa) )
#define TR_A_DSRL32(rd,rt,sa)     ( 0x0000003e | A_RD(rd) | A_RT(rt) | A_SA(sa) )
#define TR_A_DSRA32(rd,rt,sa)     ( 0x0000003f | A_RD(rd) | A_RT(rt) | A_SA(sa) )

#define TR_A_BLTZ(rs,i)           ( 0x04000000 | A_RS(rs) | A_I(i) )
#define TR_A_BGEZ(rs,i)           ( 0x04010000 | A_RS(rs) | A_I(i) )
#define TR_A_BLTZL(rs,i)          ( 0x04020000 | A_RS(rs) | A_I(i) )
#define TR_A_BGEZL(rs,i)          ( 0x04030000 | A_RS(rs) | A_I(i) )
#define TR_A_TGEI(rs,i)           ( 0x04080000 | A_RS(rs) | A_I(i) )
#define TR_A_TGEIU(rs,i)          ( 0x04090000 | A_RS(rs) | A_I(i) )
#define TR_A_TLTI(rs,i)           ( 0x040a0000 | A_RS(rs) | A_I(i) )
#define TR_A_TLTIU(rs,i)          ( 0x040b0000 | A_RS(rs) | A_I(i) )
#define TR_A_TEQI(rs,i)           ( 0x040c0000 | A_RS(rs) | A_I(i) )
#define TR_A_TNEI(rs,i)           ( 0x040e0000 | A_RS(rs) | A_I(i) )
#define TR_A_BLTZAL(rs,i)         ( 0x04100000 | A_RS(rs) | A_I(i) )
#define TR_A_BGEZAL(rs,i)         ( 0x04110000 | A_RS(rs) | A_I(i) )
#define TR_A_BLTZALL(rs,i)        ( 0x04120000 | A_RS(rs) | A_I(i) )
#define TR_A_BGEZALL(rs,i)        ( 0x04130000 | A_RS(rs) | A_I(i) )

#define TR_A_J(t)                 ( 0x08000000 | A_T(t) )
#define TR_A_JAL(t)               ( 0x0c000000 | A_T(t) )
#define TR_A_BEQ(rs,rt,i)         ( 0x10000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_BNE(rs,rt,i)         ( 0x14000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_BLEZ(rs,rt,i)        ( 0x18000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_BGTZ(rs,rt,i)        ( 0x1c000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_ADDI(rt,rs,i)        ( 0x20000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_ADDIU(rt,rs,i)       ( 0x24000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_SLTI(rt,rs,i)        ( 0x28000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_SLTIU(rt,rs,i)       ( 0x2c000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_ANDI(rt,rs,i)        ( 0x30000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_ORI(rt,rs,i)         ( 0x34000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_XORI(rt,rs,i)        ( 0x38000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_LUI(rt,i)            ( 0x3c000000 | A_RT(rt) | A_I(i) )
#define TR_A_BEQL(rs,rt,i)        ( 0x50000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_BNEL(rs,rt,i)        ( 0x54000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_BLEZL(rs,rt,i)       ( 0x58000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_BGTZL(rs,rt,i)       ( 0x5c000000 | A_RS(rs) | A_RT(rt) | A_I(i) )
#define TR_A_DADDI(rt,rs,i)       ( 0x60000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_DADDIU(rt,rs,i)      ( 0x64000000 | A_RT(rt) | A_RS(rs) | A_I(i) )
#define TR_A_LDL(rt,o,b)          ( 0x68000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LDR(rt,o,b)          ( 0x6c000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LB(rt,o,b)           ( 0x80000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LH(rt,o,b)           ( 0x84000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LWL(rt,o,b)          ( 0x88000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LW(rt,o,b)           ( 0x8c000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LBU(rt,o,b)          ( 0x90000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LHU(rt,o,b)          ( 0x94000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LWR(rt,o,b)          ( 0x98000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LWU(rt,o,b)          ( 0x9c000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SB(rt,o,b)           ( 0xa0000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SH(rt,o,b)           ( 0xa4000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SWL(rt,o,b)          ( 0xa8000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SW(rt,o,b)           ( 0xac000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SDL(rt,o,b)          ( 0xb0000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SDR(rt,o,b)          ( 0xb4000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SWR(rt,o,b)          ( 0xb8000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_CACHE(rt,o,b)        ( 0xbc000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LL(rt,o,b)           ( 0xc0000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LWC1(rt,o,b)         ( 0xc4000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LWC2(rt,o,b)         ( 0xc8000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LLD(rt,o,b)          ( 0xd0000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LDC1(rt,o,b)         ( 0xd4000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LDC2(rt,o,b)         ( 0xd8000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_LD(rt,o,b)           ( 0xdc000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SC(rt,o,b)           ( 0xe0000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SWC1(rt,o,b)         ( 0xe4000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SWC2(rt,o,b)         ( 0xe8000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SCD(rt,o,b)          ( 0xf0000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SDC1(rt,o,b)         ( 0xf4000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SDC2(rt,o,b)         ( 0xf8000000 | A_RS(b) | A_RT(rt) | A_I(o) )
#define TR_A_SD(rt,o,b)           ( 0xfc000000 | A_RS(b) | A_RT(rt) | A_I(o) )

/* cop0 (tlb/eret) */
#define TR_A_MFC0(rt,rd)          ( 0x40000000 | A_RT(rt) | A_RD(rd) )
#define TR_A_DMFC0(rt,rd)         ( 0x40200000 | A_RT(rt) | A_RD(rd) )
#define TR_A_CFC0(rt,rd)          ( 0x40400000 | A_RT(rt) | A_RD(rd) )
#define TR_A_MTC0(rt,rd)          ( 0x40800000 | A_RT(rt) | A_RD(rd) )
#define TR_A_DMTC0(rt,rd)         ( 0x40a00000 | A_RT(rt) | A_RD(rd) )
#define TR_A_CTC0(rt,rd)          ( 0x40c00000 | A_RT(rt) | A_RD(rd) )
#define TR_A_BC0F(o)              ( 0x41000000 | A_I(o) )
#define TR_A_BC0T(o)              ( 0x41010000 | A_I(o) )
#define TR_A_BC0FL(o)             ( 0x41020000 | A_I(o) )
#define TR_A_BC0TL(o)             ( 0x41030000 | A_I(o) )
#define TR_A_cop0_TLBR()          ( 0x42000001 )
#define TR_A_cop0_TLBWI()         ( 0x42000002 )
#define TR_A_cop0_TLBWR()         ( 0x42000006 )
#define TR_A_cop0_TLBP()          ( 0x42000008 )
#define TR_A_cop0_ERET()          ( 0x42000018 )

/* cop1 (fpu) */
#define TR_A_MFC1(rt,rd)          ( 0x44000000 | A_RT(rt) | A_RD(rd) )
#define TR_A_DMFC1(rt,rd)         ( 0x44200000 | A_RT(rt) | A_RD(rd) )
#define TR_A_CFC1(rt,rd)          ( 0x44400000 | A_RT(rt) | A_RD(rd) )
#define TR_A_MTC1(rt,rd)          ( 0x44800000 | A_RT(rt) | A_RD(rd) )
#define TR_A_DMTC1(rt,rd)         ( 0x44a00000 | A_RT(rt) | A_RD(rd) )
#define TR_A_CTC1(rt,rd)          ( 0x44c00000 | A_RT(rt) | A_RD(rd) )
#define TR_A_BC1F(o)              ( 0x45000000 | A_I(o) )
#define TR_A_BC1T(o)              ( 0x45010000 | A_I(o) )
#define TR_A_BC1FL(o)             ( 0x45020000 | A_I(o) )
#define TR_A_BC1TL(o)             ( 0x45030000 | A_I(o) )
#define TR_A_cop1_ADD_s(fd,fs,ft) ( 0x46000000 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_ADD_d(fd,fs,ft) ( 0x46200000 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_SUB_s(fd,fs,ft) ( 0x46000001 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_SUB_d(fd,fs,ft) ( 0x46200001 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_MUL_s(fd,fs,ft) ( 0x46000002 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_MUL_d(fd,fs,ft) ( 0x46200002 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_DIV_s(fd,fs,ft) ( 0x46000003 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_DIV_d(fd,fs,ft) ( 0x46200003 | A_FD(fd) | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_SQRT_s(fd,fs)   ( 0x46000004 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_SQRT_d(fd,fs)   ( 0x46200004 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_ABS_s(fd,fs)    ( 0x46000005 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_ABS_d(fd,fs)    ( 0x46200005 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_MOV_s(fd,fs)    ( 0x46000006 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_MOV_d(fd,fs)    ( 0x46200006 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_NEG_s(fd,fs)    ( 0x46000007 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_NEG_d(fd,fs)    ( 0x46200007 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_ROUNDl_s(fd,fs) ( 0x46000008 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_ROUNDl_d(fd,fs) ( 0x46200008 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_TRUNCl_s(fd,fs) ( 0x46000009 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_TRUNCl_d(fd,fs) ( 0x46200009 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CEILl_s(fd,fs)  ( 0x4600000a | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CEILl_d(fd,fs)  ( 0x4620000a | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_FLOORl_s(fd,fs) ( 0x4600000b | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_FLOORl_d(fd,fs) ( 0x4620000b | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_ROUNDw_s(fd,fs) ( 0x4600000c | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_ROUNDw_d(fd,fs) ( 0x4620000c | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_TRUNCw_s(fd,fs) ( 0x4600000d | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_TRUNCw_d(fd,fs) ( 0x4620000d | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CEILw_s(fd,fs)  ( 0x4600000e | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CEILw_d(fd,fs)  ( 0x4620000e | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_FLOORw_s(fd,fs) ( 0x4600000f | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_FLOORw_d(fd,fs) ( 0x4620000f | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTs_d(fd,fs)   ( 0x46200020 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTs_w(fd,fs)   ( 0x46800020 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTs_l(fd,fs)   ( 0x46a00020 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTd_s(fd,fs)   ( 0x46000021 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTd_w(fd,fs)   ( 0x46800021 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTd_l(fd,fs)   ( 0x46a00021 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTw_s(fd,fs)   ( 0x46000024 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTw_d(fd,fs)   ( 0x46200024 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTl_s(fd,fs)   ( 0x46000025 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CVTl_d(fd,fs)   ( 0x46200025 | A_FD(fd) | A_FS(fs) )
#define TR_A_cop1_CF_s(fs,ft)     ( 0x46000030 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CF_d(fs,ft)     ( 0x46200030 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CUN_s(fs,ft)    ( 0x46000031 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CUN_d(fs,ft)    ( 0x46200031 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CEQ_s(fs,ft)    ( 0x46000032 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CEQ_d(fs,ft)    ( 0x46200032 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CUEQ_s(fs,ft)   ( 0x46000033 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CUEQ_d(fs,ft)   ( 0x46200033 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_COLT_s(fs,ft)   ( 0x46000034 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_COLT_d(fs,ft)   ( 0x46200034 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CULT_s(fs,ft)   ( 0x46000035 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CULT_d(fs,ft)   ( 0x46200035 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_COLE_s(fs,ft)   ( 0x46000036 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_COLE_d(fs,ft)   ( 0x46200036 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CULE_s(fs,ft)   ( 0x46000037 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CULE_d(fs,ft)   ( 0x46200037 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CSF_s(fs,ft)    ( 0x46000038 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CSF_d(fs,ft)    ( 0x46200038 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGLE_s(fs,ft)  ( 0x46000039 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGLE_d(fs,ft)  ( 0x46200039 | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CSEQ_s(fs,ft)   ( 0x4600003a | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CSEQ_d(fs,ft)   ( 0x4620003a | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGL_s(fs,ft)   ( 0x4600003b | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGL_d(fs,ft)   ( 0x4620003b | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CLT_s(fs,ft)    ( 0x4600003c | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CLT_d(fs,ft)    ( 0x4620003c | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGE_s(fs,ft)   ( 0x4600003d | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGE_d(fs,ft)   ( 0x4620003d | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CLE_s(fs,ft)    ( 0x4600003e | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CLE_d(fs,ft)    ( 0x4620003e | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGT_s(fs,ft)   ( 0x4600003f | A_FS(fs) | A_FT(ft) )
#define TR_A_cop1_CNGT_d(fs,ft)   ( 0x4620003f | A_FS(fs) | A_FT(ft) )




/**********************\
*                      *
* mnemonic debugging   *
*                      *
\**********************/

#define PRINT_NOTHING        "%s"
#define PRINT_PAR_NOTHING    ""
#define PRINT_DIS_NOTHING    ""

#define PRINT_RD             " r%02u"
#define PRINT_PAR_RD         tr_i->r.rd
#define PRINT_DIS_RD         D_RD(code)

#define PRINT_RS             " r%02u"
#define PRINT_PAR_RS         tr_i->r.rs
#define PRINT_DIS_RS         D_RS(code)

#define PRINT_I              " $%04hx (%+hd)"
#define PRINT_PAR_I          tr_i->i.i, tr_i->i.i
#define PRINT_DIS_I          D_I(code), D_I(code)

#define PRINT_T              " $%08x ($%08x)"
#define PRINT_PAR_T          (tr_i->j.t) << 2, tr_i->j.t
#define PRINT_DIS_T          (D_T(code)) << 2, D_T(code)

#define PRINT_RD_RS          " r%02u, r%02u"
#define PRINT_PAR_RD_RS      tr_i->r.rd, tr_i->r.rs
#define PRINT_DIS_RD_RS      D_RD(code), D_RS(code)

#define PRINT_RT_RD          " r%02u, r%02u"
#define PRINT_PAR_RT_RD      tr_i->r.rt, tr_i->r.rd
#define PRINT_DIS_RT_RD      D_RT(code), D_RD(code)

#define PRINT_RS_RT          " r%02u, r%02u"
#define PRINT_PAR_RS_RT      tr_i->r.rs, tr_i->r.rt
#define PRINT_DIS_RS_RT      D_RS(code), D_RT(code)

#define PRINT_RD_RT_SA       " r%02u, r%02u, %u"
#define PRINT_PAR_RD_RT_SA   tr_i->r.rd, tr_i->r.rt, tr_i->r.sa
#define PRINT_DIS_RD_RT_SA   D_RD(code), D_RT(code), D_SA(code)

#define PRINT_RD_RT_RS       " r%02u, r%02u, r%02u"
#define PRINT_PAR_RD_RT_RS   tr_i->r.rd, tr_i->r.rt, tr_i->r.rs
#define PRINT_DIS_RD_RT_RS   D_RD(code), D_RT(code), D_RS(code)

#define PRINT_RD_RS_RT       " r%02u, r%02u, r%02u"
#define PRINT_PAR_RD_RS_RT   tr_i->r.rd, tr_i->r.rs, tr_i->r.rt
#define PRINT_DIS_RD_RS_RT   D_RD(code), D_RS(code), D_RT(code)



#define PRINT_RS_I           " r%02u, $%04hx (%+hd) [instr.]"
#define PRINT_PAR_RS_I       tr_i->i.rs, tr_i->i.i, tr_i->i.i
#define PRINT_DIS_RS_I       D_RS(code), D_I(code), D_I(code)

#define PRINT_RT_I           " r%02u, $%04hx"
#define PRINT_PAR_RT_I       tr_i->i.rt, tr_i->i.i
#define PRINT_DIS_RT_I       D_RT(code), D_I(code)

#define PRINT_RS_RT_I        " r%02u, r%02u, $%04hx (%+hd) [instr.]"
#define PRINT_PAR_RS_RT_I    tr_i->i.rs, tr_i->i.rt, tr_i->i.i, tr_i->i.i
#define PRINT_DIS_RS_RT_I    D_RS(code), D_RT(code), D_I(code), D_I(code)

#define PRINT_RT_RS_I        " r%02u, r%02u, $%04hx (%+hd)"
#define PRINT_PAR_RT_RS_I    tr_i->i.rt, tr_i->i.rs, tr_i->i.i, tr_i->i.i
#define PRINT_DIS_RT_RS_I    D_RT(code), D_RS(code), D_I(code), D_I(code)



#define PRINT_RT_O_B         " r%02u, %+hd(r%02u)"
#define PRINT_PAR_RT_O_B     tr_i->i.rt, tr_i->i.i, tr_i->i.rs
#define PRINT_DIS_RT_O_B     D_RT(code), D_I(code), D_RS(code)



#define PRINT_FS_FT          " f%02u, f%02u"
#define PRINT_PAR_FS_FT      tr_i->f.fs, tr_i->f.ft
#define PRINT_DIS_FS_FT      D_FS(code), D_FT(code)

#define PRINT_FD_FS          " f%02u, f%02u"
#define PRINT_PAR_FD_FS      tr_i->f.fd, tr_i->f.fs
#define PRINT_DIS_FD_FS      D_FD(code), D_FS(code)

#define PRINT_FD_FS_FT       " f%02u, f%02u, f%02u"
#define PRINT_PAR_FD_FS_FT   tr_i->f.fd, tr_i->f.fs, tr_i->f.ft
#define PRINT_DIS_FD_FS_FT   D_FD(code), D_FS(code), D_FT(code)


#ifdef TR_DEBUG
        #define TR_PRINT_DEBUGGING_INFO(name, format)           \
                if(prefs.debug & DBG_CPU_REG)                   \
                {                                               \
                        cpu_print_gpr();                        \
                }                                               \
                if(prefs.debug & DBG_CPU_COUNT)                 \
                {                                               \
                        printf("%08"PR32"x: ", (tr_u32)reg.cpr[0][COUNT]);\
                        printf("[%08"PR32"x]: ", C_PC());\
                }                                               \
                if(prefs.debug & DBG_CPU_INSTR)                 \
                {                                               \
                        tr_io_printf(TR_IO_INFO, TR_IO_CPU,     \
                                "%-10s" PRINT_##format,         \
                                name, PRINT_PAR_##format);      \
                        putchar('\n');                          \
                }

        #define TR_PRINT_DEBUGGING_GEN_INFO(code, name, format) \
                if(prefs.debug & DBG_RSP_INSTR)                 \
                {                                               \
                        tr_io_printf(TR_IO_INFO, TR_IO_COMPILE, \
                                "%-10s" PRINT_##format,         \
                                name, PRINT_DIS_##format);      \
                        putchar('\n');                          \
                                                                \
                }
#else
        #define TR_PRINT_DEBUGGING_INFO(name, format)
        #define TR_PRINT_DEBUGGING_GEN_INFO(code, name, format)
#endif





#ifdef TR_HALT
        #define CPU_HALT reg.halt = 1; \
                         puts(ST_FG_RED "CPU halted!" ST_FG_DEF); fflush(stdout);
#else
        #define CPU_HALT /* */
#endif





#endif /* #ifndef __TR_COMPILE_STATIC_MNEMONIC_H */

