/**
 * TrueReality - N64/exceptions.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>

#include <config.h>
#include <parser_extern.h>

#include "registers.h"
#include "exceptions.h"

#ifdef TR_STATIC_COMPILER
        #include "compile_static_mnemonic.h"
#endif

#ifdef TR_DEBUG
        #include <debug.h>
#endif
   




int tlb_load_or_store;   /* for TBL invalid and miss exceptions */





void rs4300i_trap_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Tr;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_EXC_TR)
                printf("Exception (Tr) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_trap_exception() */





void rs4300i_break_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Bp;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_EXC_BP)
                printf("Exception (Bp) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_break_exception() */





void rs4300i_syscall_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Sys;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_EXC_SYS)
                printf("Exception (Sys) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_syscall_exception() */





void rs4300i_integer_overflow_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Ov;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_EXC_OV)
                printf("Exception (Ov) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_integer_overflow_exception() */





void rs4300i_floating_point_exception()
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_FPE;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_EXC_FPE)
                printf("Exception (FPE) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_integer_overflow_exception() */





void rs4300i_reserved_instruction_exception()
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Ri;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_EXC_RI)
                printf("Exception (Ri) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_reserved_instruction_exception() */





void rs4300i_tlb_miss_exception()
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        if(tlb_load_or_store == EXC_TLB_STORE)
        {
            /* EXC_TLB_STORE */
                reg.cpr[0][CAUSE] |= EXC_CODE_TLBS;
#ifdef TR_DEBUG
                if(prefs.debug & DBG_EXC_TLB)
                        printf("Exception (TLBS (miss)) @ %08lx\n", (tr_u32)reg.cpr[0][COUNT]); fflush(stdout);
#endif
        }
        else
        {
            /* EXC_TLB_LOAD */
                reg.cpr[0][CAUSE] |= EXC_CODE_TLBL;
#ifdef TR_DEBUG
                if(prefs.debug & DBG_EXC_TLB)
                        printf("Exception (TLBL (miss)) @ %08lx\n", (tr_u32)reg.cpr[0][COUNT]); fflush(stdout);
#endif
        }

        if(reg.cpr[0][STATUS] & 0x00000002)   /* check EXL */
        {
                jump_to_common_exception_vector();
        }
        else
        {
                jump_to_tlb_exception_vector();
        }

} /* void rs4300i_tlb_miss_exception() */





void rs4300i_tlb_invalid_exception(int store)
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        if(tlb_load_or_store == EXC_TLB_STORE)
        {
            /* EXC_TLB_STORE */
                reg.cpr[0][CAUSE] |= EXC_CODE_TLBS;
#ifdef TR_DEBUG
                if(prefs.debug & DBG_EXC_TLB)
                        printf("Exception (TLBS (invalid)) @ %08lx\n", (tr_u32)reg.cpr[0][COUNT]); fflush(stdout);
#endif
        }
        else
        {
            /* EXC_TLB_LOAD */
                reg.cpr[0][CAUSE] |= EXC_CODE_TLBL;
#ifdef TR_DEBUG
                if(prefs.debug & DBG_EXC_TLB)
                        printf("Exception (TLBL (invalid)) @ %08lx\n", (tr_u32)reg.cpr[0][COUNT]); fflush(stdout);
#endif
        }

        jump_to_common_exception_vector();

} /* void rs4300i_tlb_invalid_exception() */





void rs4300i_tlb_modification_exception()
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Mod;

#ifdef TR_DEBUG
                if(prefs.debug & DBG_EXC_TLB)
                        printf("Exception (TLB (modification)) @ %08lx\n", (tr_u32)reg.cpr[0][COUNT]); fflush(stdout);
#endif

        jump_to_common_exception_vector();

} /* void rs4300i_tlb_modification_exception() */











void jump_to_tlb_exception_vector()
{
        reg.cpr[0][STATUS] |= 0x00000002;               /* set EXL in the STATUS reg        */
#ifdef TR_STATIC_COMPILER
        reg.cpr[0][EPC]  = C_PC();                        /* store PC in EPC                  */
#else
        reg.cpr[0][EPC]  = reg.pc;                      /* store PC in EPC                  */
#endif

        if(reg.delay == EXEC_DELAY)
        {
                reg.cpr[0][CAUSE] |= 0x80000000;                /* set BD bit in CAUSE reg          */
                reg.cpr[0][EPC]   -= 4;                         /* store pc before branch instr     */
        }
        else
        {
                reg.cpr[0][CAUSE] &= ~0x80000000;               /* clr BD bit in CAUSE reg          */
        }

    /* set pc to 0x80000000 or 0xbfc00200 (tlb exception vector) */
        if(reg.cpr[0][STATUS] & 0x00400000)   /* check BEV */
        {
#ifdef TR_STATIC_COMPILER
/* BUG: this must not happen! */
		printf("\n\nBUG in jump_to_tlb_exception_vector!!!\n\n");
		reg.halt = 1;
/*
		??? tr_i = &mem.rd_ram_instr[0x/4];
		reg.delay_jump = 0x80000180;
		reg.delay = NO_DELAY;
*/
                reg.delay_jump = 0xbfc00200;
#else
                reg.pc = 0xbfc00200;
#endif
        }
        else
        {
#ifdef TR_STATIC_COMPILER
		tr_i = &mem.rd_ram_instr[0];
                reg.delay_jump = 0x80000000;
#else
                reg.pc = 0x80000000;
#endif
        }
        reg.delay = NO_DELAY;  /* next instruction must NOT be the instr in the delay slot !!! */

} /* void jump_to_tlb_refill_vector() */





void jump_to_common_exception_vector()
{
        reg.cpr[0][STATUS] |= 0x00000002;               /* set EXL in the STATUS reg        */
#ifdef TR_STATIC_COMPILER
        reg.cpr[0][EPC]  = C_PC();
#else
        reg.cpr[0][EPC]  = reg.pc;                      /* store PC in EPC                  */
#endif

        if(reg.delay == EXEC_DELAY)
        {
                reg.cpr[0][CAUSE] |= 0x80000000;                /* set BD bit in CAUSE reg          */
                reg.cpr[0][EPC]   -= 4;                         /* store pc before branch instr     */

        }
        else
        {
                reg.cpr[0][CAUSE] &= ~0x80000000;               /* clr BD bit in CAUSE reg          */
        }

#ifdef TR_STATIC_COMPILER
	tr_i = &mem.rd_ram_instr[0x180/4];
        reg.delay_jump = 0x80000180;
        reg.delay = NO_DELAY;
#else
        reg.pc = 0x80000180;    /* set pc to 0x80000180 (common exception vector) */
        reg.delay = NO_DELAY;   /* next instruction must NOT be the instr in the delay slot !!! */
#endif

} /* void jump_to_common_exception_vector() */

