/**
 * TrueReality - N64/exceptions.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#ifndef __TR_EXCEPTIONS_H
#define __TR_EXCEPTIONS_H





#include <type_sizes.h>





#define EXC_CODE(x)     ((x)<<2)

#define EXC_CODE_Int    EXC_CODE(0)
#define EXC_CODE_Mod    EXC_CODE(1)     /* TLB Modification exception */
#define EXC_CODE_TLBL   EXC_CODE(2)     /* TLB Miss exception (load or instruction fetch) */
#define EXC_CODE_TLBS   EXC_CODE(3)     /* TLB Miss exception (store) */
#define EXC_CODE_AdEL   EXC_CODE(4)
#define EXC_CODE_AdES   EXC_CODE(5)
#define EXC_CODE_IBE    EXC_CODE(6)
#define EXC_CODE_DBE    EXC_CODE(7)
#define EXC_CODE_Sys    EXC_CODE(8)
#define EXC_CODE_Bp     EXC_CODE(9)
#define EXC_CODE_Ri     EXC_CODE(10)
#define EXC_CODE_CpU    EXC_CODE(11)
#define EXC_CODE_Ov     EXC_CODE(12)
#define EXC_CODE_Tr     EXC_CODE(13)
#define EXC_CODE_FPE    EXC_CODE(15)
#define EXC_CODE_WATCH  EXC_CODE(23)

#define EXC_CODE__MASK  0x7c
#define EXC_CODE__SHIFT 2



#define EXC_TLB_STORE 1
#define EXC_TLB_LOAD  0





extern int tlb_load_or_store;   /* for TBL invalid and miss exceptions */





void    rs4300i_syscall_exception();
void    rs4300i_trap_exception();
void    rs4300i_break_exception();
void    rs4300i_integer_overflow_exception();
void    rs4300i_floating_point_exception();
void    rs4300i_reserved_instruction_exception();
void    rs4300i_tlb_miss_exception();
void    rs4300i_tlb_invalid_exception();
void    rs4300i_tlb_modification_exception();




void    jump_to_common_exception_vector();
void    jump_to_tlb_exception_vector();





#endif /* #ifndef __TR_EXCEPTIONS_H */

