/**
 * TrueReality - N64/mnemonic.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#ifndef __TR_MNEMONIC_H
#define __TR_MNEMONIC_H





#include <type_sizes.h>

#include <file_loader_extern.h>

#include <ansi_esc_codes.h>





/**********************\
*                      *
* instruction reader   *
*                      *
\**********************/


#define READ_CPU_INSTRUCTION(addr)                                                             \
(                                                                                              \
        ( (addr < 0xa4000040) )                                                                \
        ?                                                                                      \
                ( *((tr_u32 *)(mem.rd_ram + (addr & 0x1fffffff))) )     /* instr in prgcode */ \
        :                                                                                      \
                ( (addr < 0xa4002000) )                                                        \
                ?                                                                              \
                        ( *(tr_u32 *)(mem.sp_dmem + (addr & 0xfff)) )   /* instr in bootcode */\
                :                                                                              \
                        ( *(tr_u32 *)(rom.image + (addr & 0x0fffffff)) )         /* for test */\
)





/**********************\
*                      *
* mnemonic definitions *
*                      *
\**********************/

#define __CODE          (reg.code = READ_CPU_INSTRUCTION(reg.pc))
#define __NEXT_CODE     (READ_CPU_INSTRUCTION((reg.pc + 4)))
#define __PREV_CODE     (READ_CPU_INSTRUCTION((reg.pc - 4)))

#define __OPCODE        ((tr_u8)(__CODE >> 26))

#define __RS            (((tr_u8)(reg.code >> 21)) & 0x1f)

#define __RT            (((tr_u8)(reg.code >> 16)) & 0x1f)

#define __RD            (((tr_u8)(reg.code >> 11)) & 0x1f)

#define __SA            (((tr_u8)(reg.code >>  6)) & 0x1f)

#define __F             ( (tr_u8)(reg.code)       & 0x3f)

#define __I             ( (tr_s32)(tr_s16)reg.code )
#define __O             ( reg.pc + 4 + (__I << 2) )

#define ____T           (reg.code & 0x03ffffff)
#define __T             ( (reg.pc & 0xf0000000) | (____T << 2) )





/******************************\
*                              *
* mnemonic definitions for fpu *
*                              *
\******************************/

#define FMT_S   16      /* 32bit binary floating-point */
#define FMT_D   17      /* 64bit binary floating-point */
#define FMT_W   20      /* 32bit binary fixed-point    */
#define FMT_L   21      /* 64bit binary fixed-point    */


#define __FMT   __RS

#define __FT    __RT

#define __FS    __RD

#define __FD    __SA

#define FT                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FT]                                        \
        :                                                               \
                ((reg.cpr[1][__FT+1] << 32) | (reg.cpr[1][__FT] & 0xffffffff))       \
)

#define FD                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FD]                                        \
        :                                                               \
                ((reg.cpr[1][__FD+1] << 32) | (reg.cpr[1][__FD] & 0xffffffff))       \
)


#define FS                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FS]                                        \
        :                                                               \
                ((reg.cpr[1][__FS+1] << 32) | (reg.cpr[1][__FS] & 0xffffffff))       \
)


#define STORE_FPR(value)                                                \
{                                                                       \
        if(reg.cpr[0][12] & 0x04000000)                                 \
        {                                                               \
                reg.cpr[1][__FD] = (value);                             \
        }                                                               \
        else                                                            \
        {                                                               \
                reg.cpr[1][__FD+0] = (value);                           \
                reg.cpr[1][__FD+0] = (tr_s64)(tr_s32)reg.cpr[1][__FD+0]; \
                reg.cpr[1][__FD+1] = (value) >> 32;                     \
                reg.cpr[1][__FD+1] = (tr_s64)(tr_s32)reg.cpr[1][__FD+1]; \
        }                                                               \
}

        




/**********************\
*                      *
* mnemonic debugging   *
*                      *
\**********************/

#define PRINT_NOTHING

#define PRINT_O         printf(" $%08"PR32"x", __O);
#define PRINT_RD        printf(" r%02"PR8"u", __RD);
#define PRINT_RS        printf(" r%02"PR8"u", __RS);
#define PRINT_T         printf(" $%08"PR32"x", __T); PRINT_SYMBOL(__T)

#define PRINT_RD_RS     printf(" r%02"PR8"u, r%02"PR8"u", __RD, __RS);
#define PRINT_RS_I      printf(" r%02"PR8"u, $%08"PR32"x (%"PR32"d)", __RS, (tr_s32)__I, (tr_s32)__I);
#define PRINT_RS_O      printf(" r%02"PR8"u, $%08"PR32"x", __RS, __O);
#define PRINT_RS_RD     printf(" r%02"PR8"u, r%02"PR8"u", __RS, __RD);
#define PRINT_RS_RT     printf(" r%02"PR8"u, r%02"PR8"u", __RS, __RT);
#define PRINT_RT_I      printf(" r%02"PR8"u, $%04"PR16"x ($%08"PR32"x)", __RT, (tr_u16)__I, (tr_s32)(__I << 16));
#define PRINT_RT_RD     printf(" r%02"PR8"u, r%02"PR8"u", __RT, __RD);

#define PRINT_RD_RS_RT  printf(" r%02"PR8"u, r%02"PR8"u, r%02"PR8"u", __RD, __RS, __RT);
#define PRINT_RD_RT_SA  printf(" r%02"PR8"u, r%02"PR8"u, %02"PR8"u", __RD, __RT, __SA);
#define PRINT_RD_RT_RS  printf(" r%02"PR8"u, r%02"PR8"u, r%02"PR8"u", __RD, __RT, __RS);
#define PRINT_RS_RT_O   printf(" r%02"PR8"u, r%02"PR8"u, $%08"PR32"x", __RS, __RT, __O);
#define PRINT_RT_O_B    printf(" r%02"PR8"u, $%04"PR32"x(r%02"PR8"u)", __RT, __I, __RS);
#define PRINT_RT_RS_I   printf(" r%02"PR8"u, r%02"PR8"u, $%04"PR16"x (%"PR16"d)", __RT, __RS, (tr_s16)__I, (tr_s16)__I);

#define PRINT_FD_FS_FT  printf(" f%02"PR8"u, f%02"PR8"u, f%02"PR8"u", __FD, __FS, __FT);
#define PRINT_FD_FS     printf(" f%02"PR8"u, f%02"PR8"u", __FD, __FS);
#define PRINT_FT_O_B    printf(" f%02"PR8"u, $%04"PR32"x(r%02"PR8"u)", __FT, __I, __RS);
#define PRINT_RT_FS     printf(" r%02"PR8"u, f%02"PR8"u", __RT, __FS);

#define PRINT_C0        printf(" r%02"PR8"u, cpr0[%02"PR8"u]", __RT, __RD);
#define PRINT_C0C       printf(" r%02"PR8"u, ccr0[%02"PR8"u]", __RT, __RD);
#define PRINT_RT_C1     printf(" r%02"PR8"u, ccr1[%02"PR8"u]", __RT, __FS);
#define PRINT_C2_O_B    printf(" cpr2[%02"PR8"u], $%04"PR32"x(r%02"PR8"u)", __RT, __I, __RS);




#define PRINT_FMT               \
        switch(__FMT)           \
        {                       \
            case FMT_S:         \
                putchar('s');   \
                break;          \
            case FMT_D:         \
                putchar('d');   \
                break;          \
            case FMT_W:         \
                putchar('w');   \
                break;          \
            case FMT_L:         \
                putchar('l');   \
                break;          \
            default:            \
                putchar('?');   \
        }





#ifdef TR_DEBUG_NOTIMPLEMENTED
        #define PRINT_NOT_IMPLEMENTED(x)        \
                puts(x" - NOT IMPLEMENTED");    \
                fflush(stdout);
#else
        #define PRINT_NOT_IMPLEMENTED(x)
#endif




#ifdef TR_DEBUG_XINST
        #define PRINT_XINST(x)                  \
                if(prefs.debug & DBG_CPU_INSTR) \
                {                               \
                        puts(x); fflush(stdout);\
                }
#else
        #define PRINT_XINST(x)
#endif





#ifdef TR_DEBUG
        #define PRINT_DEBUG_HEAD                                        \
                if(prefs.debug & DBG_CPU_COUNT)                         \
                        printf("CPU(%08"PR32"x): ", (tr_u32)reg.cpr[0][COUNT]);\
                else                                                    \
                        fputs("CPU: ", stdout);                         \
                printf("[$%08"PR32"x]: ", reg.pc);                           \
                printf("$%08"PR32"x: ", reg.code);

        #define PRINT_DEBUG_FOOT                                        \
                if(prefs.debug & DBG_CPU_REG)                           \
                        cpu_print_gpr();
#else
        #define PRINT_DEBUG_HEAD
        #define PRINT_DEBUG_FOOT
#endif





#ifdef TR_DEBUG
        #define PRINT_DEBUGGING_INFO(name, format)              \
                if(prefs.debug & DBG_CPU_INSTR)                 \
                {                                               \
                        PRINT_DEBUG_HEAD                        \
                        printf("%-10s", name);                  \
                        PRINT_##format                          \
                        putchar('\n');                          \
                        PRINT_DEBUG_FOOT                        \
                                                                \
                }
#else
        #define PRINT_DEBUGGING_INFO(name, format)
#endif





#ifdef TR_DEBUG
        #define PRINT_DEBUGGING_INFO_FPU(name, format)  \
                if(prefs.debug & DBG_CPU_INSTR)         \
                {                                       \
                        int i;                          \
                                                        \
                        PRINT_DEBUG_HEAD                \
                        printf(" %s", name);            \
                        PRINT_FMT                       \
                        for(i=0; i<9-strlen(name); i++) \
                                putchar(' ');           \
                        PRINT_##format                  \
                        putchar('\n');                  \
                        PRINT_DEBUG_FOOT                \
                }
#else
        #define PRINT_DEBUGGING_INFO_FPU(name, format)
#endif





#ifdef TR_DEBUG
        #define PRINT_DEBUGGING_INFO_FPU_COND                   \
                if(prefs.debug & DBG_CPU_INSTR)                 \
                {                                               \
                        PRINT_DEBUG_HEAD                        \
                                                                \
                        switch(__F)                             \
                        {                                       \
                            int i;                              \
                                                                \
                            case 0x30:                          \
                                fputs(" C.f.", stdout);         \
                                PRINT_FMT                       \
                                for(i=0; i<5; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x31:                          \
                                fputs(" C.un.", stdout);        \
                                PRINT_FMT                       \
                                for(i=0; i<4; i++)              \
                                        putchar(' ');           \
                               break;                           \
                            case 0x32:                          \
                                fputs(" C.eq.", stdout);        \
                                PRINT_FMT                       \
                                for(i=0; i<4; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x33:                          \
                                fputs(" C.ueq.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x34:                          \
                                fputs(" C.olt.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x35:                          \
                                fputs(" C.ult.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x36:                          \
                                fputs(" C.ole.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x37:                          \
                                fputs(" C.ule.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x38:                          \
                                fputs(" C.sf.", stdout);        \
                                PRINT_FMT                       \
                                for(i=0; i<4; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x39:                          \
                                fputs(" C.ngle.", stdout);      \
                                PRINT_FMT                       \
                                for(i=0; i<2; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x3a:                          \
                                fputs(" C.seq.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x3b:                          \
                                fputs(" C.ngl.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x3c:                          \
                                fputs(" C.lt.", stdout);        \
                                PRINT_FMT                       \
                                for(i=0; i<4; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x3d:                          \
                                fputs(" C.nge.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x3e:                          \
                                fputs(" C.le.", stdout);        \
                                PRINT_FMT                       \
                                for(i=0; i<4; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            case 0x3f:                          \
                                fputs(" C.ngt.", stdout);       \
                                PRINT_FMT                       \
                                for(i=0; i<3; i++)              \
                                        putchar(' ');           \
                                break;                          \
                            default:                            \
                                fputs(" C.?.", stdout);         \
                                PRINT_FMT                       \
                                for(i=0; i<5; i++)              \
                                        putchar(' ');           \
                        }                                       \
                        PRINT_FD_FS                             \
                        putchar('\n');                          \
                        PRINT_DEBUG_FOOT                        \
                }
#else
        #define PRINT_DEBUGGING_INFO_FPU_COND /* */
#endif





#ifdef TR_HALT
        #define CPU_HALT reg.halt = 1; \
                         puts(ST_FG_RED "CPU halted!" ST_FG_DEF); fflush(stdout);
#else
        #define CPU_HALT /* */
#endif




#define PRINT_SYMBOL(a)                                                 \
{                                                                       \
        int     i;                                                      \
                                                                        \
        for(i=0; prefs.symbols==USE_SYMBOLS && i<prefs.n_symbols; i++)  \
        {                                                               \
                if(((prefs.symtab)[i]).addr == a)                       \
                {                                                       \
                        printf(" (%s)", ((prefs.symtab)[i]).name);      \
                }                                                       \
        }                                                               \
}





#endif /* #ifndef __TR_MNEMONIC_H */

