/**
 * TrueReality - N64/rdp.c
 * Copyright (C) 1998, 1999 by Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#define TR_UCODE 0




/* includes - everyone needs them ;-) */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <config.h>
#include <parser_extern.h>
#include <memory_extern.h>
#include <type_sizes.h>

#ifdef TR_DEBUG
        #include <debug.h>
#endif

#include "../output/display/renderer.h"
#include "../output/display/display.h"

#include "rsp.h"
#include "rdp.h"
#include "rsp_registers.h"
#include "rdp_registers.h"
#include "mnemonic.h"
#include "registers.h"
#include "memory.h"










/* define macros - to speed up things and to make things looking nice */

#ifdef TR_DEBUG
        #define PRINT_RDP_NOT_IMPLEMENTED(name)                                                 \
                /* printf(ST_FG_BRO "$%03"RP32"x %s - NOT IMPLEMENTED\n" ST_FG_DEF, rdp_reg.cmd0>>24, name); */    \
                printf("$%02"PR32"x %s - NOT IMPLEMENTED\n", rdp_reg.cmd0>>24, name);    \
                fflush(stdout);

        #define PRINT_RSP_AUDIO_NOT_IMPLEMENTED(name)                                   \
                printf("$%02"PR32"x %s (AUDIO) - NOT IMPLEMENTED\n", rdp_reg.cmd0>>24, name);\
                fflush(stdout);

        #define PRINT_RDP_WARNING(text)                                         \
                if(prefs.debug & DBG_RDP)                                       \
                {                                                               \
                        printf(ST_FG_BRO "RDP WARNING: %s\n" ST_FG_DEF, text);  \
                        fflush(stdout);                                         \
                }

        #define PRINT_RDP_INFO(name)                                    \
                if(prefs.debug & DBG_RDP)                               \
                {                                                       \
                        printf(ST_BG_WHI ST_FG_BLU                      \
                                "$%02"PR32"x %s"                             \
                                ST_FG_DEF ST_BG_DEF,                    \
                                rdp_reg.cmd0>>24, name);                \
                        fflush(stdout);                                 \
                }

        void    PRINT_RDP_MNEMONIC(char *fmt, ...)
        {
                va_list ap;

                if(prefs.debug & DBG_RDP)
                {
                        va_start(ap, fmt);
                        vprintf(fmt, ap);
                        fflush(stdout);
                        va_end(ap);
                }
        }
#ifdef MACOS
        #undef PRINT_RDP_NOT_IMPLEMENTED
        #define PRINT_RDP_NOT_IMPLEMENTED(name)                 \
                if(!IsInIgnoreList(rdp_reg.cmd0))               \
                {                                               \
                        sprintf(gHaltInstrStr, "RDP: %s", name);\
                                gIgnoreAddr = rdp_reg.cmd0;     \
                        reg.halt = 1;                           \
                }                                               \
                return;
        #undef PRINT_RDP_AUDIO_NOT_IMPLEMENTED
        #define PRINT_RDP_AUDIO_NOT_IMPLEMENTED(name)           \
                printf("$%02"PR32"x %s (AUDIO) - NOT IMPLEMENTED\n", rdp_reg.cmd0>>24, name);\
                if(!IsInIgnoreList(-1))                         \
                {                                               \
                        sprintf(gHaltInstrStr, "RDP: 0x%02"PR32"x %s (AUDIO) - NOT IMPLEMENTED", rdp_reg.cmd0>>24, name);\
                                gIgnoreAddr = -1;               \
                        reg.halt = 1;                           \
                }                                               \
                return;
#endif


#else
        #define PRINT_RSP_AUDIO_NOT_IMPLEMENTED(name)
        #define PRINT_RDP_NOT_IMPLEMENTED(name)
        #define PRINT_RDP_AUDIO_NOT_IMPLEMENTED(name)
        #define PRINT_RDP_WARNING(text)
        #define PRINT_RDP_INFO(name)
        void    PRINT_RDP_MNEMONIC(char *fmt, ...)
        {
        }
#endif










/* static prototypes */

#ifdef HLE
        static  tr_u32 segoffset2addr(tr_u32 so);   /* calcs the real addr out of seg and offset */
#endif





/* hle rsp instructions (audio & graphics) */

static void fixme();
static void spnoop();

/* rdp graphics instructions */

static void rdp_noop();
static void rdp_texrect();
static void rdp_texrectflip();
static void rdp_loadsync();
static void rdp_pipesync();
static void rdp_tilesync();
static void rdp_fullsync();
static void rdp_setkeygb();
static void rdp_setkeyr();
static void rdp_setconvert();
static void rdp_setscissor();
static void rdp_setprimdepth();
static void rdp_setothermode();
static void rdp_loadtlut();
static void rdp_settilesize();
static void rdp_loadblock();
static void rdp_loadtile();
static void rdp_settile();
static void rdp_fillrect();
static void rdp_setfillcolor();
static void rdp_setfogcolor();
static void rdp_setblendcolor();
static void rdp_setprimcolor();
static void rdp_setenvcolor();
static void rdp_setcombine();
static void rdp_settextureimage();
static void rdp_setdepthimage();
static void rdp_setcolorimage();
static void rdp_trifill();
static void rdp_trishade();
static void rdp_tritxtr();
static void rdp_trishadetxtr();
static void rdp_trifillz();
static void rdp_trishadez();
static void rdp_tritxtrz();
static void rdp_trishadetxtrz();










/* variables section */

struct rdp_reg rdp_reg;





/* special section - it includes the HLE include files. the
   files should contain the prototypes and the RSP HLE functions
   itself - so we can use them in the following funktion pointer
   array. */

#ifdef HLE

#include "hle_uc.incl"     /* here are the pointer arrays  */
#include "hle_uc00.incl"
#include "hle_uc01.incl"
#include "hle_uc02.incl"

#endif /* #ifdef HLE */





#ifndef HLE

/* rdp graphics instruction function pointer */
/* in HLE mode the pointer array is in hle_ucX.incl! */

static rdp_instr gfx_instruction[1][256] =
{
    {
      /**
      * DMA commands:
      * We are in native RSP emulation - here are no commands!
      **/
        fixme,                      fixme,                      fixme,                      fixme,
        fixme,                      fixme,                      fixme,                      fixme,
        fixme,                      fixme,                      fixme,                      fixme,
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
      /**
      * The next 64 commands are not used (as far as i know).
      **/
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
      /**
      * IMMEDIATE commands:
      * We are in native RSP emulation - here are no commands!
      **/
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,        
        fixme,                      fixme,                      fixme,                      fixme,
        fixme,                      fixme,                      fixme,                      fixme,
        fixme,                      fixme,                      fixme,                      fixme,
        fixme,                      fixme,                      fixme,                      fixme,
      /**
      * RDP commands:
      * These 64 commands are the real RDP commands.
      * So they are used here ;-)
      **/
        rdp_noop,               fixme,                  fixme,                  fixme,    
        fixme,                  fixme,                  fixme,                  fixme,    
        rdp_trifill,            rdp_trifillz,           rdp_tritxtr,            rdp_tritxtrz,
        rdp_trishade,           rdp_trishadez,          rdp_trishadetxtr,       rdp_trishadetxtrz,
        fixme,                  fixme,                  fixme,                  fixme,    
        fixme,                  fixme,                  fixme,                  fixme,    
        fixme,                  fixme,                  fixme,                  fixme,    
        fixme,                  fixme,                  fixme,                  fixme,    
        fixme,                  fixme,                  fixme,                  fixme,    
        rdp_texrect,            rdp_texrectflip,        rdp_loadsync,           rdp_pipesync,
        rdp_tilesync,           rdp_fullsync,           rdp_setkeygb,           rdp_setkeyr,
        rdp_setconvert,         rdp_setscissor,         rdp_setprimdepth,       rdp_setothermode,
        rdp_loadtlut,           fixme,                  rdp_settilesize,        rdp_loadblock,
        rdp_loadtile,           rdp_settile,            rdp_fillrect,           rdp_setfillcolor,
        rdp_setfogcolor,        rdp_setblendcolor,      rdp_setprimcolor,       rdp_setenvcolor,
        rdp_setcombine,         rdp_settextureimage,    rdp_setdepthimage,      rdp_setcolorimage
    }
};

#endif /* #ifndef HLE */









/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - global routines                          *
*                                                                              *
\******************************************************************************/




void rdp_go()
{
#ifdef HLE
        return;
#endif
        /* we should do a fork() here - but!!! how could we stay portable? */
        /* maybe all these commands should be buffered and executed in a vert retr */

        rdp_reg.start = rdp_reg.curr = ((tr_u32 *)mem.dpc_reg)[0] & 0xfff;
        rdp_reg.stop  =                ((tr_u32 *)mem.dpc_reg)[1] & 0xfff;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_RDP)
                printf("start: $%"PR32"x; stop: $%"PR32"x;\n", rdp_reg.start, rdp_reg.stop);
#endif



        ((tr_u32 *)mem.dpc_reg)[2] = rdp_reg.curr;

        if(rdp_reg.start == rdp_reg.stop)
                return;

        rdp_reg.curr = rdp_reg.stop - 8;

#ifdef TR_DEBUG
        if(prefs.debug & DBG_RDP)
                printf("RDP: [$%"PR32"x]: ", rdp_reg.curr);
#endif

        rdp_reg.cmd0 = ((tr_u32 *)mem.sp_dmem)[(rdp_reg.curr>>2)+0];
        rdp_reg.cmd1 = ((tr_u32 *)mem.sp_dmem)[(rdp_reg.curr>>2)+1];

#ifdef HLE
        switch(rdp_reg.task)
        {
            case AUDIO_TASK:
                aud_instruction[rdp_reg.ucode_nr][rdp_reg.cmd0>>24]();
                break;

            case GFX_TASK:
                gfx_instruction[rdp_reg.ucode_nr][rdp_reg.cmd0>>24]();
                break;
        }
#else
        gfx_instruction[0][rdp_reg.cmd0>>24]();
#endif

        ((tr_u32 *)mem.dpc_reg)[2] = (rdp_reg.curr += 8);
/*
#ifdef TR_DEBUG
        if(prefs.debug & DBG_RDP)
        {
                fputs("Press a key!", stdout);
                fflush(stdout);
                getchar();
        }
#endif
*/
} /* void rdp_go() */





void rdp_dis()
{

} /* void rdp_dis() */





void rdp_reset()
{
#ifdef HLE
        rdp_reg.pc_i = 0;
        rdp_reg.halt = 1;
#endif

        rdp_reg.lights = 1;

        rdp_reg.tile = 0;
        rdp_reg.loadtile = 7;
        rdp_reg.td[0].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[1].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[2].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[3].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[4].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[5].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[6].set_by = RDPTD_LOAD_NOT_SET;
        rdp_reg.td[7].set_by = RDPTD_LOAD_NOT_SET;

} /* void rdp_reset() */





#ifdef HLE

static int hle_wait = 0;

void rdp_step()
{
        tr_u32   a;
#ifdef TR_DEBUG
        int     i;
#endif



    /* hehe - this makes lcars.bin work - otherwise rdp is too fast :) */
    /* this should be a command line option - i know ... */
        if(hle_wait++ < 130)
        {
                return;
        }


        a = rdp_reg.pc[rdp_reg.pc_i];

        rdp_reg.cmd0 = ((tr_u32 *)mem.rd_ram)[(a>>2)+0];
        rdp_reg.cmd1 = ((tr_u32 *)mem.rd_ram)[(a>>2)+1];

#ifdef TR_DEBUG
        if(prefs.debug & DBG_RDP)
        {
                i = 0;
                while(i <= rdp_reg.pc_i)
                {
                        printf("$%08"PR32"x ", rdp_reg.pc[i]);
                        i++;
                }
                putchar('\n');

                printf("$%08"PR32"x $%08"PR32"x\n", rdp_reg.cmd0, rdp_reg.cmd1);
                
                fflush(stdout);
        }
#endif

        rdp_reg.curr = ((tr_u32 *)mem.dpc_reg)[2] = \
        rdp_reg.pc[rdp_reg.pc_i] = a + 8;   /* point to next instruction */

        switch(rdp_reg.task)
        {
            case AUDIO_TASK:
#ifdef AUDIO
                aud_instruction[0][rdp_reg.cmd0>>24]();
#endif
                rdp_reg.size -= 8;
                if(rdp_reg.size == 0)
                {
                        PRINT_RDP_MNEMONIC("RDP stops here - RSP BREAK will be simulated\n");
                        rdp_fullsync();   /* from jeff */
                
                        rsp_reg.code = 0x0000000d;
                        rsp_dis(0x0000000d);   /* RSP BREAK instruction */
                        rdp_reg.halt = 1;
                /* seems not to work with this
                        reg.do_or_check_sthg = RS4300I_DO_AI_INTERRUPT;
                */
                }
                break;

            case GFX_TASK:
                gfx_instruction[rdp_reg.ucode_nr][rdp_reg.cmd0>>24]();
                break;
        }

/*
#ifdef TR_DEBUG
        if(prefs.debug & DBG_RDP)
        {
                fputs("Press a key!", stdout);
                fflush(stdout);
                getchar();
        }
#endif
*/
} /* void rdp_step() */





void rdp_dl_check()
{
        tr_u32 us, uds, dss;



        us  = ((tr_u32 *)mem.sp_dmem)[0xfd4>>2];  /* ucode size */
        uds = ((tr_u32 *)mem.sp_dmem)[0xfdc>>2];  /* ucode data size */
        dss = ((tr_u32 *)mem.sp_dmem)[0xfe4>>2];  /* dram stack size */

        switch(((tr_u32 *)mem.sp_dmem)[0xfc0>>2])
        {
            case 1:                             /* Graphic Task */
                PRINT_RDP_MNEMONIC("GRAPHICS TASK -");
#if 1
                printf("GRAPHICS TASK -");
#endif

                rdp_reg.task = GFX_TASK;

                if(us != 0x1000)
                {
                        PRINT_RDP_MNEMONIC("\n        UCODE SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(uds != 0x800)
                {
                        PRINT_RDP_MNEMONIC("\n        UCODE DATA SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(dss < 640)
                {
                        PRINT_RDP_MNEMONIC("\n        DRAM STACK SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
#ifdef DISPLAY
                if(prefs.display != NO_DISPLAY && rdp_reg.task != INVALID_TASK)
                {
                        InitVisualRenderBuffer();
                        /* rdp_reset(); */
                }
#endif
                break;

            case 2:                             /* Audio Task */
                PRINT_RDP_MNEMONIC("AUDIO TASK -");
#if 1
                printf("AUDIO TASK -");
#endif

                rdp_reg.task = AUDIO_TASK;

                if(us > 0x1000)
                {
                        PRINT_RDP_MNEMONIC("\n        UCODE SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(uds > 0x800)
                {
                        PRINT_RDP_MNEMONIC("\n        UCODE DATA SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(dss != 0)
                {
                        PRINT_RDP_MNEMONIC("\n        DRAM STACK SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                break;

            default:
                PRINT_RDP_MNEMONIC("UNKNOWN TASK - WILL NOT RUN HLE\n");
                return;

        } /* switch(mem.sp_dmem[0xfc0>>2]) */

    
    
    /* for audio and graphic task */

        hle_wait = 0;

        if(rdp_reg.task == INVALID_TASK)
        {
                puts("\n        WILL NOT RUN IN HLE !!!");
        }
        else
        {
                rdp_reg.pc[0] = rdp_reg.curr = ((tr_u32 *)mem.dpc_reg)[2] =
                        ((tr_u32 *)mem.sp_dmem)[0xff0>>2];
                rdp_reg.size = ((tr_u32 *)mem.sp_dmem)[0xff4>>2];  /* we need this in audio task */
                
                rdp_reg.ucode_nr = TR_UCODE;   /* later we have to make this dynamic */
                
                rdp_reg.pc_i = 0;
                rdp_reg.halt = 0;
                PRINT_RDP_MNEMONIC(" WILL RUN HLE\n");
#if 1
                printf(" WILL RUN HLE\n");
#endif
        }

} /* void rdp_dl_check() */

#endif /* #ifdef HLE */










/******************************************************************************\
*                                                                              *
*   HLE (High Level Emulation) - helper routine.                               *
*                                                                              *
\******************************************************************************/

#ifdef HLE

static tr_u32 segoffset2addr(tr_u32 so)
{
        return( rdp_reg.segment[(so>>24)&0x0f] + (so&0x00ffffff) );

} /* static tr_u32 segoffset2addr(tr_u32 so) */

#endif /* #ifdef HLE */














/******************************************************************************\
*                                                                              *
*   RSP (Reality Signal Processor) HLE - static routines (graphics and audio)  *
*                                                                              *
\******************************************************************************/



static void fixme()
{
        puts("FIXME");

        /* CPU_HALT */

} /* static void fixme() */





static void spnoop()
{
        PRINT_RDP_INFO("SPNOOP\n")

} /* static void spnoop() */










/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) emulation.                                 *
*                                                                              *
\******************************************************************************/





static void rdp_noop()
{
        PRINT_RDP_INFO("NOOP\n")

} /* static void rdp_noop() */





static void rdp_texrect()
{
#ifdef HLE
        tr_u32   a = rdp_reg.pc[rdp_reg.pc_i];
        tr_u32   cmd2 = ((tr_u32 *)mem.rd_ram)[(a>>2)+1];
        tr_u32   cmd3 = ((tr_u32 *)mem.rd_ram)[(a>>2)+3];
#else
        tr_u32   cmd2 = ((tr_u32 *)mem.sp_dmem)[(rdp_reg.curr>>2)+2];
        tr_u32   cmd3 = ((tr_u32 *)mem.sp_dmem)[(rdp_reg.curr>>2)+3];
#endif

        tr_u16  lrx     = (rdp_reg.cmd0 & 0x00fff000) >> 14;
        tr_u16  lry     = (rdp_reg.cmd0 & 0x00000fff) >> 2;
        tr_u16  ulx     = (rdp_reg.cmd1 & 0x00fff000) >> 14;
        tr_u16  uly     = (rdp_reg.cmd1 & 0x00000fff) >> 2;
        int     tile    = (rdp_reg.cmd1 & 0x07000000) >> 24;

        float   s       = ((float)(tr_s16)((cmd2 >> 16) & 0xffff)) / 32.0f;     /* s10.5 */
        float   t       = ((float)(tr_s16)((cmd2      ) & 0xffff)) / 32.0f;     /* s10.5 */

        float   dsdx    = ((float)(tr_s16)((cmd3 >> 16) & 0xffff)) / 1024.0f;   /* s5.10 */
        float   dtdy    = ((float)(tr_s16)((cmd3      ) & 0xffff)) / 1024.0f;   /* s5.10 */



        PRINT_RDP_INFO("TEXRECT ")
        PRINT_RDP_MNEMONIC("ulx=%4hu; uly=%4hu; lrx=%4hu; lry=%4hu; tile=%u; s=%+9.4f; t=%+9.4f; dsdx=%+7.4f; dtdy=%+7.4f\n",
               ulx, uly, lrx, lry, tile, s, t, dsdx, dtdy);
/*
        printf("TEXRECT ");
        printf("ulx=%4hu; uly=%4hu; lrx=%4hu; lry=%4hu; tile=%u; s=%+9.4f; t=%+9.4f; dsdx=%+7.4f; dtdy=%+7.4f\n",
               ulx, uly, lrx, lry, tile, s, t, dsdx, dtdy);
*/


    /* gil: in this case i am correct! */
    /* 1cyc mode (0x00000000): w = lrx - ulx     */
    /* 2cyc mode (0x00100000): w = lrx - ulx     */
    /* copy mode (0x00200000): w = lrx - ulx + 1 */
    /* fill mode (0x00300000): not allowed       */
        if(rdp_reg.mode_h & 0x00200000)
        {
            /* we are in copy mode */
            /* 4 texels are drawn per cycle */
                dsdx /= 4.0;
        }
        else
        {
            /* we are in 1 or 2 cycle mode (fill mode is not allowed)! */
            /* we must not draw the bottom and right edges */
                lrx--;
                lry--;
        }

    /*
        DrawVisualTexRectangle(
                ulx, uly,
                lrx, lry,
                tile,
                s, t, dsdx, dtdy
                );
    */
    /*    tr_ren_rect(???);
*/
} /* static void rdp_texrect() */





static void rdp_texrectflip()
{
        PRINT_RDP_NOT_IMPLEMENTED("TEXRECTFLIP")

        CPU_HALT

} /* static void rdp_texrectflip() */





static void rdp_loadsync()
{
        PRINT_RDP_INFO("LOADSYNC\n")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_loadsync() */





static void rdp_pipesync()
{
        PRINT_RDP_INFO("PIPESYNC\n")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_pipesync() */





static void rdp_tilesync()
{
        PRINT_RDP_INFO("TILESYNC\n")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_tilesync() */





static void rdp_fullsync()
{
        PRINT_RDP_INFO("FULLSYNC\n")

        reg.do_or_check_sthg = RS4300I_DO_DP_INTERRUPT;
        /* ((tr_u32 *)mem.dpc_reg)[2] = 0; */

        /* CPU_HALT */

} /* static void rdp_fullsync() */





static void rdp_setkeygb()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETKEYGB")

        CPU_HALT

} /* static void rdp_setkeygb() */





static void rdp_setkeyr()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETKEYR")

        CPU_HALT

} /* static void rdp_setkeyr() */





static void rdp_setconvert()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETCONVERT")

        CPU_HALT

} /* static void rdp_setconvert() */





static void rdp_setscissor()
{
        static char *interlace_modes[] = { "no", "?", "even", "odd" };

        rdp_reg.scissor.ulx             = (rdp_reg.cmd0 & 0x00fff000) >> 14;
        rdp_reg.scissor.uly             = (rdp_reg.cmd0 & 0x00000fff) >> 2;
        rdp_reg.scissor.interlace_mode  = (rdp_reg.cmd1 & 0x03000000) >> 24;
        rdp_reg.scissor.lrx             = (rdp_reg.cmd1 & 0x00fff000) >> 14;
        rdp_reg.scissor.lry             = (rdp_reg.cmd1 & 0x00000fff) >> 2;

    /* It seems that most demo coders thought that SETSCISSOR uses:
          x, y, width, height
       arguments. That is NOT correct. SETSCISSOR uses:
          ulx, uly, lrx, lry
       args!!! */

        PRINT_RDP_INFO("SETSCISSOR ")
        PRINT_RDP_MNEMONIC("ulx=%lu; uly=%lu; lrx=%lu; lry=%lu; interlace_mode=%s\n",
                rdp_reg.scissor.ulx,
                rdp_reg.scissor.uly,
                rdp_reg.scissor.lrx,
                rdp_reg.scissor.lry,
                interlace_modes[rdp_reg.scissor.interlace_mode]
                );
        PRINT_RDP_WARNING("interlace_mode is ignored")

#ifdef DISPLAY
        if(prefs.display != NO_DISPLAY)
        {
                SetVisualClipRectangle(rdp_reg.scissor.ulx, rdp_reg.scissor.uly, rdp_reg.scissor.lrx, rdp_reg.scissor.lry);
        }
#endif

} /* static void rdp_setscissor() */





static void rdp_setprimdepth()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETPRIMDEPTH")

        CPU_HALT

} /* static void rdp_setprimdepth() */





static void rdp_setothermode()
{
        PRINT_RDP_INFO("SETOTHERMODE\n")
        PRINT_RDP_WARNING("this should clear/init SETOTHERMODE_[HL] - is not done")
        fputs("************** SETOTHERMODE ", stdout);
        puts("this should clear/init SETOTHERMODE_[HL] - is not done");

} /* static void rdp_setothermode() */





static void rdp_loadtlut()
{
        tr_u32   tile   = (rdp_reg.cmd1 >> 24) & 0x07;
        tr_u16  count  = (rdp_reg.cmd1 >> 14) & 0x03ff;
/*
        int     i;
*/


        PRINT_RDP_INFO("LOADTLUT ")
        PRINT_RDP_MNEMONIC("tile=%ld (->tmem=%lx), count=%hx\n", 
               tile,
               rdp_reg.td[tile].tmem,
               count
               );

/*
        for(i = rdp_reg.td[tile].tmem; i < (rdp_reg.td[tile].tmem+count); i++)
                ((tr_u32 *)mem.tmem)[i] = ((tr_u32 *)mem.rd_ram)[i];
*/

        rdp_reg.tlut_8_fmt  = rdp_reg.textureimg_fmt;
        rdp_reg.tlut_8_size = rdp_reg.textureimg_size;
        rdp_reg.tlut_8_addr = rdp_reg.textureimg_addr;

        rdp_reg.tlut_4_fmt[rdp_reg.td[tile].palette]  = rdp_reg.textureimg_fmt; 
        rdp_reg.tlut_4_size[rdp_reg.td[tile].palette] = rdp_reg.textureimg_size;
        rdp_reg.tlut_4_addr[rdp_reg.td[tile].palette] = rdp_reg.textureimg_addr;

} /* static void rdp_loadtlut() */





static void rdp_settilesize()
{
        tr_u32 tile = (rdp_reg.cmd1 >> 24) & 0x07;
        t_td  *tmp_tile = &(rdp_reg.td[tile]);

    /* i think this is 10.2 format */
        tmp_tile->sts_uls = ((float)((rdp_reg.cmd0 >> 12) & 0x0fff)) / 4.0f;
        tmp_tile->sts_ult = ((float)( rdp_reg.cmd0        & 0x0fff)) / 4.0f;
        tmp_tile->sts_lrs = ((float)((rdp_reg.cmd1 >> 12) & 0x0fff)) / 4.0f;
        tmp_tile->sts_lrt = ((float)( rdp_reg.cmd1        & 0x0fff)) / 4.0f;

        PRINT_RDP_INFO("SETTILESIZE ")
        PRINT_RDP_MNEMONIC("uls=%6.2f; ult=%6.2f; lrs=%6.2f; lrt=%6.2f; tile=%ld\n",
                tmp_tile->sts_uls,
                tmp_tile->sts_ult,
                tmp_tile->sts_lrs,
                tmp_tile->sts_lrt,
                tile
                );

/*
        printf("SETTILESIZE ");
        printf("uls=%6.2f; ult=%6.2f; lrs=%6.2f; lrt=%6.2f; tile=%ld\n",
                tmp_tile->sts_uls,
                tmp_tile->sts_ult,
                tmp_tile->sts_lrs,
                tmp_tile->sts_lrt,
                tile
                );
*/

} /* static void rdp_settilesize() */





static void rdp_loadblock()
{
        tr_u32  tile      = (rdp_reg.cmd1 >> 24) & 0x07;
        tr_u16 dxt       = rdp_reg.cmd1 & 0x0fff;   /* 1.11 fixed point format */

        t_td   *tmp_tile = &(rdp_reg.td[tile]);



    /** 
     * We do not check if there is no TEXRECT or TRI1/2 command after
     * a LOADBLOCK/-TILE.
     * That means that the LOADBLOCK/-TILE command before has written
     * its values to the tile descriptor but the tile was not loaded or
     * cached.
     * Normally this should not happen (2 or more LOADBLOCKs/-TILEs
     * after each other make no sense).
    **/

        tmp_tile->set_by = RDPTD_LOADBLOCK;
        tmp_tile->addr = rdp_reg.textureimg_addr;
        rdp_reg.loadtile = tile;

        tmp_tile->uls = (rdp_reg.cmd0 >> 12) & 0x0fff;
        tmp_tile->ult = (rdp_reg.cmd0      ) & 0x0fff;
        tmp_tile->lrs = (rdp_reg.cmd1 >> 12) & 0x0fff;

        /* we calcutate the real dxt value here (real value is in 64bit format) */
        dxt = (tr_u16)(((float)1.0) / (((float)(dxt) / ((float)(2048.0)))));
        tmp_tile->dxt = dxt;

        PRINT_RDP_INFO("LOADBLOCK ")
        PRINT_RDP_MNEMONIC("uls=%hu; ult=%hu; lrs=%hu; (real line width)dxt=%hu(*64bit); tile=%lu (->tmem=%hx)\n",
                tmp_tile->uls,
                tmp_tile->ult,
                tmp_tile->lrs,
                tmp_tile->dxt,
                tile,
                tmp_tile->tmem
                );
/*
        printf("LOADBLOCK ");
        printf("uls=%hu; ult=%hu; lrs=%hu; (real line width)dxt=%hu(*64bit); tile=%lu (->tmem=%hx)\n",
                tmp_tile->uls,
                tmp_tile->ult,
                tmp_tile->lrs,
                tmp_tile->dxt,
                tile,
                tmp_tile->tmem
                );
*/

} /* static void rdp_loadblock() */





static void rdp_loadtile()
{
        tr_u32  tile      = (rdp_reg.cmd1 >> 24) & 0x07;

        t_td   *tmp_tile = &(rdp_reg.td[tile]);



    /** 
     * We do not check if there is no TEXRECT or TRI1/2 command after
     * a LOADBLOCK/-TILE.
     * That means that the LOADBLOCK/-TILE command before has written
     * its values to the tile descriptor but the tile was not loaded or
     * cached.
     * Normally this should not happen (2 or more LOADBLOCKs/-TILEs
     * after each other make no sense).
    **/

        tmp_tile->set_by = RDPTD_LOADTILE;
        tmp_tile->addr = rdp_reg.textureimg_addr;
        rdp_reg.loadtile = tile;

        tmp_tile->uls = (rdp_reg.cmd0 >> 12) & 0x0fff;
        tmp_tile->ult = (rdp_reg.cmd0      ) & 0x0fff;
        tmp_tile->lrs = (rdp_reg.cmd1 >> 12) & 0x0fff;
        tmp_tile->lrt = (rdp_reg.cmd1      ) & 0x0fff;

        PRINT_RDP_INFO("LOADTILE ")
        PRINT_RDP_MNEMONIC("uls=%hu; ult=%hu; lrs=%hu; lrt=%hu; tile=%lu (->tmem=%hx)\n",
                tmp_tile->uls,
                tmp_tile->ult,
                tmp_tile->lrs,
                tmp_tile->lrt,
                tile,
                tmp_tile->tmem
                );

} /* static void rdp_loadtile() */





static void rdp_settile()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };
        static char *cm[]       = { "NOMIRROR/WARP(NOCLAMP)", "MIRROR", "CLAMP", "MIRROR&CLAMP" };

        tr_u32 tile = (rdp_reg.cmd1 >> 24) & 0x07;
        t_td  *tmp_tile = &(rdp_reg.td[tile]);



        tmp_tile->format  = (rdp_reg.cmd0 >> 21) & 0x07;
        tmp_tile->size    = (rdp_reg.cmd0 >> 19) & 0x03;   /* thanx Paul! */
        tmp_tile->line    = (rdp_reg.cmd0 >>  9) & 0x01ff;
        tmp_tile->tmem    = (rdp_reg.cmd0        & 0x01ff) << 3;   /* *8! */
        tmp_tile->palette = (rdp_reg.cmd1 >> 20) & 0x0f;
        tmp_tile->clampt  = (rdp_reg.cmd1 >> 19) & 0x01;
        tmp_tile->mirrort = (rdp_reg.cmd1 >> 18) & 0x01;
        tmp_tile->maskt   = (rdp_reg.cmd1 >> 14) & 0x0f;
        tmp_tile->shiftt  = (rdp_reg.cmd1 >> 10) & 0x0f;
        tmp_tile->clamps  = (rdp_reg.cmd1 >>  9) & 0x01;
        tmp_tile->mirrors = (rdp_reg.cmd1 >>  8) & 0x01;
        tmp_tile->masks   = (rdp_reg.cmd1 >>  4) & 0x0f;
        tmp_tile->shifts  =  rdp_reg.cmd1        & 0x0f;

        /* setup line from "64bit pixel" to correct 32, 16, 8 or 4bit pixel */
        switch(tmp_tile->size)
        {
                case 0:   /* 4bit */
                        tmp_tile->line <<= 4;
                        break;
                case 1:   /* 8bit */
                        tmp_tile->line <<= 3;
                        break;
                case 2:   /* 16bit */
                        tmp_tile->line <<= 2;
                        break;
                case 3:   /* 32bit */
                        tmp_tile->line <<= 1;
                        break;
        }

        PRINT_RDP_INFO("SETTILE ")
        PRINT_RDP_MNEMONIC(
                "format=%s; size=%s; line=%lu (pixel); tmem=%lu; tile=%lu; palette=$%lx; "
                "cmt=%s; maskt=%lu; shiftt=%lu; cms=%s; masks=%lu; shifts=%lu\n",
                format[tmp_tile->format],
                size[tmp_tile->size],
                tmp_tile->line,
                tmp_tile->tmem,
                tile,
                tmp_tile->palette,
                cm[(tmp_tile->clampt << 1) | tmp_tile->mirrort],
                tmp_tile->maskt,
                tmp_tile->shiftt,
                cm[(tmp_tile->clamps << 1) | tmp_tile->mirrors],
                tmp_tile->masks,
                tmp_tile->shifts);

/*
        printf("SETTILE ");
        printf(
                "format=%s; size=%s; line=%lu (pixel); tmem=%lu; tile=%lu; palette=$%lx; "
                "cmt=%s; maskt=%lu; shiftt=%lu; cms=%s; masks=%lu; shifts=%lu\n",
                format[tmp_tile->format],
                size[tmp_tile->size],
                tmp_tile->line,
                tmp_tile->tmem,
                tile,
                tmp_tile->palette,
                cm[(tmp_tile->clampt << 1) | tmp_tile->mirrort],
                tmp_tile->maskt,
                tmp_tile->shiftt,
                cm[(tmp_tile->clamps << 1) | tmp_tile->mirrors],
                tmp_tile->masks,
                tmp_tile->shifts);
*/

} /* static void rdp_settile() */





static void rdp_fillrect()
{
        tr_u32   ulx, uly, lrx, lry;
#ifdef HLE
        tr_u32   color;
#endif



        lrx = (rdp_reg.cmd0 & 0x00ffc000) >> 14;
        lry = (rdp_reg.cmd0 & 0x00000ffc) >> 2;
        ulx = (rdp_reg.cmd1 & 0x00ffc000) >> 14;
        uly = (rdp_reg.cmd1 & 0x00000ffc) >> 2;


        
        PRINT_RDP_INFO("FILLRECT ")
        PRINT_RDP_MNEMONIC("ulx=%lu; uly=%lu; lrx=%lu; lry=%lu\n",
                           ulx,
                           uly,
                           lrx,
                           lry);
        PRINT_RDP_WARNING("modes are ignored")


    /* modes: copy mode is not used (not allowed?) */
        if(!(rdp_reg.mode_h & 0x00200000))
        {
            /* if we are not in 1 or 2 cycle mode (not in copy or fill mode) */
            /* we must not draw the bottom and right edges                   */
                lrx--;
                lry--;
                color = rdp_reg.blendcolor;
        }
        else
        {
                color = rdp_reg.fillcolor;
        }

#ifdef DISPLAY
        if(prefs.display != NO_DISPLAY)
        {
#ifdef JUST3D
                SetVisualColor(0xffffffff);
                DrawVisualRectangle(ulx, uly, lrx, lry);
#else
                SetVisualColor(rdp_reg.fillcolor);
                DrawVisualFillRect(ulx, uly, lrx, lry);
#endif
        }
#endif

} /* static void rdp_fillrect() */





static void rdp_setfillcolor()
{
        PRINT_RDP_INFO("SETFILLCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.fillcolor = rdp_reg.cmd1;

} /* static void rdp_setfillcolor() */





static void rdp_setfogcolor()
{
        PRINT_RDP_INFO("SETFOGCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.fogcolor = rdp_reg.cmd1;

} /* static void rdp_setfogcolor() */





static void rdp_setblendcolor()
{
        PRINT_RDP_INFO("SETBLENDCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.blendcolor = rdp_reg.cmd1;

} /* static void rdp_setblendcolor() */





static void rdp_setprimcolor()
{
        PRINT_RDP_INFO("SETPRIMCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.lod_min   = ((float)((rdp_reg.cmd0 >> 8) & 0xff)) / 256.0f;
        rdp_reg.lod_frac  = ((float)((rdp_reg.cmd0 >> 0) & 0xff)) / 256.0f;
        rdp_reg.primcolor = rdp_reg.cmd1;

} /* static void rdp_setprimcolor() */





static void rdp_setenvcolor()
{
        PRINT_RDP_INFO("SETENVCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.envcolor = rdp_reg.cmd1;

} /* static void rdp_setenvcolor() */





static void rdp_setcombine()
{
        static char *color_combiner[] =
        {
                "combined",     "texel0",       "texel1",          "primitive",
                "shade",        "environment",  "center",          "combined_alpha",
                "texel0_alpha", "texel1_alpha", "primitive_alpha", "shade_alpha", 
                "env_alpha",    "lod_fraction", "prim_lod_frac",   "k5",
                "?",            "?",            "?",               "?",
                "?",            "?",            "?",               "?",
                "?",            "?",            "?",               "?",
                "?",            "?",            "?",               "0"
        };

        static char *alpha_combiner[] =
        {
                "combined", "texel0",      "texel1" ,       "primitive",
                "shade",    "environment", "prim_lod_frac", "0"
        };



        rdp_reg.mode_cycle1_c1 = (rdp_reg.cmd0 >> 20) & 0x0f;
        rdp_reg.mode_cycle1_c2 = (rdp_reg.cmd1 >> 28) & 0x0f;
        rdp_reg.mode_cycle1_c3 = (rdp_reg.cmd0 >> 15) & 0x1f;
        rdp_reg.mode_cycle1_c4 = (rdp_reg.cmd1 >> 15) & 0x07;
        rdp_reg.mode_cycle1_a1 = (rdp_reg.cmd0 >> 12) & 0x07;
        rdp_reg.mode_cycle1_a2 = (rdp_reg.cmd1 >> 12) & 0x07;
        rdp_reg.mode_cycle1_a3 = (rdp_reg.cmd0 >>  9) & 0x07;
        rdp_reg.mode_cycle1_a4 = (rdp_reg.cmd1 >>  9) & 0x07;

        rdp_reg.mode_cycle2_c1 = (rdp_reg.cmd0 >>  5) & 0x0f;
        rdp_reg.mode_cycle2_c2 = (rdp_reg.cmd1 >> 24) & 0x0f;
        rdp_reg.mode_cycle2_c3 = (rdp_reg.cmd0 >>  0) & 0x1f;
        rdp_reg.mode_cycle2_c4 = (rdp_reg.cmd1 >>  6) & 0x07;
        rdp_reg.mode_cycle2_a1 = (rdp_reg.cmd0 >> 21) & 0x07;
        rdp_reg.mode_cycle2_a2 = (rdp_reg.cmd1 >>  3) & 0x07;
        rdp_reg.mode_cycle2_a3 = (rdp_reg.cmd0 >> 18) & 0x07;
        rdp_reg.mode_cycle2_a4 = (rdp_reg.cmd1 >>  0) & 0x07;

        PRINT_RDP_INFO("SETCOMBINE ")
        PRINT_RDP_MNEMONIC("mode0=$%06lx; mode1=$%08lx\n",
               (rdp_reg.cmd0 & 0x00ffffff),
               (rdp_reg.cmd1 & 0xffffffff));
        PRINT_RDP_MNEMONIC("    cycle1: %s %s %s %s %s %s %s %s\n",
                color_combiner[rdp_reg.mode_cycle1_c1],
                color_combiner[rdp_reg.mode_cycle1_c2],
                color_combiner[rdp_reg.mode_cycle1_c3],
                color_combiner[rdp_reg.mode_cycle1_c4],
                alpha_combiner[rdp_reg.mode_cycle1_a1],
                alpha_combiner[rdp_reg.mode_cycle1_a2],
                alpha_combiner[rdp_reg.mode_cycle1_a3],
                alpha_combiner[rdp_reg.mode_cycle1_a4]
                );
        PRINT_RDP_MNEMONIC("    cycle2: %s %s %s %s %s %s %s %s\n",
                color_combiner[rdp_reg.mode_cycle2_c1],
                color_combiner[rdp_reg.mode_cycle2_c2],
                color_combiner[rdp_reg.mode_cycle2_c3],
                color_combiner[rdp_reg.mode_cycle2_c4],
                alpha_combiner[rdp_reg.mode_cycle2_a1],
                alpha_combiner[rdp_reg.mode_cycle2_a2],
                alpha_combiner[rdp_reg.mode_cycle2_a3],
                alpha_combiner[rdp_reg.mode_cycle2_a4]
                );

        printf("SETCOMBINE ");
        printf("mode0=$%06"PR32"x; mode1=$%08"PR32"x\n",
               (rdp_reg.cmd0 & 0x00ffffff),
               (rdp_reg.cmd1 & 0xffffffff));
        printf("    cycle1: %s %s %s %s %s %s %s %s\n",
                color_combiner[rdp_reg.mode_cycle1_c1],
                color_combiner[rdp_reg.mode_cycle1_c2],
                color_combiner[rdp_reg.mode_cycle1_c3],
                color_combiner[rdp_reg.mode_cycle1_c4],
                alpha_combiner[rdp_reg.mode_cycle1_a1],
                alpha_combiner[rdp_reg.mode_cycle1_a2],
                alpha_combiner[rdp_reg.mode_cycle1_a3],
                alpha_combiner[rdp_reg.mode_cycle1_a4]
                );
        printf("    cycle2: %s %s %s %s %s %s %s %s\n",
                color_combiner[rdp_reg.mode_cycle2_c1],
                color_combiner[rdp_reg.mode_cycle2_c2],
                color_combiner[rdp_reg.mode_cycle2_c3],
                color_combiner[rdp_reg.mode_cycle2_c4],
                alpha_combiner[rdp_reg.mode_cycle2_a1],
                alpha_combiner[rdp_reg.mode_cycle2_a2],
                alpha_combiner[rdp_reg.mode_cycle2_a3],
                alpha_combiner[rdp_reg.mode_cycle2_a4]
                );

    /* this is a hack to set up texturing */
#if 0
        if((rdp_reg.mode_cycle1_c1 == 0x03) || rdp_reg.mode_cycle1_c1 == 0x04 || rdp_reg.mode_cycle1_c1 == 0x05)
                rdp_reg.do_textures = 0;
        else if((rdp_reg.mode_cycle1_c2 == 0x03) || rdp_reg.mode_cycle1_c2 == 0x04 || rdp_reg.mode_cycle1_c2 == 0x05)
                rdp_reg.do_textures = 0;
        else if((rdp_reg.mode_cycle1_c3 == 0x03) || rdp_reg.mode_cycle1_c3 == 0x04 || rdp_reg.mode_cycle1_c3 == 0x05)
                rdp_reg.do_textures = 0;
        else if((rdp_reg.mode_cycle1_c4 == 0x03) || rdp_reg.mode_cycle1_c4 == 0x04 || rdp_reg.mode_cycle1_c4 == 0x05)
                rdp_reg.do_textures = 0;
        else
                rdp_reg.do_textures = 1;
#endif

        if(rdp_reg.mode_cycle1_c1 == 0x01)
                rdp_reg.do_textures = 1;
        else if(rdp_reg.mode_cycle1_c2 == 0x01)
                rdp_reg.do_textures = 1;
        else if(rdp_reg.mode_cycle1_c3 == 0x01)
                rdp_reg.do_textures = 1;
        else if(rdp_reg.mode_cycle1_c4 == 0x01)
                rdp_reg.do_textures = 1;
        else
                rdp_reg.do_textures = 0;

} /* static void rdp_setcombine() */





static void rdp_settextureimage()
{
        static char *format[]           = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]             = { "4bit", "8bit", "16bit", "32bit" };


        rdp_reg.textureimg_fmt          = (rdp_reg.cmd0 >> 21) & 0x07;
        rdp_reg.textureimg_size         = (rdp_reg.cmd0 >> 19) & 0x03;
        rdp_reg.textureimg_width        = 1 + (rdp_reg.cmd0 & 0x00000fff);
#ifndef HLE
        rdp_reg.textureimg_addr         = rdp_reg.cmd1;
#else
        rdp_reg.textureimg_addr         = segoffset2addr(rdp_reg.cmd1) & 0x003fffff;
        rdp_reg.textureimg_loadaddr     = rdp_reg.cmd1;
        rdp_reg.textureimg_id           = 0;
#endif

        PRINT_RDP_INFO("SETTEXTUREIMAGE ")
        PRINT_RDP_MNEMONIC("format=%s; size=%s; width=%hu; addr=$%08lx\n",
               format[rdp_reg.textureimg_fmt],
               size[rdp_reg.textureimg_size],
               rdp_reg.textureimg_width,
               rdp_reg.textureimg_addr);
/*
        printf("SETTEXTUREIMAGE ");
        printf("format=%s; size=%s; width=%hu; addr=$%08lx\n",
               format[rdp_reg.textureimg_fmt],
               size[rdp_reg.textureimg_size],
               rdp_reg.textureimg_width,
               rdp_reg.textureimg_addr);
*/

} /* static void rdp_settextureimage() */





static void rdp_setdepthimage()
{
        PRINT_RDP_INFO("SETDEPTHIMAGE ")
        PRINT_RDP_MNEMONIC("addr=$%lx\n", rdp_reg.cmd1);

#ifndef HLE
        rdp_reg.depthimg_addr = rdp_reg.cmd1;
#else
        rdp_reg.depthimg_addr = segoffset2addr(rdp_reg.cmd1);
#endif

} /* static void rdp_setdepthimage() */





static void rdp_setcolorimage()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };



        rdp_reg.colorimg_fmt   = (rdp_reg.cmd0 & 0x00e00000) >> 21;
        rdp_reg.colorimg_size  = (rdp_reg.cmd0 & 0x00180000) >> 19;
        rdp_reg.colorimg_width = 1 + (rdp_reg.cmd0 & 0x00000fff);
#ifndef HLE
        rdp_reg.colorimg_addr  = rdp_reg.cmd1;
#else
        rdp_reg.colorimg_addr = segoffset2addr(rdp_reg.cmd1) & 0x003fffff;
#endif

        PRINT_RDP_INFO("SETCOLORIMAGE ")
        PRINT_RDP_MNEMONIC("format=%s; size=%s; width=%hu; addr=$%lx\n",
                           format[rdp_reg.colorimg_fmt],
                           size[rdp_reg.colorimg_size],
                           rdp_reg.colorimg_width,
                           rdp_reg.colorimg_addr);

#ifdef DISPLAY
        if(prefs.display != NO_DISPLAY)
        {
        /* printf("%x\n", rdp_reg.colorimg_addr); */
                SetVisualRenderBuffer(
                        rdp_reg.colorimg_fmt, rdp_reg.colorimg_size,
                        rdp_reg.colorimg_width, rdp_reg.colorimg_addr + (tr_u8 *)mem.rd_ram);
        }
#endif

        /* CPU_HALT */

} /* static void rdp_setcolorimage() */





static void rdp_trifill()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRIFILL")

        CPU_HALT

} /* static void rdp_trifill() */





static void rdp_trishade()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADE")

        CPU_HALT

} /* static void rdp_trishade() */





static void rdp_tritxtr()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRITXTR")

        CPU_HALT

} /* static void rdp_tritxtr() */





static void rdp_trishadetxtr()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADETXTR")

        CPU_HALT

} /* static void rdp_trishadetxtr() */





static void rdp_trifillz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRIFILLZ")

        CPU_HALT

} /* static void rdp_trifillz() */





static void rdp_trishadez()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADEZ")

        CPU_HALT

} /* static void rdp_trishadez() */





static void rdp_tritxtrz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRITXTRZ")

        CPU_HALT

} /* static void rdp_tritxtrz() */





static void rdp_trishadetxtrz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADETXTRZ")

        CPU_HALT

} /* static void rdp_trishadetxtrz() */










