/**
 * TrueReality - N64/rdp_registers.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/* every 'export' in this file resides in 'rdp.c' */
/* this file could also be named as 'rdp_export.h' */





#ifndef __TR_RDP_REGISTERS_H
#define __TR_RDP_REGISTERS_H





#include "type_sizes.h"






extern struct rdp_reg rdp_reg;





typedef struct
{
        float   x, y, z;        /* s15 */
        tr_u16  flags;
        float   s, t;           /* s10.5 */
        tr_u8   r, g, b, a;
}
t_vtx;

typedef struct
{
    /* SETTILE */
        tr_u8   format;
        tr_u8   size;
        tr_u16  line;
        tr_u16  tmem;
        tr_u8   palette;
        tr_u8    clampt;
        tr_u8    mirrort;
        tr_u8   maskt;
        tr_u8   shiftt;
        tr_u8    clamps;
        tr_u8    mirrors;
        tr_u8   masks;
        tr_u8   shifts;

    /* LOADBLOCK / LOADTILE */
#define RDPTD_LOAD_NOT_SET -1
#define RDPTD_LOADBLOCK 1
#define RDPTD_LOADTILE 2
        int     set_by;
        tr_u32   addr;
        tr_u16  uls; /* Upper Left Texture X-Coordinate (10.2) */
        tr_u16  ult; /* Upper Left Texture Y-Coordinate (10.2) */
        tr_u16  lrs; /* Lower Right Texture X-Coordinate (10.2) */
        tr_u16  lrt; /* Lower Right Texture Y-Coordinate (10.2) */
        tr_u16  dxt; /* Increment dxt per scanline */

    /* SETTILESIZE */
        float   sts_uls; /* Upper Left Texture X-Coordinate (10.2) */
        float   sts_ult; /* Upper Left Texture Y-Coordinate (10.2) */
        float   sts_lrs; /* Lower Right Texture X-Coordinate (10.2) */
        float   sts_lrt; /* Lower Right Texture Y-Coordinate (10.2) */
}
t_td;

struct rdp_reg
{
        /* general */

        tr_u32    cmd0, cmd1;
        tr_u32    start, stop, curr;

        t_td    td[8];          /* 8 possible tiles */
        int     tile;           /* the actual tile (set by SETTILE) */
        int     loadtile;       /* the load tile (set by LOADBLOCK/-TILE) */


        /* color register */
        tr_u32  fillcolor;     /* memory interface: 32bit */
        tr_u32  primcolor;     /* color combiner: rgba */
        float  lod_min;
        float  lod_frac;
        tr_u32  envcolor;      /* color combiner: rgba */
        tr_u32  fogcolor;      /* blender: rgba */
        tr_u32  blendcolor;    /* blender: rgba */

        /* This is the frame buffer location to draw into */
        /* Used by rdp_setcolorimage() */
        tr_u8  colorimg_fmt;          /* Frame Format (3 bits) */
        tr_u8  colorimg_size;         /* Bits per pixel (2 bits) */
        tr_u16 colorimg_width;        /* width of frame in pixels (12 bits) */
        tr_u32  colorimg_addr;         /* RDRAM pointer to Image (32 bits) */

        /* This is the frame buffer location to draw into */
        /* Used by rdp_settextureimage() */
        tr_u8  textureimg_fmt;        /* see above */
        tr_u8  textureimg_size;
        tr_u16 textureimg_width;
        tr_u32  textureimg_addr;
        tr_u32  textureimg_loadaddr;
        tr_u32  textureimg_id;

        /* Used by rdp_setdepthimage() */
        tr_u32  depthimg_addr;

        /* This is the location of an 8 bit TLUT */
        /* Used by rdp_loadtlut() */
        tr_u8  tlut_8_fmt;   /* TLUT Format (3 bits) */
        tr_u8  tlut_8_size;  /* Bits per pixel (2 bits) */
        tr_u32  tlut_8_addr;  /* RDRAM pointer to TLUT (32 bits) */

        /* This is the location of an 4 bit TLUT */
        /* Used by rdp_loadtlut() */
        tr_u8  tlut_4_fmt[16];   /* TLUT Format (3 bits) */
        tr_u8  tlut_4_size[16];  /* Bits per pixel (2 bits) */
        tr_u32  tlut_4_addr[16];  /* RDRAM pointer to TLUT (32 bits) */

        struct
        {
                tr_u32    ulx;
                tr_u32    uly;
                tr_u32    lrx;
                tr_u32    lry;
                tr_u32    interlace_mode;
        } scissor;


#ifdef HLE

        /**
        *
        * This might be part of some sort of High Level Emulation (HLE).
        * HLE was first hacked out by Epsilon and RealityMan - the authors of UltraHLE.
        *
        **/

        tr_u32 pc[10];                            /* DL stack is 10 */
        tr_u32 pc_i;                              /* PC index (current DL depth) */

#define INVALID_TASK    -1
#define GFX_TASK        1
#define AUDIO_TASK      2

        int task;
        int ucode_nr;
        int size;   /* length of task */



        struct
        {
                tr_u32  mipmap_level;
                tr_u32  on;               /* 1: on 0: off */
                float scale_t;
                float scale_s;
        }
        texture[8];

        struct
        {
                float   r;
                float   g;
                float   b;
                float   a;
                float   r_copy;
                float   g_copy;
                float   b_copy;
                float   a_copy;
                float   x;
                float   y;
                float   z;
                float   w;
        }
        light[8], lookat_x, lookat_y;

        tr_u32 geometrymode;
        tr_u32 mode_h;
        tr_u32 mode_l;
        tr_u32 mode_cycle1_c1;
        tr_u32 mode_cycle1_c2;
        tr_u32 mode_cycle1_c3;
        tr_u32 mode_cycle1_c4;
        tr_u32 mode_cycle1_a1;
        tr_u32 mode_cycle1_a2;
        tr_u32 mode_cycle1_a3;
        tr_u32 mode_cycle1_a4;
        tr_u32 mode_cycle2_c1;
        tr_u32 mode_cycle2_c2;
        tr_u32 mode_cycle2_c3;
        tr_u32 mode_cycle2_c4;
        tr_u32 mode_cycle2_a1;
        tr_u32 mode_cycle2_a2;
        tr_u32 mode_cycle2_a3;
        tr_u32 mode_cycle2_a4;
        tr_u32 do_textures;
        tr_u32 lights;
        tr_u32 segment[16];                      /* segment pointer */

        float vp[8];                            /* viewport (scale, translate) */

        /*
        float perspective_normal;
        */

        t_vtx vtx[64];                          /* 32 vertices */

        struct
        {
                tr_s32   nx;
                tr_s32   ny;
                tr_s32   px;
                tr_s32   py;
        }
        clip;                                   /* clipping stuff */


#endif /* #ifdef HLE */


        int halt;

#ifdef MACOS
        int graphics_changes;
#endif
    
}; /* struct rsp_reg */





#endif /* #ifndef __TR_RDP_REGISTERS_H */
