/**
 * TrueReality - N64/registers.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/* every 'export' in this file resides in 'cpu.c' */
/* this file could also be named as 'cpu_export.h' */





#ifndef __TR_REGISTERS_H
#define __TR_REGISTERS_H





#include "type_sizes.h"





struct rs4300i_tlb
{
    /* these regs store the cop0 register */
        tr_u32   hh;   /* PageMask (5) */
        tr_u32   hl;   /* EntryHi (10) */
        tr_u32   lh;   /* EntryLo1 (3) */
        tr_u32   ll;   /* EntryLo0 (2) */
    /* the next values are for speed. they are set when
       the above 4 values are set. */
        tr_u8    g;
        tr_u32   mask;
        tr_u32   invmask;
        tr_u32   even_odd_checkbit;
};





/* all "real" values are in cpu.c */

/* rs4300i registers */

struct rs4300i_reg
{
        tr_s64 r[32];        /* General Purpose Registers     (GPR) */
        tr_s64 cpr[3][32];   /* CoProcessor Registers         (CPR) */
        tr_s64 ccr[3][32];   /* Coprocessor Control Registers (CCR) */

        tr_u64 hi, lo;       /* multiply and divide registers */

        tr_u32 pc;           /* program counter */



        /* add emulation reg */
        
        tr_u32 pc_delay;   /* delayed program counter */
        tr_s16 delay_branch;   /* delayed offset for program counter (branch instr) */
        tr_u32 delay_jump;   /* delayed jump for program counter */
        int    delay;      /* this is for correct emulation of the 6 stage pipeline */
                            /* if(delay == NO_DELAY)   { next instr is pc + 4 } */
                            /* if(delay == DO_DELAY)   { next instr is the delayed instruction } */
                            /* if(delay == EXEC_DELAY) { exec instr @ delayed pc - set in switch} */
        
        int    do_or_check_sthg;
                            /* if '!0' then the rs4300i has to do or check sthg */
                            /* look at the defines that follow this struct */

        int    halt;       /* if '!0' then stop execution */
        tr_u32 code;       /* current code (needed for speed) */
        
        tr_u8  llbit;

        struct rs4300i_tlb tlb[32];

}; /* struct rs4300i_reg */



/* rs4300i_reg.delay */

#define NO_DELAY                0
#define DO_DELAY                1
#define EXEC_DELAY              2
#define DO_LOOP                 3
#define DELAY_BRANCH            4
#define DELAY_JUMP              5
#define DELAY_LOOP              6
#define DELAY_BRANCH_EXEC       7
#define DELAY_JUMP_EXEC         8
#define DELAY_LOOP_EXEC         9





/* defines for 'rs4300i_reg.do_or_check_sthg' */
/* (Gil Pedersen) */
/* I've added more possible exceptions, and prioritized them to what I think is right */
/* This work's well with my ppc assembly optimised exception checking code */
/* (Niki W. Waibel) */
/* Okay Gil. I've no problem with this. */

#define RS4300I_DO_OR_CHECK_NOTHING                  0
#define RS4300I_CHECK_CPU_INTERRUPTS                (1 <<  0)
#define RS4300I_CHECK_MI_INTERRUPTS                 (1 <<  1)
#define RS4300I_DO_SI_INTERRUPT                     (1 <<  2)
#define RS4300I_DO_SP_INTERRUPT                     (1 <<  3)
#define RS4300I_DO_DP_INTERRUPT                     (1 <<  4)
#define RS4300I_DO_AI_INTERRUPT                     (1 <<  5)
#define RS4300I_DO_PI_INTERRUPT                     (1 <<  6)
#define RS4300I_DO_BUS_ERROR_ACCESS_EXCEPTION       (1 << 12)
#define RS4300I_DO_WATCH_EXCEPTION                  (1 << 13)
#define RS4300I_DO_TLB_MODIFICATION_EXCEPTION       (1 << 14)
#define RS4300I_DO_TLB_INVALID_EXCEPTION            (1 << 15)
#define RS4300I_DO_TLB_MISS_EXCEPTION               (1 << 16)
#define RS4300I_DO_ADDR_ERROR_ACCESS_EXCEPTION      (1 << 17)
#define RS4300I_DO_FLOATING_POINT_EXCEPTION         (1 << 18)
#define RS4300I_DO_INTEGER_OVERFLOW_EXCEPTION       (1 << 19)
#define RS4300I_DO_TRAP_EXCEPTION                   (1 << 20)
#define RS4300I_DO_RESERVED_INSTRUCTION_EXCEPTION   (1 << 21)
#define RS4300I_DO_COP_UNUSABLE_EXCEPTION           (1 << 22)
#define RS4300I_DO_BREAK_EXCEPTION                  (1 << 23)
#define RS4300I_DO_SYSCALL_EXCEPTION                (1 << 24)
#define RS4300I_DO_BUS_ERROR_FETCH_EXCEPTION        (1 << 25)
#define RS4300I_DO_ADDR_ERROR_FETCH_EXCEPTION       (1 << 28)
#define RS4300I_DO_SOFT_NMI_EXCEPTION               (1 << 29)
#define RS4300I_DO_SOFT_RESET_EXCEPTION             (1 << 30)
#define RS4300I_DO_RESET_EXCEPTION                  (1 << 31)





extern struct rs4300i_reg reg;
extern tr_instruction *tr_i;





#define INDEX           0
#define RANDOM          1
#define ENTRYLO0        2
#define ENTRYLO1        3
#define CONTEXT         4
#define PAGEMASK        5
#define WIRED           6
#define BADVADDR        8
#define COUNT           9
#define ENTRYHI         10
#define COMPARE         11
#define STATUS          12
#define CAUSE           13
#define EPC             14
#define PRID            15
#define CONFIG          16
#define LLADDR          17
#define WATCHLO         18
#define WATCHHI         19
#define XCONTEXT        20
#define ECC             26
#define CACHEERR        27
#define TAGLO           28
#define TAGHI           29
#define ERROREPC        30





#endif /* #ifndef __TR_REGISTERS_H */
