/**
 * Copyright (C) 1998 by Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 * Version information:
 *
 * 1999032000:
 *      Minor changes in JAL and JALR (the bit indicating the IMEM was not
 *      properly set).
 *
 * 1999031900:
 *      Changed rsp_special_break():
 *              ((tr_u32 *)mem.sp_reg)[4] |= 0x00000003;
 *      ->
 *              ((tr_u32 *)mem.sp_reg)[4] |= 0x00000203;
 *      Otherwise RSP stops in some cases.
 *      Thanks a lot Jeff!!!
 *
**/





#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <config.h>
#include <parser_extern.h>
#include <memory_extern.h>
#include <type_sizes.h>

#ifdef TR_DEBUG
        #include <debug.h>
#endif

#include "memory.h"
#include "registers.h"
#include "rsp_registers.h"
#include "rsp_mnemonic.h"
#include "rsp.h"





#define FIXED2DOUBLE(x) (((double)x) / 65536.0)
#define DOUBLE2FIXED(x) ((tr_u32)(x * 65536.0 + 0.5))
#define FIXED2FLOAT(x) (((float)x) / 65536.0)
#define FLOAT2FIXED(x) ((tr_u32)(x * 65536.0 + 0.5))





/* rsp registers */

struct rsp_reg rsp_reg;


#ifndef TR_DEBUG
#ifdef TR_MACOS
static char *ELEM[32] =
{
        /* The first 16 elements cannot be accessed because the elements
           are encoded in the cop2 function (rs) field itself. */
           
        "???", "???",  "???",  "???",  "???",  "???",  "???",  "???",
        "???", "???",  "???",  "???",  "???",  "???",  "???",  "???",
         "[]", " ? ", "[0q]", "[1q]", "[0h]", "[1h]", "[2h]", "[3h]",
        "[0]", "[1]",  "[2]",  "[3]",  "[4]",  "[5]",  "[6]",  "[7]"
};
#endif /* TR_MACOS */
#endif /* !TR_DEBUG */

#ifdef TR_DEBUG
static char *ELEM[32] =
{
        /* The first 16 elements cannot be accessed because the elements
           are encoded in the cop2 function (rs) field itself. */
           
        "???", "???",  "???",  "???",  "???",  "???",  "???",  "???",
        "???", "???",  "???",  "???",  "???",  "???",  "???",  "???",
         "[]", " ? ", "[0q]", "[1q]", "[0h]", "[1h]", "[2h]", "[3h]",
        "[0]", "[1]",  "[2]",  "[3]",  "[4]",  "[5]",  "[6]",  "[7]"
};





static char *EL_SP[32] =
{
        /* The first 16 elements are normally not accessed because the elements
           are encoded in the cop2 function (rs) field itself. But if we have a
           instruction which uses 2 element specifiers, such as:
           VMOV v1[2], v3[5]
           then the first 16 elements are used because the 5th bit of the __RD
           fiels seems to be 0! */
           
        "[00]", "[01]", "[02]", "[03]", "[04]", "[05]", "[06]", "[07]",
        "[08]", "[09]", "[10]", "[11]", "[12]", "[13]", "[14]", "[15]",
        "[00]", "[01]", "[02]", "[03]", "[04]", "[05]", "[06]", "[07]",
        "[08]", "[09]", "[10]", "[11]", "[12]", "[13]", "[14]", "[15]"
};





static char *EL_C0[32] =
{
        "sp_mem_addr", "sp_dram_addr",    "sp_rd_len",    "sp_wr_len",
          "sp_status",  "sp_dma_full",  "sp_dma_busy", "sp_semaphore",
          "dpc_start",      "dpc_end",  "dpc_current",   "dpc_status",
          "dpc_clock",  "dpc_bufbusy", "dpc_pipebusy",     "dpc_tmem",
                " ? ",          " ? ",          " ? ",          " ? ",
                " ? ",          " ? ",          " ? ",          " ? ",
                " ? ",          " ? ",          " ? ",          " ? ",
                " ? ",          " ? ",          " ? ",          " ? "
};





static char *EL_C2[32] =
{
        "[00]", "[01]", "[02]", "[03]", "[04]", "[05]", "[06]", "[07]",
        "[08]", "[09]", "[10]", "[11]", "[12]", "[13]", "[14]", "[15]",
         " ? ",  " ? ",  " ? ",  " ? ",  " ? ",  " ? ",  " ? ",  " ? ",
         " ? ",  " ? ",  " ? ",  " ? ",  " ? ",  " ? ",  " ? ",  " ? "
};
#endif /* #ifdef TR_DEBUG */




/* rsp instructions */

static void rsp_special();
static void rsp_regimm();
static void rsp_j();
static void rsp_jal();
static void rsp_beq();
static void rsp_bne();
static void rsp_blez();
static void rsp_bgtz();
static void rsp_addi();
static void rsp_addiu();
static void rsp_slti();
static void rsp_sltiu();
static void rsp_andi();
static void rsp_ori();
static void rsp_xori();
static void rsp_lui();
static void rsp_cop0();
static void rsp_cop2();
static void rsp_lb();
static void rsp_lh();
static void rsp_lw();
static void rsp_lbu();
static void rsp_lhu();
static void rsp_sb();
static void rsp_sh();
static void rsp_sw();
static void rsp_lwc2();
static void rsp_swc2();
static void rsp_reserved();



/* rsp special instructions */

static void rsp_special_sll();
static void rsp_special_srl();
static void rsp_special_sra();
static void rsp_special_sllv();
static void rsp_special_srlv();
static void rsp_special_srav();
static void rsp_special_jr();
static void rsp_special_jalr();
static void rsp_special_break();
static void rsp_special_add();
static void rsp_special_addu();
static void rsp_special_sub();
static void rsp_special_subu();
static void rsp_special_and();
static void rsp_special_or();
static void rsp_special_xor();
static void rsp_special_nor();
static void rsp_special_slt();
static void rsp_special_sltu();
static void rsp_special_reserved();



/* rsp regimm instructions */

static void rsp_regimm_bltz();
static void rsp_regimm_bgez();
static void rsp_regimm_bltzal();
static void rsp_regimm_bgezal();
static void rsp_regimm_reserved();



/* rsp cop0 instructions */

static void rsp_cop0_mf();
static void rsp_cop0_mt();
static void rsp_cop0_reserved();



/* rsp cop2 instructions */

static void rsp_cop2_mf();
static void rsp_cop2_cf();
static void rsp_cop2_mt();
static void rsp_cop2_ct();
static void rsp_cop2_vectop();
static void rsp_cop2_reserved();



/* rsp cop2 vectop instructions */

static void rsp_cop2_vectop_vmulf();
static void rsp_cop2_vectop_vmulu();
static void rsp_cop2_vectop_vrndp();
static void rsp_cop2_vectop_vmulq();
static void rsp_cop2_vectop_vmudl();
static void rsp_cop2_vectop_vmudm();
static void rsp_cop2_vectop_vmudn();
static void rsp_cop2_vectop_vmudh();
static void rsp_cop2_vectop_vmacf();
static void rsp_cop2_vectop_vmacu();
static void rsp_cop2_vectop_vrndn();
static void rsp_cop2_vectop_vmacq();
static void rsp_cop2_vectop_vmadl();
static void rsp_cop2_vectop_vmadm();
static void rsp_cop2_vectop_vmadn();
static void rsp_cop2_vectop_vmadh();
static void rsp_cop2_vectop_vadd();
static void rsp_cop2_vectop_vsub();
static void rsp_cop2_vectop_vsut();
static void rsp_cop2_vectop_vabs();
static void rsp_cop2_vectop_vaddc();
static void rsp_cop2_vectop_vsubc();
static void rsp_cop2_vectop_vaddb();
static void rsp_cop2_vectop_vsubb();
static void rsp_cop2_vectop_vaccb();
static void rsp_cop2_vectop_vsucb();
static void rsp_cop2_vectop_vsad();
static void rsp_cop2_vectop_vsac();
static void rsp_cop2_vectop_vsum();
static void rsp_cop2_vectop_vsaw();
static void rsp_cop2_vectop_vlt();
static void rsp_cop2_vectop_veq();
static void rsp_cop2_vectop_vne();
static void rsp_cop2_vectop_vge();
static void rsp_cop2_vectop_vcl();
static void rsp_cop2_vectop_vch();
static void rsp_cop2_vectop_vcr();
static void rsp_cop2_vectop_vmrg();
static void rsp_cop2_vectop_vand();
static void rsp_cop2_vectop_vnand();
static void rsp_cop2_vectop_vor();
static void rsp_cop2_vectop_vnor();
static void rsp_cop2_vectop_vxor();
static void rsp_cop2_vectop_vnxor();
static void rsp_cop2_vectop_vrcp();
static void rsp_cop2_vectop_vrcpl();
static void rsp_cop2_vectop_vrcph();
static void rsp_cop2_vectop_vmov();
static void rsp_cop2_vectop_vrsq();
static void rsp_cop2_vectop_vrsql();
static void rsp_cop2_vectop_vrsqh();
static void rsp_cop2_vectop_vnoop();
static void rsp_cop2_vectop_vextt();
static void rsp_cop2_vectop_vextq();
static void rsp_cop2_vectop_vextn();
static void rsp_cop2_vectop_vinst();
static void rsp_cop2_vectop_vinsq();
static void rsp_cop2_vectop_vinsn();
static void rsp_cop2_vectop_reserved();



/* rsp instruction function pointer */

static rsp_instr instruction[64] =
{
        rsp_special,
        rsp_regimm,
        rsp_j,
        rsp_jal,
        rsp_beq,
        rsp_bne,
        rsp_blez,
        rsp_bgtz,
        rsp_addi,
        rsp_addiu,
        rsp_slti,
        rsp_sltiu,
        rsp_andi,
        rsp_ori,
        rsp_xori,
        rsp_lui,
        rsp_cop0,
        rsp_reserved,
        rsp_cop2,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_lb,
        rsp_lh,
        rsp_reserved,
        rsp_lw,
        rsp_lbu,
        rsp_lhu,
        rsp_reserved,
        rsp_reserved,
        rsp_sb,
        rsp_sh,
        rsp_reserved,
        rsp_sw,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_lwc2,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_swc2,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved
};





/* rsp special function pointer */

static rsp_instr special_instruction[64] =
{
        rsp_special_sll,
        rsp_special_reserved,
        rsp_special_srl,
        rsp_special_sra,
        rsp_special_sllv,
        rsp_special_reserved,
        rsp_special_srlv,
        rsp_special_srav,
        rsp_special_jr,
        rsp_special_jalr,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_break,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_add,
        rsp_special_addu,
        rsp_special_sub,
        rsp_special_subu,
        rsp_special_and,
        rsp_special_or,
        rsp_special_xor,
        rsp_special_nor,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_slt,
        rsp_special_sltu,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved
};



/* rsp regimm function pointer */

static rsp_instr regimm_instruction[32] =
{
        rsp_regimm_bltz,
        rsp_regimm_bgez,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_bltzal,
        rsp_regimm_bgezal,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved
};



/* rsp cop0 function pointer */

static rsp_instr cop0_instruction[32] =
{
        rsp_cop0_mf,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_mt,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved
};



/* rsp cop2 function pointer */

static rsp_instr cop2_instruction[32] =
{
        rsp_cop2_mf,
        rsp_cop2_reserved,
        rsp_cop2_cf,
        rsp_cop2_reserved,
        rsp_cop2_mt,
        rsp_cop2_reserved,
        rsp_cop2_ct,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop
};



/* rsp cop2 vectop function pointer */

static rsp_instr cop2_vectop_instruction[64] =
{
        rsp_cop2_vectop_vmulf,
        rsp_cop2_vectop_vmulu,
        rsp_cop2_vectop_vrndp,
        rsp_cop2_vectop_vmulq,
        rsp_cop2_vectop_vmudl,
        rsp_cop2_vectop_vmudm,
        rsp_cop2_vectop_vmudn,
        rsp_cop2_vectop_vmudh,
        rsp_cop2_vectop_vmacf,
        rsp_cop2_vectop_vmacu,
        rsp_cop2_vectop_vrndn,
        rsp_cop2_vectop_vmacq,
        rsp_cop2_vectop_vmadl,
        rsp_cop2_vectop_vmadm,
        rsp_cop2_vectop_vmadn,
        rsp_cop2_vectop_vmadh,
        rsp_cop2_vectop_vadd,
        rsp_cop2_vectop_vsub,
        rsp_cop2_vectop_vsut,
        rsp_cop2_vectop_vabs,
        rsp_cop2_vectop_vaddc,
        rsp_cop2_vectop_vsubc,
        rsp_cop2_vectop_vaddb,
        rsp_cop2_vectop_vsubb,
        rsp_cop2_vectop_vaccb,
        rsp_cop2_vectop_vsucb,
        rsp_cop2_vectop_vsad,
        rsp_cop2_vectop_vsac,
        rsp_cop2_vectop_vsum,
        rsp_cop2_vectop_vsaw,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_vlt,
        rsp_cop2_vectop_veq,
        rsp_cop2_vectop_vne,
        rsp_cop2_vectop_vge,
        rsp_cop2_vectop_vcl,
        rsp_cop2_vectop_vch,
        rsp_cop2_vectop_vcr,
        rsp_cop2_vectop_vmrg,
        rsp_cop2_vectop_vand,
        rsp_cop2_vectop_vnand,
        rsp_cop2_vectop_vor,
        rsp_cop2_vectop_vnor,
        rsp_cop2_vectop_vxor,
        rsp_cop2_vectop_vnxor,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_vrcp,
        rsp_cop2_vectop_vrcpl,
        rsp_cop2_vectop_vrcph,
        rsp_cop2_vectop_vmov,
        rsp_cop2_vectop_vrsq,
        rsp_cop2_vectop_vrsql,
        rsp_cop2_vectop_vrsqh,
        rsp_cop2_vectop_vnoop,
        rsp_cop2_vectop_vextt,
        rsp_cop2_vectop_vextq,
        rsp_cop2_vectop_vextn,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_vinst,
        rsp_cop2_vectop_vinsq,
        rsp_cop2_vectop_vinsn,
        rsp_cop2_vectop_reserved
};







void rsp_step()
{
        /* execute instruction */

        (instruction[__OPCODE])();



        /* calculate next pc */

        switch(rsp_reg.delay)
        {
            case EXEC_RSP_DELAY:   /* next instruction is at pc_delay */
                rsp_reg.delay = NO_RSP_DELAY;
                mem.sp_reg_pc = rsp_reg.pc_delay;
                break;

            case DO_RSP_DELAY:   /* a delayed pc was stored into pc_delay      */
                       /* next instr is at pc+4 (case 0 is executed) */
                       /* next instr is at pc_delay                  */
                rsp_reg.delay = EXEC_RSP_DELAY;

            case NO_RSP_DELAY:    /* normal execution */
                mem.sp_reg_pc += 4;

        } /* switch(rsp_reg.delay) */



        /* do common things */

        rsp_reg.r[0] = 0;
        rsp_reg.count++;

} /* void rsp_step() */





void rsp_dis(tr_u32 code)
{
        (instruction[code>>26])();
}





void rsp_print_r()
{
        int i;



        for(i=0; i<16; i++)
        {
                printf("rsp[%2"PR8"u]=%8"PR32"x  ", i,    rsp_reg.r[i]);
                printf("rsp[%2"PR8"u]=%8"PR32"x\n", i+16, rsp_reg.r[i+16]);
        }

        printf("vrcph_source=%04"PR16"x         ", rsp_reg.vrcph_source);
        printf("       vrcph_result=%04"PR16"x\n", rsp_reg.vrcph_result);
        
        printf("flg[0]=%04"PR16"x  ", rsp_reg.flag[0]);
        printf("flg[1]=%04"PR16"x  ", rsp_reg.flag[1]);
        printf("flg[2]=%04"PR16"x  ", rsp_reg.flag[2]);
        printf("flg[3]=%04"PR16"x\n", rsp_reg.flag[3]);

        printf("[pc]=%08"PR32"x\n\n", mem.sp_reg_pc);
        
}





void rsp_print_v()
{
        int i;



        for(i=0; i<32; i++)
        {
                printf("v[%2"PR8"d] = %4"PR16"x %4"PR16"x %4"PR16"x %4"PR16"x %4"PR16"x %4"PR16"x %4"PR16"x %4"PR16"x\n",
                       i,
                       rsp_reg.v[i][0],
                       rsp_reg.v[i][1],
                       rsp_reg.v[i][2],
                       rsp_reg.v[i][3],
                       rsp_reg.v[i][4],
                       rsp_reg.v[i][5],
                       rsp_reg.v[i][6],
                       rsp_reg.v[i][7]
                      );
        }
        putchar('\n');
}





void rsp_reset()
{
        int i, j;



        /* debugging */

        rsp_reg.halt = 1;
        rsp_reg.count = 0;



        /* gpr registers */

        for(i=0; i<32; i++)
                rsp_reg.r[i] = 0;



        /* flags (ccr) */

        for(i=0; i<4; i++)
                rsp_reg.flag[i] = 0;



        /* vector registers */

        for(j=0; j<32; j++)
                for(i=0; i<8; i++)
                        rsp_reg.v[j][i] = 0;



        /* hidden registers */

        rsp_reg.vrcph_source = 0;
        rsp_reg.vrcph_result = 0;

} /* void rsp_reset() */










/******************************************************************************\
*                                                                              *
*   RSP (Reality Signal Processor)                                             *
*                                                                              *
\******************************************************************************/





static void rsp_special()
{
        PRINT_XINST("special instruction")

        (special_instruction[__F])();

}





static void rsp_regimm()
{
        PRINT_XINST("regimm instruction")

        (regimm_instruction[__RT])();

}





static void rsp_j()
{
        PRINT_DEBUGGING_INFO("J", T)

        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = __T;

}





static void rsp_jal()
{
        PRINT_DEBUGGING_INFO(ST_BG_BLU "JAL" ST_BG_DEF, T)

        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = __T;
        rsp_reg.r[31] = (mem.sp_reg_pc + 8) & 0x1fff;

}





static void rsp_beq()
{
        PRINT_DEBUGGING_INFO("BEQ", RS_RT_O)

        if(rsp_reg.r[__RS] == rsp_reg.r[__RT])
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }

}





static void rsp_bne()
{
        PRINT_DEBUGGING_INFO("BNE", RS_RT_O)

        if(rsp_reg.r[__RS] != rsp_reg.r[__RT])
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }

}





static void rsp_blez()
{
        PRINT_DEBUGGING_INFO("BLEZ", RS_O)

        if(((signed)rsp_reg.r[__RS]) <= 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }

}





static void rsp_bgtz()
{
        PRINT_DEBUGGING_INFO("BGTZ", RS_O)

        if(((signed)rsp_reg.r[__RS]) > 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }

}





static void rsp_addi()
{
        PRINT_DEBUGGING_INFO("ADDI", RT_RS_I)

        rsp_reg.r[__RT] = rsp_reg.r[__RS] + (tr_s32)__I;

}





static void rsp_addiu()
{
        PRINT_DEBUGGING_INFO("ADDIU", RT_RS_I)

        rsp_reg.r[__RT] = rsp_reg.r[__RS] + (tr_s32)__I;

}





static void rsp_slti()
{
        PRINT_DEBUGGING_INFO("SLTI", RT_RS_I)

        if(rsp_reg.r[__RS] < __I)
                rsp_reg.r[__RT] = 1;
        else
                rsp_reg.r[__RT] = 0;

}





static void rsp_sltiu()
{
        PRINT_DEBUGGING_INFO("SLTIU", RT_RS_I)

        if((tr_u32)rsp_reg.r[__RS] < (tr_u32)__I)
                rsp_reg.r[__RT] = 1;
        else
                rsp_reg.r[__RT] = 0;

}





static void rsp_andi()
{
        PRINT_DEBUGGING_INFO("ANDI", RT_RS_I)

        rsp_reg.r[__RT] = rsp_reg.r[__RS] & (tr_u16)__I;

}





static void rsp_ori()
{
        PRINT_DEBUGGING_INFO("ORI", RT_RS_I)

        rsp_reg.r[__RT] = rsp_reg.r[__RS] | (tr_u16)__I;

}





static void rsp_xori()
{
        PRINT_DEBUGGING_INFO("XORI", RT_RS_I)

        rsp_reg.r[__RT] = rsp_reg.r[__RS] ^ (tr_u16)__I;

}





static void rsp_lui()
{
        PRINT_DEBUGGING_INFO("LUI", RT_I)

        rsp_reg.r[__RT] = __I << 16;
      
}





static void rsp_cop0()
{
#ifdef TR_DEBUG
#ifdef TR_DEBUG_XINST
        if(prefs.debug & DBG_RSP_INSTR)
        {
                puts("cop0 instruction");
        }
#endif
#endif

        (cop0_instruction[__RS])();
   
}





static void rsp_cop2()
{
#ifdef TR_DEBUG
#ifdef TR_DEBUG_XINST
        if(prefs.debug & DBG_RSP_INSTR)
        {
                puts("cop2 instruction");
        }
#endif
#endif

        (cop2_instruction[__RS])();

}





static void rsp_lb()
{
        PRINT_DEBUGGING_INFO("LB", RT_O_B)

#ifdef TR_ENDIAN_IS_LITTLE
        rsp_reg.r[__RT] = (tr_s32)(tr_s8)((tr_u8 *)mem.sp_dmem)[((rsp_reg.r[__RS]+__I)&0xfff)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
        rsp_reg.r[__RT] = (tr_s32)(tr_s8)((tr_u8 *)mem.sp_dmem)[(rsp_reg.r[__RS]+__I)&0xfff];
#endif

}





static void rsp_lh()
{
        tr_u32 addr;

        PRINT_DEBUGGING_INFO("LH", RT_O_B)

        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        rsp_reg.r[__RT] = (tr_s32)(tr_s16)
#ifdef TR_ENDIAN_IS_LITTLE
                            (mem.sp_dmem[(addr+0)^3] << 8)
                          | (mem.sp_dmem[(addr+1)^3]);
#endif
#ifdef TR_ENDIAN_IS_BIG
                            (mem.sp_dmem[addr+0] << 8)
                          | (mem.sp_dmem[addr+1]);
#endif

}





static void rsp_lw()
{
        tr_u32 addr;

        PRINT_DEBUGGING_INFO("LW", RT_O_B)

        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        /* printf("0: %02lx\n1: %02lx\n2: %02lx\n3: %02lx\n", (addr+0)^3, (addr+1)^3, (addr+2)^3, (addr+3)^3); */
        /* printf("0: %02x\n1: %02x\n2: %02x\n3: %02x\n", mem.sp_dmem[(addr+0)^3], mem.sp_dmem[(addr+1)^3], mem.sp_dmem[(addr+2)^3], mem.sp_dmem[(addr+3)^3]); */

        rsp_reg.r[__RT] = (tr_s32)(tr_s32)
#ifdef TR_ENDIAN_IS_LITTLE
                            (mem.sp_dmem[(addr+0)^3] << 24)
                          | (mem.sp_dmem[(addr+1)^3] << 16)
                          | (mem.sp_dmem[(addr+2)^3] << 8)
                          |  mem.sp_dmem[(addr+3)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
                            (mem.sp_dmem[addr+0] << 24)
                          | (mem.sp_dmem[addr+1] << 16)
                          | (mem.sp_dmem[addr+2] << 8)
                          |  mem.sp_dmem[addr+3];
#endif

if(prefs.debug & DBG_RSP_INSTR)
printf("   =%08"PR32"x\n", rsp_reg.r[__RT]);

}





static void rsp_lbu()
{
        PRINT_DEBUGGING_INFO("LBU", RT_O_B)

#ifdef TR_ENDIAN_IS_LITTLE
        rsp_reg.r[__RT] = (tr_s32)(tr_u8)((tr_u8 *)mem.sp_dmem)[((rsp_reg.r[__RS]+__I)&0xfff)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
        rsp_reg.r[__RT] = (tr_s32)(tr_u8)((tr_u8 *)mem.sp_dmem)[(rsp_reg.r[__RS]+__I)&0xfff];
#endif

}





static void rsp_lhu()
{
        tr_u32 addr;

        PRINT_DEBUGGING_INFO("LHU", RT_O_B)

        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        rsp_reg.r[__RT] = (tr_s32)(tr_u16)
#ifdef TR_ENDIAN_IS_LITTLE
                            (mem.sp_dmem[(addr+0)^3] << 8)
                          | (mem.sp_dmem[(addr+1)^3]);
#endif
#ifdef TR_ENDIAN_IS_BIG
                            (mem.sp_dmem[addr+0] << 8)
                          | (mem.sp_dmem[addr+1]);
#endif

}





static void rsp_sb()
{
        PRINT_DEBUGGING_INFO("SB", RT_O_B)

#ifdef TR_ENDIAN_IS_LITTLE
        mem.sp_dmem[((rsp_reg.r[__RS]+__I)&0xfff)^3] = (tr_u8)rsp_reg.r[__RT];
#endif
#ifdef TR_ENDIAN_IS_BIG
        mem.sp_dmem[((rsp_reg.r[__RS]+__I)&0xfff)] = (tr_u8)rsp_reg.r[__RT];
#endif

}





static void rsp_sh()
{
        tr_u32 data, addr;

        PRINT_DEBUGGING_INFO("SH", RT_O_B)

        data = rsp_reg.r[__RT];
        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
        mem.sp_dmem[(addr+0)^3] = (tr_u8)(data >> 8);
        mem.sp_dmem[(addr+1)^3] = (tr_u8)(data);
#endif
#ifdef TR_ENDIAN_IS_BIG
        mem.sp_dmem[addr+0] = (tr_u8)(data >> 8);
        mem.sp_dmem[addr+1] = (tr_u8)(data);
#endif

}





static void rsp_sw()
{
        tr_u32 data, addr;

        PRINT_DEBUGGING_INFO("SW", RT_O_B)

        data = rsp_reg.r[__RT];
        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
        mem.sp_dmem[(addr+0)^3] = (tr_u8)(data >> 24);
        mem.sp_dmem[(addr+1)^3] = (tr_u8)(data >> 16);
        mem.sp_dmem[(addr+2)^3] = (tr_u8)(data >> 8);
        mem.sp_dmem[(addr+3)^3] = (tr_u8)(data >> 0);
#endif
#ifdef TR_ENDIAN_IS_BIG
        mem.sp_dmem[addr+0] = (tr_u8)(data >> 24);
        mem.sp_dmem[addr+1] = (tr_u8)(data >> 16);
        mem.sp_dmem[addr+2] = (tr_u8)(data >> 8);
        mem.sp_dmem[addr+3] = (tr_u8)(data >> 0);
#endif

if(prefs.debug & DBG_RSP_INSTR)
printf("   =%08"PR32"x\n", data);

}





static void rsp_lwc2()
{
        tr_s32   addr;
        int     offset = __SA >> 1;
        int     i;



        /* printf("RS: %x\nRT: %x\n(RD: %x)\nSA: %x\nF:  %x\n", __RS, __RT, __RD, __SA, __F); */

        switch(__RD)
        {
            case 0x00:
                PRINT_DEBUGGING_INFO("LBV", VE_O_B(0))

                addr = __RSP_O(0) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                if(offset & 1)
                {
                        rsp_reg.v[__RT][offset>>1] &= 0xff00;
                        rsp_reg.v[__RT][offset>>1] |= mem.sp_dmem[addr^3];
                }
                else
                {
                        rsp_reg.v[__RT][offset>>1] &= 0x00ff;
                        rsp_reg.v[__RT][offset>>1] |= mem.sp_dmem[addr^3] << 8;
                }
#endif
#ifdef TR_ENDIAN_IS_BIG
                if(offset & 1)
                {
                        rsp_reg.v[__RT][offset>>1] &= 0xff00;
                        rsp_reg.v[__RT][offset>>1] |= mem.sp_dmem[addr];
                }
                else
                {
                        rsp_reg.v[__RT][offset>>1] &= 0x00ff;
                        rsp_reg.v[__RT][offset>>1] |= mem.sp_dmem[addr] << 8;
                }
#endif
                break;

            case 0x01:
                PRINT_DEBUGGING_INFO("LSV", VE_O_B(1))

                addr = __RSP_O(1) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[addr^3] << 8) | mem.sp_dmem[(addr+1)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[addr] << 8) | mem.sp_dmem[addr+1];
#endif
                break;

            case 0x02:
                PRINT_DEBUGGING_INFO("LLV", VE_O_B(2))

                addr = __RSP_O(2) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[(addr)^3] << 8) | mem.sp_dmem[(addr+1)^3];
                rsp_reg.v[__RT][(offset>>1)+1] = (mem.sp_dmem[(addr+2)^3] << 8) | mem.sp_dmem[(addr+3)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[addr] << 8) | mem.sp_dmem[addr+1];
                rsp_reg.v[__RT][(offset>>1)+1] = (mem.sp_dmem[addr+2] << 8) | mem.sp_dmem[addr+3];
#endif
                break;

            case 0x03:
                PRINT_DEBUGGING_INFO("LDV", VE_O_B(3))

                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[(addr)^3] << 8) | mem.sp_dmem[(addr+1)^3];
                rsp_reg.v[__RT][(offset>>1)+1] = (mem.sp_dmem[(addr+2)^3] << 8) | mem.sp_dmem[(addr+3)^3];
                rsp_reg.v[__RT][(offset>>1)+2] = (mem.sp_dmem[(addr+4)^3] << 8) | mem.sp_dmem[(addr+5)^3];
                rsp_reg.v[__RT][(offset>>1)+3] = (mem.sp_dmem[(addr+6)^3] << 8) | mem.sp_dmem[(addr+7)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[addr+0] << 8) | mem.sp_dmem[addr+1];
                rsp_reg.v[__RT][(offset>>1)+1] = (mem.sp_dmem[addr+2] << 8) | mem.sp_dmem[addr+3];
                rsp_reg.v[__RT][(offset>>1)+2] = (mem.sp_dmem[addr+4] << 8) | mem.sp_dmem[addr+5];
                rsp_reg.v[__RT][(offset>>1)+3] = (mem.sp_dmem[addr+6] << 8) | mem.sp_dmem[addr+7];
#endif
                break;

            case 0x04:
                PRINT_DEBUGGING_INFO("LQV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (16 - (addr & 0xf)); i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        if((offset+i) & 1)
                        {
                                rsp_reg.v[__RT][((offset+i)>>1)] &= 0xff00;
                                rsp_reg.v[__RT][((offset+i)>>1)] |= mem.sp_dmem[(addr+i)^3];
                        }
                        else
                        {
                                rsp_reg.v[__RT][((offset+i)>>1)] &= 0x00ff;
                                rsp_reg.v[__RT][((offset+i)>>1)] |= mem.sp_dmem[(addr+i)^3] << 8;
                        }
#endif
#ifdef TR_ENDIAN_IS_BIG
                        if((offset+i) & 1)
                        {
                                rsp_reg.v[__RT][((offset+i)>>1)] &= 0xff00;
                                rsp_reg.v[__RT][((offset+i)>>1)] |= mem.sp_dmem[addr+i];
                        }
                        else
                        {
                                rsp_reg.v[__RT][((offset+i)>>1)] &= 0x00ff;
                                rsp_reg.v[__RT][((offset+i)>>1)] |= mem.sp_dmem[addr+i] << 8;
                        }
#endif
                }
                break;

            case 0x05:
                PRINT_DEBUGGING_INFO("LRV", VE_O_B(4))
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (addr & 0xf); i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        if(i & 1)
                        {
                                rsp_reg.v[__RT][7-(i>>1)] &= 0x00ff;
                                rsp_reg.v[__RT][7-(i>>1)] |= mem.sp_dmem[(addr-i-1)^3] << 8;
                        }
                        else
                        {
                                rsp_reg.v[__RT][7-(i>>1)] &= 0xff00;
                                rsp_reg.v[__RT][7-(i>>1)] |= mem.sp_dmem[(addr-i-1)^3];
                        }
#endif
#ifdef TR_ENDIAN_IS_BIG
                        if(i & 1)
                        {
                                rsp_reg.v[__RT][7-(i>>1)] &= 0x00ff;
                                rsp_reg.v[__RT][7-(i>>1)] |= mem.sp_dmem[addr-i-1] << 8;
                        }
                        else
                        {
                                rsp_reg.v[__RT][7-(i>>1)] &= 0xff00;
                                rsp_reg.v[__RT][7-(i>>1)] |= mem.sp_dmem[addr-i-1];
                        }
#endif
                }
                break;

            case 0x06:
                PRINT_DEBUGGING_INFO("LPV", VE_O_B(3))

                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
#ifdef TR_ENDIAN_IS_LITTLE
                rsp_reg.v[__RT][offset>>1] = mem.sp_dmem[addr^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+1] = mem.sp_dmem[(addr+1)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+2] = mem.sp_dmem[(addr+2)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+3] = mem.sp_dmem[(addr+3)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+4] = mem.sp_dmem[(addr+4)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+5] = mem.sp_dmem[(addr+5)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+6] = mem.sp_dmem[(addr+6)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+7] = mem.sp_dmem[(addr+7)^3] << 8;
#endif
#ifdef TR_ENDIAN_IS_BIG
                rsp_reg.v[__RT][offset>>1] = mem.sp_dmem[addr^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+1] = mem.sp_dmem[addr+1] << 8;
                rsp_reg.v[__RT][(offset>>1)+2] = mem.sp_dmem[addr+2] << 8;
                rsp_reg.v[__RT][(offset>>1)+3] = mem.sp_dmem[addr+3] << 8;
                rsp_reg.v[__RT][(offset>>1)+4] = mem.sp_dmem[addr+4] << 8;
                rsp_reg.v[__RT][(offset>>1)+5] = mem.sp_dmem[addr+5] << 8;
                rsp_reg.v[__RT][(offset>>1)+6] = mem.sp_dmem[addr+6] << 8;
                rsp_reg.v[__RT][(offset>>1)+7] = mem.sp_dmem[addr+7] << 8;
#endif
                break;

            case 0x07:
                PRINT_DEBUGGING_INFO("LUV", VE_O_B(3))

                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i<8; i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        rsp_reg.v[__RT][((offset)>>1)+i] = mem.sp_dmem[(addr+i)^3] << 7;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        rsp_reg.v[__RT][((offset)>>1)+i] = mem.sp_dmem[addr+i] << 7;
#endif
                }
                break;

            case 0x08:
                PRINT_DEBUGGING_INFO("LHV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i < (16 - (addr & 0xf)); i+=2)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        rsp_reg.v[__RT][(offset+i)>>1] = mem.sp_dmem[(addr+i)^3] << 7;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        rsp_reg.v[__RT][(offset+i)>>1] = mem.sp_dmem[addr+i] << 7;
#endif
                }
                break;

            case 0x09:
                PRINT_DEBUGGING_INFO("LFV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (16 - (addr & 0xf)); i+=4)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        rsp_reg.v[__RT][(offset>>1)+(i>>2)] = mem.sp_dmem[(addr+i)^3] << 7;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        rsp_reg.v[__RT][(offset>>1)+(i>>2)] = mem.sp_dmem[addr+i] << 7;
#endif
                }
                break;

            case 0x0a:
                PRINT_DEBUGGING_INFO("LWV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i<16; i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        if(i >= offset)
                        {
                                if(i & 1)
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0xff00;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i-offset)^3];
                                }
                                else
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0x00ff;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i-offset)^3] << 8;
                                }
                        }
                        else
                        {
                                if(i & 1)
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0xff00;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i+16-offset)^3];
                                }
                                else
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0x00ff;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i+16-offset)^3] << 8;
                                }
                        }
#endif
#ifdef TR_ENDIAN_IS_BIG
                        if(i >= offset)
                        {
                                if(i & 1)
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0xff00;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[addr+i-offset];
                                }
                                else
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0x00ff;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[addr+i-offset] << 8;
                                }
                        }
                        else
                        {
                                if(i & 1)
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0xff00;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[addr+i+16-offset];
                                }
                                else
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0x00ff;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[addr+i+16-offset] << 8;
                                }
                        }
#endif
                }
                break;

            case 0x0b:
                PRINT_DEBUGGING_INFO("LTV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i<8; i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7] = (mem.sp_dmem[(addr+(i<<1))^3] << 8) | mem.sp_dmem[(addr+(i<<1)+1)^3];
#endif
#ifdef TR_ENDIAN_IS_BIG
                        rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7] = (mem.sp_dmem[addr+(i<<1)] << 8) | mem.sp_dmem[addr+(i<<1)+1];
#endif
                }
                break;

            default:
                PRINT_NOT_IMPLEMENTED("L?V")

                RSP_HALT
                break;

        } /* switch(__RD) */

}





static void rsp_swc2()
{
        tr_s32   addr;
        int     offset = __SA >> 1;
        int     i;


        /* printf("RS: %x\nRT: %x\n(RD: %x)\nSA: %x\nF:  %x\n", __RS, __RT, __RD, __SA, __F); */

        switch(__RD)
        {
            case 0x00:
                PRINT_DEBUGGING_INFO("SBV", VE_O_B(0))

                addr = __RSP_O(0) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                if(offset & 1)
                        mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1];
                else
                        mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
#endif
#ifdef TR_ENDIAN_IS_BIG
                if(offset & 1)
                        mem.sp_dmem[addr] = rsp_reg.v[__RT][offset>>1];
                else
                        mem.sp_dmem[addr] = rsp_reg.v[__RT][offset>>1] >> 8;
#endif
                break;

            case 0x01:
                PRINT_DEBUGGING_INFO("SSV", VE_O_B(1))

                addr = __RSP_O(1) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][offset>>1];
                
#endif
#ifdef TR_ENDIAN_IS_BIG
                mem.sp_dmem[addr] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[addr+1] = rsp_reg.v[__RT][offset>>1];
#endif
                break;

            case 0x02:
                PRINT_DEBUGGING_INFO("SLV", VE_O_B(2))

                addr = __RSP_O(2) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][offset>>1];
                mem.sp_dmem[(addr+2)^3] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[(addr+3)^3] = rsp_reg.v[__RT][(offset>>1)+1];
#endif
#ifdef TR_ENDIAN_IS_BIG
                mem.sp_dmem[addr] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[addr+1] = rsp_reg.v[__RT][offset>>1];
                mem.sp_dmem[addr+2] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[addr+3] = rsp_reg.v[__RT][(offset>>1)+1];
#endif
                break;

            case 0x03:
                PRINT_DEBUGGING_INFO("SDV", VE_O_B(3))

                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

#ifdef TR_ENDIAN_IS_LITTLE
                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][offset>>1];
                mem.sp_dmem[(addr+2)^3] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[(addr+3)^3] = rsp_reg.v[__RT][(offset>>1)+1];
                mem.sp_dmem[(addr+4)^3] = rsp_reg.v[__RT][(offset>>1)+2] >> 8;
                mem.sp_dmem[(addr+5)^3] = rsp_reg.v[__RT][(offset>>1)+2];
                mem.sp_dmem[(addr+6)^3] = rsp_reg.v[__RT][(offset>>1)+3] >> 8;
                mem.sp_dmem[(addr+7)^3] = rsp_reg.v[__RT][(offset>>1)+3];
#endif
#ifdef TR_ENDIAN_IS_BIG
                mem.sp_dmem[addr] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[addr+1] = rsp_reg.v[__RT][offset>>1];
                mem.sp_dmem[addr+2] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[addr+3] = rsp_reg.v[__RT][(offset>>1)+1];
                mem.sp_dmem[addr+4] = rsp_reg.v[__RT][(offset>>1)+2] >> 8;
                mem.sp_dmem[addr+5] = rsp_reg.v[__RT][(offset>>1)+2];
                mem.sp_dmem[addr+6] = rsp_reg.v[__RT][(offset>>1)+3] >> 8;
                mem.sp_dmem[addr+7] = rsp_reg.v[__RT][(offset>>1)+3];
#endif
                break;

            case 0x04:
                PRINT_DEBUGGING_INFO("SQV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (16 - (addr & 0xf)); i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        if((offset+i) & 1)
                                mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset+i)>>1];
                        else
                                mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset+i)>>1] >> 8;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        if((offset+i) & 1)
                                mem.sp_dmem[addr+i] = rsp_reg.v[__RT][(offset+i)>>1];
                        else
                                mem.sp_dmem[addr+i] = rsp_reg.v[__RT][(offset+i)>>1] >> 8;
#endif
                }
                break;

            case 0x05:
                PRINT_DEBUGGING_INFO("SRV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (addr & 0xf); i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        if((offset+i) & 1)
                                mem.sp_dmem[(addr-i-1)^3] = rsp_reg.v[__RT][7-(i>>1)] >> 8;
                        else
                                mem.sp_dmem[(addr-i-1)^3] = rsp_reg.v[__RT][7-(i>>1)];
#endif
#ifdef TR_ENDIAN_IS_BIG
                        if((offset+i) & 1)
                                mem.sp_dmem[addr-i-1] = rsp_reg.v[__RT][7-(i>>1)] >> 8;
                        else
                                mem.sp_dmem[addr-i-1] = rsp_reg.v[__RT][7-(i>>1)];
#endif
                }
                break;

            case 0x06:
                PRINT_DEBUGGING_INFO("SPV", VE_O_B(3))

                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
#ifdef TR_ENDIAN_IS_LITTLE
                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[(addr+2)^3] = rsp_reg.v[__RT][(offset>>1)+2] >> 8;
                mem.sp_dmem[(addr+3)^3] = rsp_reg.v[__RT][(offset>>1)+3] >> 8;
                mem.sp_dmem[(addr+4)^3] = rsp_reg.v[__RT][(offset>>1)+4] >> 8;
                mem.sp_dmem[(addr+5)^3] = rsp_reg.v[__RT][(offset>>1)+5] >> 8;
                mem.sp_dmem[(addr+6)^3] = rsp_reg.v[__RT][(offset>>1)+6] >> 8;
                mem.sp_dmem[(addr+7)^3] = rsp_reg.v[__RT][(offset>>1)+7] >> 8;
#endif
#ifdef TR_ENDIAN_IS_BIG
                mem.sp_dmem[addr] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[addr+1] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[addr+2] = rsp_reg.v[__RT][(offset>>1)+2] >> 8;
                mem.sp_dmem[addr+3] = rsp_reg.v[__RT][(offset>>1)+3] >> 8;
                mem.sp_dmem[addr+4] = rsp_reg.v[__RT][(offset>>1)+4] >> 8;
                mem.sp_dmem[addr+5] = rsp_reg.v[__RT][(offset>>1)+5] >> 8;
                mem.sp_dmem[addr+6] = rsp_reg.v[__RT][(offset>>1)+6] >> 8;
                mem.sp_dmem[addr+7] = rsp_reg.v[__RT][(offset>>1)+7] >> 8;
#endif
                break;

            case 0x07:
                PRINT_DEBUGGING_INFO("SUV", VE_O_B(3))

                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i<8; i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][((offset)>>1)+i] >> 7;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        mem.sp_dmem[addr+i] = rsp_reg.v[__RT][((offset)>>1)+i] >> 7;
#endif
                }
                break;

            case 0x08:
                PRINT_DEBUGGING_INFO("SHV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i < (16 - (addr & 0xf)); i+=2)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset+i)>>1] >> 7;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        mem.sp_dmem[addr+i] = rsp_reg.v[__RT][(offset+i)>>1] >> 7;
#endif
                }
                break;

            case 0x09:
                PRINT_DEBUGGING_INFO("SFV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i < (16 - (addr & 0xf)); i+=4)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset>>1)+(i>>2)] >> 7;
#endif
#ifdef TR_ENDIAN_IS_BIG
                        mem.sp_dmem[addr+i] = rsp_reg.v[__RT][(offset>>1)+(i>>2)] >> 7;
#endif
                }
                break;

            case 0x0a:
                PRINT_DEBUGGING_INFO("SWV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i<16; i++)
                {
#ifdef TR_ENDIAN_IS_LITTLE
                        if(i >= offset)
                        {
                                if(i & 1)
                                        mem.sp_dmem[(addr+i-offset)^3] = rsp_reg.v[__RT][i>>1];
                                else
                                        mem.sp_dmem[(addr+i-offset)^3] = rsp_reg.v[__RT][i>>1] >> 8;
                        }
                        else
                        {
                                if(i & 1)
                                        mem.sp_dmem[(addr+i+16-offset)^3] = rsp_reg.v[__RT][i>>1];
                                else
                                        mem.sp_dmem[(addr+i+16-offset)^3] = rsp_reg.v[__RT][i>>1] >> 8;
                        }
#endif
#ifdef TR_ENDIAN_IS_BIG
                        if(i >= offset)
                        {
                                if(i & 1)
                                        mem.sp_dmem[addr+i-offset] = rsp_reg.v[__RT][i>>1];
                                else
                                        mem.sp_dmem[addr+i-offset] = rsp_reg.v[__RT][i>>1] >> 8;
                        }
                        else
                        {
                                if(i & 1)
                                        mem.sp_dmem[addr+i+16-offset] = rsp_reg.v[__RT][i>>1];
                                else
                                        mem.sp_dmem[addr+i+16-offset] = rsp_reg.v[__RT][i>>1] >> 8;
                        }
#endif
                }
                break;

            case 0x0b:
                PRINT_DEBUGGING_INFO("STV", VE_O_B(4))

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                /* The following was VERY VERY hard to figure out!!! */
                for(i=0; i<8; i++)
                {
                        tr_s32   o = offset ? addr + (16-offset) : addr;
#ifdef TR_ENDIAN_IS_LITTLE
                        mem.sp_dmem[((o & ~0x0f) | ((o + (i<<1)) & 0x0f))^3] = rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7] >> 8;
                        mem.sp_dmem[(((o & ~0x0f) | ((o + (i<<1)) & 0x0f)) + 1)^3] = rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7];
#endif
#ifdef TR_ENDIAN_IS_BIG
                        mem.sp_dmem[(o & ~0x0f) | ((o + (i<<1)) & 0x0f)] = rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7] >> 8;
                        mem.sp_dmem[((o & ~0x0f) | ((o + (i<<1)) & 0x0f)) + 1] = rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7];
#endif
                }
                break;

            default:
                PRINT_NOT_IMPLEMENTED("S?V")

                RSP_HALT
                break;

        } /* switch(__RD) */

}






static void rsp_reserved()
{
        PRINT_NOT_IMPLEMENTED("reserved")

        RSP_HALT

}





static void rsp_special_sll()
{
        PRINT_DEBUGGING_INFO("SLL", RD_RT_SA)

        rsp_reg.r[__RD] = (tr_u32)rsp_reg.r[__RT] << __SA;

}





static void rsp_special_srl()
{
        PRINT_DEBUGGING_INFO("SRL", RD_RT_SA)

        rsp_reg.r[__RD] = (tr_u32)rsp_reg.r[__RT] >> __SA;

}





static void rsp_special_sra()
{
        PRINT_DEBUGGING_INFO("SRA", RD_RT_SA)

        rsp_reg.r[__RD] = (tr_s32)rsp_reg.r[__RT] >> __SA;

}





static void rsp_special_sllv()
{
        PRINT_DEBUGGING_INFO("SLLV", RD_RT_RS)

        rsp_reg.r[__RD] = rsp_reg.r[__RT] << (rsp_reg.r[__RS] & 0x1f);

}





static void rsp_special_srlv()
{
        PRINT_DEBUGGING_INFO("SRLV", RD_RT_RS)

        rsp_reg.r[__RD] = (tr_u32)rsp_reg.r[__RT] >> (rsp_reg.r[__RS] & 0x1f);

}





static void rsp_special_srav()
{
        PRINT_DEBUGGING_INFO("SRAV", RD_RT_RS)

        rsp_reg.r[__RD] = (tr_s32)rsp_reg.r[__RT] >> (rsp_reg.r[__RS] & 0x1f);

}





static void rsp_special_jr()
{
        PRINT_DEBUGGING_INFO(ST_BG_BLU "JR" ST_BG_DEF, RS)

        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = 0x04000000 | (rsp_reg.r[__RS] & 0x1fff);

}





static void rsp_special_jalr()
{
        PRINT_DEBUGGING_INFO("JALR", RD_RS)

        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = 0x04000000 | (rsp_reg.r[__RS] & 0x1fff);
        rsp_reg.r[__RD] = (mem.sp_reg_pc + 8) & 0x1fff;

}





static void rsp_special_break()
{
        PRINT_DEBUGGING_INFO("BREAK", NOTHING)

        ((tr_u32 *)mem.sp_reg)[4] |= 0x00000203;

        if(((tr_u32 *)mem.sp_reg)[4] & 0x40)
                reg.do_or_check_sthg |= RS4300I_DO_SP_INTERRUPT;

        rsp_reg.halt = 1;
        /* puts(ST_FG_BRO "RSP BREAK: RSP halted!" ST_FG_DEF); */
        /* fflush(stdout); */

}





static void rsp_special_add()
{
        PRINT_DEBUGGING_INFO("ADD", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] + rsp_reg.r[__RT];

}





static void rsp_special_addu()
{
        PRINT_DEBUGGING_INFO("ADDU", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] + rsp_reg.r[__RT];

}





static void rsp_special_sub()
{
        PRINT_DEBUGGING_INFO("SUB", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] - rsp_reg.r[__RT];

}





static void rsp_special_subu()
{
        PRINT_DEBUGGING_INFO("SUBU", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] - rsp_reg.r[__RT];

}





static void rsp_special_and()
{
        PRINT_DEBUGGING_INFO("AND", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] & rsp_reg.r[__RT];

}





static void rsp_special_or()
{
        PRINT_DEBUGGING_INFO("OR", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] | rsp_reg.r[__RT];

}





static void rsp_special_xor()
{
        PRINT_DEBUGGING_INFO("XOR", RD_RS_RT)

        rsp_reg.r[__RD] = rsp_reg.r[__RS] ^ rsp_reg.r[__RT];

}





static void rsp_special_nor()
{
        PRINT_DEBUGGING_INFO("NOR", RD_RS_RT)

        rsp_reg.r[__RD] = ~(rsp_reg.r[__RS] | rsp_reg.r[__RT]);
}





static void rsp_special_slt()
{
        PRINT_DEBUGGING_INFO("SLT", RD_RS_RT)

        if(rsp_reg.r[__RS] < rsp_reg.r[__RT])
                rsp_reg.r[__RD] = 1;
        else
                rsp_reg.r[__RD] = 0;

}





static void rsp_special_sltu()
{
        PRINT_DEBUGGING_INFO("SLTU", RD_RS_RT)



        if((tr_u32)rsp_reg.r[__RS] < (tr_u32)rsp_reg.r[__RT])
                rsp_reg.r[__RD] = 1;
        else
                rsp_reg.r[__RD] = 0;

}





static void rsp_special_reserved()
{
        PRINT_NOT_IMPLEMENTED("special reserved")

        RSP_HALT

}




static void rsp_regimm_bltz()
{
        PRINT_DEBUGGING_INFO("BLTZ", RS_O)

        if(((tr_s32)rsp_reg.r[__RS]) < 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }

}





static void rsp_regimm_bgez()
{
        PRINT_DEBUGGING_INFO("BGEZ", RS_O)

        if(((tr_s32)rsp_reg.r[__RS]) >= 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }

}





static void rsp_regimm_bltzal()
{
        PRINT_DEBUGGING_INFO("BLTZAL", RS_O)

        rsp_reg.r[31] = mem.sp_reg_pc + 8;

        if(((tr_s32)rsp_reg.r[__RS]) < 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }

}





static void rsp_regimm_bgezal()
{
        PRINT_DEBUGGING_INFO("BGEZAL", RS_O)

        rsp_reg.r[31] = mem.sp_reg_pc + 8;

        if(((tr_s32)rsp_reg.r[__RS]) >= 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }

}





static void rsp_regimm_reserved()
{
        PRINT_NOT_IMPLEMENTED("regrimm reserved")

        RSP_HALT

}





static void rsp_cop0_mf()
{
        PRINT_DEBUGGING_INFO("MFC0", RT_C0)

        if(__RD < 8)
                rsp_reg.r[__RT] = doReadMemWord(0x04040000 | (__RD << 2));
        else
                rsp_reg.r[__RT] = doReadMemWord(0x04100000 | ((__RD & ~0x08) << 2));

if(prefs.debug & DBG_RSP_INSTR)
printf("   =%08"PR32"x\n", rsp_reg.r[__RT]);

}





static void rsp_cop0_mt()
{
        PRINT_DEBUGGING_INFO("MTC0", RT_C0)

        if(__RD < 8)
                doWriteMemWord(rsp_reg.r[__RT], 0x04040000 | (__RD << 2));
        else
                doWriteMemWord(rsp_reg.r[__RT], 0x04100000 | ((__RD & ~0x08) << 2));

if(prefs.debug & DBG_RSP_INSTR)
printf("   =%08"PR32"x\n", rsp_reg.r[__RT]);

}





static void rsp_cop0_reserved()
{
        PRINT_NOT_IMPLEMENTED("cop0 reserved")

        RSP_HALT

}





static void rsp_cop2_mf()
{
        tr_s16  data;
        tr_u8    offs;

        PRINT_DEBUGGING_INFO("MFC2", RT_C2)

        offs = (__SA >> 1) & 0x0f;

        if(offs & 0x01)   /* odd */
        {
                data =  rsp_reg.v[__RD][(offs >> 1) + 1] & 0x00ff;
                data += rsp_reg.v[__RD][offs >> 1]       & 0xff00;
        }
        else              /* even */
        {
                data = rsp_reg.v[__RD][offs >> 1];
        }

        rsp_reg.r[__RT] = data;

}





static void rsp_cop2_cf()
{
        PRINT_DEBUGGING_INFO("CFC2", RT_FLAG)

        rsp_reg.r[__RT] = rsp_reg.flag[__RD & 0x03];

}





static void rsp_cop2_mt()
{
        tr_u16   data;
        tr_u8    offs;

        PRINT_DEBUGGING_INFO("MTC2", RT_C2)

        data = rsp_reg.r[__RT];
        offs = (__SA >> 1) & 0x0f;

        if(offs & 0x01)   /* odd */
        {
                rsp_reg.v[__RD][offs >> 1]       &= 0xff00;
                rsp_reg.v[__RD][offs >> 1]       |= data >> 8;
                rsp_reg.v[__RD][(offs >> 1) + 1] &= 0x00ff;
                rsp_reg.v[__RD][(offs >> 1) + 1] |= data << 8;
        }
        else              /* even */
        {
                rsp_reg.v[__RD][offs >> 1] = data;
        }

}





static void rsp_cop2_ct()
{
        PRINT_DEBUGGING_INFO("CTC2", RT_FLAG)

        rsp_reg.flag[__RD & 0x03] = rsp_reg.r[__RT];
   
}





static void rsp_cop2_vectop()
{
        (cop2_vectop_instruction[__F])();

}





static void rsp_cop2_reserved()
{
        PRINT_NOT_IMPLEMENTED("cop2 reserved")

        RSP_HALT

}





/* rsp cop2 vectop instructions */

static void rsp_cop2_vectop_vmulf()
{
        PRINT_DEBUGGING_INFO("VMULF", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VMULF")

        RSP_HALT

} /* static void rsp_cop2_vectop_vmulf() */





static void rsp_cop2_vectop_vmulu()
{
        PRINT_DEBUGGING_INFO("VMULU", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VMULU")

        RSP_HALT

} /* static void rsp_cop2_vectop_vmulu() */





static void rsp_cop2_vectop_vrndp()
{
        PRINT_DEBUGGING_INFO("VRNDP", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VRNDP")

        RSP_HALT

} /* static void rsp_cop2_vectop_vrndp() */





static void rsp_cop2_vectop_vmulq()
{
        PRINT_DEBUGGING_INFO("VMULQ", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VMULQ")

        RSP_HALT

} /* static void rsp_cop2_vectop_vmulq() */





static void rsp_cop2_vectop_vmudl()
{
        PRINT_DEBUGGING_INFO("VMUDL", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(ah, , (tr_u16), *, >> 16)

        ACCU_2_VREG(ah);

} /* static void rsp_cop2_vectop_vmudl() */





static void rsp_cop2_vectop_vmudm()
{
        PRINT_DEBUGGING_INFO("VMUDM", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(ah, , (tr_s16), *, >> 16)

        ACCU_2_VREG(ah);

} /* static void rsp_cop2_vectop_vmudm() */





static void rsp_cop2_vectop_vmudn()
{
        PRINT_DEBUGGING_INFO("VMUDN", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(al, , (tr_u16), *, )

        ACCU_2_VREG(al);

} /* static void rsp_cop2_vectop_vmudn() */





static void rsp_cop2_vectop_vmudh()
{
        PRINT_DEBUGGING_INFO("VMUDH", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(ah, , (tr_s16), *, )

        ACCU_2_VREG(ah);

} /* static void rsp_cop2_vectop_vmudh() */





static void rsp_cop2_vectop_vmacf()
{
        PRINT_DEBUGGING_INFO("VMACF", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VMACF")

        RSP_HALT

} /* static void rsp_cop2_vectop_vmacf() */





static void rsp_cop2_vectop_vmacu()
{
        PRINT_DEBUGGING_INFO("VMACU", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VMACU")

        RSP_HALT

} /* static void rsp_cop2_vectop_vmacu() */





static void rsp_cop2_vectop_vrndn()
{
        PRINT_DEBUGGING_INFO("VRNDN", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VRNDN")

        RSP_HALT

} /* static void rsp_cop2_vectop_vrndn() */





static void rsp_cop2_vectop_vmacq()
{
        PRINT_DEBUGGING_INFO("VMACQ", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VMACQ")

        RSP_HALT

} /* static void rsp_cop2_vectop_vmacq() */





static void rsp_cop2_vectop_vmadl()
{
        PRINT_DEBUGGING_INFO("VMADL", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (tr_u16), *, >> 16)

        ACCU_2_VREG(ah);

} /* static void rsp_cop2_vectop_vmadl() */





static void rsp_cop2_vectop_vmadm()
{
        PRINT_DEBUGGING_INFO("VMADM", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (tr_s16), *, >> 16)

        ACCU_2_VREG(ah);

} /* static void rsp_cop2_vectop_vmadm() */





static void rsp_cop2_vectop_vmadn()
{
        PRINT_DEBUGGING_INFO("VMADN", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(al, +, (tr_u16), *, )

        ACCU_2_VREG(al);

} /* static void rsp_cop2_vectop_vmadn() */





static void rsp_cop2_vectop_vmadh()
{
        PRINT_DEBUGGING_INFO("VMADH", V_V_VE)

        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (tr_s16), *, )

        ACCU_2_VREG(ah);

} /* static void rsp_cop2_vectop_vmadh() */





static void rsp_cop2_vectop_vadd()
{
        PRINT_DEBUGGING_INFO("VADD", V_V_VE)

        STORE_REGISTER
        VECT_OP(+)
        ADD_CARRY(__SA)
        CLIP_RESULT_ADD

} /* static void rsp_cop2_vectop_vadd() */





static void rsp_cop2_vectop_vsub()
{
        PRINT_DEBUGGING_INFO("VSUB", V_V_VE)

        STORE_REGISTER
        VECT_OP(-)
        SUB_CARRY(__SA)
        CLIP_RESULT_SUB

} /* static void rsp_cop2_vectop_vsub() */





static void rsp_cop2_vectop_vsut()
{
        PRINT_DEBUGGING_INFO("VSUT", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSUT")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsut() */





static void rsp_cop2_vectop_vabs()
{
        PRINT_DEBUGGING_INFO("VABS", V_V_VE)

        STORE_REGISTER
        VECT_OP_ABS

} /* static void rsp_cop2_vectop_vabs() */





static void rsp_cop2_vectop_vaddc()
{
        PRINT_DEBUGGING_INFO("VADDC", V_V_VE)

        STORE_REGISTER
        VECT_OP(+)
        SET_CARRY_ADD

} /* static void rsp_cop2_vectop_vaddc() */





static void rsp_cop2_vectop_vsubc()
{
        PRINT_DEBUGGING_INFO("VSUBC", V_V_VE)

        STORE_REGISTER
        VECT_OP(-)
        SET_CARRY_SUB
        SET_ZERO_FLAGS(__SA)

} /* static void rsp_cop2_vectop_vsubc() */





static void rsp_cop2_vectop_vaddb()
{
        PRINT_DEBUGGING_INFO("VADDB", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VADDB")

        RSP_HALT

} /* static void rsp_cop2_vectop_vaddb() */





static void rsp_cop2_vectop_vsubb()
{
        PRINT_DEBUGGING_INFO("VSUBB", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSUBB")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsubb() */





static void rsp_cop2_vectop_vaccb()
{
        PRINT_DEBUGGING_INFO("VACCB", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VACCB")

        RSP_HALT

} /* static void rsp_cop2_vectop_vaccb() */





static void rsp_cop2_vectop_vsucb()
{
        PRINT_DEBUGGING_INFO("VSUCB", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSUCB")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsucb() */





static void rsp_cop2_vectop_vsad()
{
        PRINT_DEBUGGING_INFO("VSAD", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSAD")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsad() */





static void rsp_cop2_vectop_vsac()
{
        PRINT_DEBUGGING_INFO("VSAC", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSAC")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsac() */





static void rsp_cop2_vectop_vsum()
{
        PRINT_DEBUGGING_INFO("VSUM", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSUM")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsum() */





static void rsp_cop2_vectop_vsaw()
{
        PRINT_DEBUGGING_INFO("VSAW", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VSAW")

        RSP_HALT

} /* static void rsp_cop2_vectop_vsaw() */





static void rsp_cop2_vectop_vlt()
{
        PRINT_DEBUGGING_INFO("VLT", V_V_VE)

        STORE_REGISTER
        VECT_OP_COMP(<)
        VECT_OP_COPY(rsp_reg.flag[1])
        rsp_reg.flag[0] = 0;

} /* static void rsp_cop2_vectop_vlt() */





static void rsp_cop2_vectop_veq()
{
        PRINT_DEBUGGING_INFO("VEQ", V_V_VE)

        STORE_REGISTER
        VECT_OP_COMP(==)
        VECT_OP_COPY(0)
        rsp_reg.flag[0] = 0;

} /* static void rsp_cop2_vectop_veq() */





static void rsp_cop2_vectop_vne()
{
        PRINT_DEBUGGING_INFO("VNE", V_V_VE)

        STORE_REGISTER
        VECT_OP_COMP(!=)
        VECT_OP_COPY(0xff)
        rsp_reg.flag[0] = 0;

} /* static void rsp_cop2_vectop_vne() */





static void rsp_cop2_vectop_vge()
{
        PRINT_DEBUGGING_INFO("VGE", V_V_VE)

        STORE_REGISTER
        VECT_OP_COMP(>=)
        VECT_OP_COPY(rsp_reg.flag[1])
        rsp_reg.flag[0] = 0;

} /* static void rsp_cop2_vectop_vge() */





static void rsp_cop2_vectop_vcl()
{
        PRINT_DEBUGGING_INFO("VCL", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VCL")

        RSP_HALT

} /* static void rsp_cop2_vectop_vcl() */





static void rsp_cop2_vectop_vch()
{
        PRINT_DEBUGGING_INFO("VCH", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VCH")

        RSP_HALT

} /* static void rsp_cop2_vectop_vch() */





static void rsp_cop2_vectop_vcr()
{
        PRINT_DEBUGGING_INFO("VCR", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VCR")

        RSP_HALT

} /* static void rsp_cop2_vectop_vcr() */





static void rsp_cop2_vectop_vmrg()
{
        int     i;

        PRINT_DEBUGGING_INFO("VMRG", V_V_VE)

        STORE_REGISTER

        for(i=0; i<8; i++)
        {
                if( (1<<i) & rsp_reg.flag[1] )
                        rsp_reg.v[__SA][i] = rsp_reg.v_src1[i];
                else
                        rsp_reg.v[__SA][i] = rsp_reg.v_src2[i];
        }

} /* static void rsp_cop2_vectop_vmrg() */





static void rsp_cop2_vectop_vand()
{
        PRINT_DEBUGGING_INFO("VAND", V_V_VE)

        STORE_REGISTER
        VECT_OP(&)

} /* static void rsp_cop2_vectop_vand() */





static void rsp_cop2_vectop_vnand()
{
        PRINT_DEBUGGING_INFO("VNAND", V_V_VE)

        STORE_REGISTER
        VECT_N_OP(&)

} /* static void rsp_cop2_vectop_vnand() */





static void rsp_cop2_vectop_vor()
{
        PRINT_DEBUGGING_INFO("VOR", V_V_VE)

        STORE_REGISTER
        VECT_OP(|)

} /* static void rsp_cop2_vectop_vor() */





static void rsp_cop2_vectop_vnor()
{
        PRINT_DEBUGGING_INFO("VNOR", V_V_VE)

        STORE_REGISTER
        VECT_N_OP(|)

} /* static void rsp_cop2_vectop_vnor() */





static void rsp_cop2_vectop_vxor()
{
        PRINT_DEBUGGING_INFO("VXOR", V_V_VE)

        STORE_REGISTER
        VECT_OP(^)

} /* static void rsp_cop2_vectop_vxor() */





static void rsp_cop2_vectop_vnxor()
{
        PRINT_DEBUGGING_INFO("VNXOR", V_V_VE)

        STORE_REGISTER
        VECT_N_OP(^)

} /* static void rsp_cop2_vectop_vnxor() */





static void rsp_cop2_vectop_vrcp()
{
        PRINT_DEBUGGING_INFO("VRCP", VE_VE)

        rsp_reg.v[__SA][__RD & 0x07] = rsp_reg.vrcph_source;
        rsp_reg.vrcph_result = rsp_reg.v[__RT][__RS & 0x07];

} /* static void rsp_cop2_vectop_vrcp() */





static void rsp_cop2_vectop_vrcpl()
{
        tr_u32 result;
        tr_u32 source;

        PRINT_DEBUGGING_INFO("VRCPL", VE_VE)

        source = (((tr_u32)rsp_reg.vrcph_source) << 16) | (((tr_u32)rsp_reg.v[__RT][__RS & 0x07]) & 0xffff);

        result = DOUBLE2FIXED(0.4999961852 / FIXED2DOUBLE(source));
        
        if(source > 0xffff)
                result--;
        if(source == 0)   /* special case ? */
                result = 0x7fffffff;

        rsp_reg.vrcph_result         = (result & 0xffff0000) >> 16;
        rsp_reg.v[__SA][__RD & 0x07] =  result & 0x0000ffff;

} /* static void rsp_cop2_vectop_vrcpl() */





static void rsp_cop2_vectop_vrcph()
{
        PRINT_DEBUGGING_INFO("VRCPH", VE_VE)

        rsp_reg.v[__SA][__RD & 0x07] = rsp_reg.vrcph_result;
        rsp_reg.vrcph_source = (tr_u16)rsp_reg.v[__RT][__RS & 0x07];

} /* static void rsp_cop2_vectop_vrcph() */





static void rsp_cop2_vectop_vmov()
{
        PRINT_DEBUGGING_INFO("VMOV", VE_VE)

        rsp_reg.v[__SA][__RD & 0x07] = rsp_reg.v[__RT][__RS & 0x07];

} /* static void rsp_cop2_vectop_vmov() */





static void rsp_cop2_vectop_vrsq()
{
        PRINT_DEBUGGING_INFO("VRSQ", VE_VE)

        PRINT_NOT_IMPLEMENTED("VRSQ")

        RSP_HALT

} /* static void rsp_cop2_vectop_vrsq() */





static void rsp_cop2_vectop_vrsql()
{
        PRINT_DEBUGGING_INFO("VRSQL", VE_VE)

        PRINT_NOT_IMPLEMENTED("VRSQL")

        RSP_HALT

} /* static void rsp_cop2_vectop_vrsql() */





static void rsp_cop2_vectop_vrsqh()
{
        PRINT_DEBUGGING_INFO("VRSQH", VE_VE)

        PRINT_NOT_IMPLEMENTED("VRSQH")

        RSP_HALT

} /* static void rsp_cop2_vectop_vrsqh() */





static void rsp_cop2_vectop_vnoop()
{
        PRINT_DEBUGGING_INFO("VNOOP", V_V_VE)

} /* static void rsp_cop2_vectop_vnoop() */





static void rsp_cop2_vectop_vextt()
{
        PRINT_DEBUGGING_INFO("VEXTT", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VEXTT")

        RSP_HALT

} /* static void rsp_cop2_vectop_vextt() */





static void rsp_cop2_vectop_vextq()
{
        PRINT_DEBUGGING_INFO("VEXTQ", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VEXTQ")

        RSP_HALT

} /* static void rsp_cop2_vectop_vextq() */





static void rsp_cop2_vectop_vextn()
{
        PRINT_DEBUGGING_INFO("VEXTN", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VEXTN")

        RSP_HALT

} /* static void rsp_cop2_vectop_vextn() */





static void rsp_cop2_vectop_vinst()
{
        PRINT_DEBUGGING_INFO("VINST", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VINST")

        RSP_HALT

} /* static void rsp_cop2_vectop_vinst() */





static void rsp_cop2_vectop_vinsq()
{
        PRINT_DEBUGGING_INFO("VINSQ", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VINSQ")

        RSP_HALT

} /* static void rsp_cop2_vectop_vinsq() */





static void rsp_cop2_vectop_vinsn()
{
        PRINT_DEBUGGING_INFO("VINSN", V_V_VE)

        PRINT_NOT_IMPLEMENTED("VINSN")

        RSP_HALT

} /* static void rsp_cop2_vectop_vinsn() */





static void rsp_cop2_vectop_reserved()
{
        PRINT_DEBUGGING_INFO("RESERVED", V_V_VE)

        PRINT_NOT_IMPLEMENTED("RESERVED")

        RSP_HALT

} /* static void rsp_cop2_vectop_reserved() */





