/**
 * TrueReality - N64/rsp_registers.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/* every 'export' in this file resides in 'rsp.c' */
/* this file could also be named as 'rsp_export.h' */





#ifndef __TR_RSP_REGISTERS_H
#define __TR_RSP_REGISTERS_H





#include "type_sizes.h"





/* all "real" values are in rsp.c */

/* rsp registers */

struct rsp_reg
{
        
        tr_s32  r[32];           /* general purpose registers */

        tr_u16  v[32][8];       /* vector registers */
        tr_u16  v_src1[8];      /* dummy source 1 vector registers */
        tr_u16  v_src2[8];      /* dummy source 2 vector registers */
        tr_u16  dummy[8];      /* dummy source 2 vector registers */

        tr_s16  ah[8];          /* high portion of accu */
        tr_s16  al[8];          /* low  portion of accu */
        tr_s16  flag[4];        /* cop2 control registers */

        tr_s16  vrcph_result;       /* calculated by VRCPL - stored by VRCPH */
        tr_s16  vrcph_source;       /* used by VRCPL - stored by VRCPH */



        tr_u32    pc_delay;   /* delayed program counter */
        int     delay;      /* this is for correct emulation of the 6 stage pipeline */
                            /* if(delay ==  0) { next instr is pc + 4 } */
                            /* if(delay ==  1) { next instr is delayed pc} */
                            /* if(delay == -1) { exec delayed pc} */
        
        tr_u32    code;       /* current code (needed for speed) */
        int     halt;       /* if '!0' then stop execution */
        int     do_or_check_sthg; /* if '!0' then the rs4300i has to do or check sthg */
                            /* look at the defines that follow this struct */
        tr_u32    count;      /* just for debugging */

}; /* struct rsp_reg */



/* rsp_reg.delay */

#define NO_RSP_DELAY    0
#define DO_RSP_DELAY    1
#define EXEC_RSP_DELAY  2





extern struct rsp_reg rsp_reg;





#endif /* #ifndef __TR_RSP_REGISTERS_H */
