/**
 * TrueReality - include/file_loader_extern.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#ifndef __TR_FILE_LOADER_EXTERN_H
#define __TR_FILE_LOADER_EXTERN_H





#include <type_sizes.h>

#ifdef TR_STATIC_COMPILER
        #include "../N64/compile_static.h"
#endif




struct tr_rominfo
{       
        char  *real_name;
        char  *dmem_name;
        char  *imem_name;
        
        tr_u16 validation;       /* 0x00 */
        tr_u8  compression;      /* 0x02 */
        tr_u8  unknown1;         /* 0x03 */
        tr_u32  clockrate;        /* 0x04 */
        tr_u32  programcounter;   /* 0x08 */
        tr_u32  release;          /* 0x0c */
        
        tr_u32  crc1;             /* 0x10 */
        tr_u32  crc2;             /* 0x14 */
        tr_u32  unknown2_h;       /* 0x18 */
        tr_u32  unknown2_l;       /* 0x1c */
        
        tr_u8  name[20];         /* 0x20 - 0x33 */
        
        tr_u8  unknown3;         /* 0x34 */
        tr_u8  unknown4;         /* 0x35 */
        tr_u8  unknown5;         /* 0x36 */
        tr_u8  unknown6;         /* 0x37 */
        tr_u8  unknown7;         /* 0x38 */
        tr_u8  unknown8;         /* 0x39 */
        tr_u8  unknown9;         /* 0x3a */
        tr_u8  manufacturerid;   /* 0x3b */
        tr_u16 cartridgeid;      /* 0x3c */
        tr_u8  countrycode;      /* 0x3e */
        tr_u8  unknown10;        /* 0x3f */
};



extern struct tr_rominfo rominfo; 





struct tr_rom
{
        char   *image;                 /* pointer to rom image                 */
#ifdef TR_STATIC_COMPILER
        tr_instruction *image_instr;   /* pointer to prgcode instructions      */
#endif
        long   length;                 /* length of rom image                  */

        char   *header;                /* pointer to header (same as 'image')  */
        char   *bootcode;              /* pointer to bootcode                  */
        tr_u32 *prgcode;               /* pointer to prgcode                   */

        tr_u32 prgcode_base_orig;      /* base of prgcode (from header)        */
        tr_u32 prgcode_base;           /* base of (prgcode & 0x1fffffff)       */
        long   prgcode_length;         /* length of prgcode                    */
};



extern  struct  tr_rom  rom;





#endif /* #ifndef __TR_FILE_LOADER_EXTERN_H */
