/**
 * TrueReality - include/parser_extern.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#ifndef __PARSER_EXTERN_H
#define __PARSER_EXTERN_H





#include <type_sizes.h>





/*** struct section ***/



typedef struct
{
        tr_u32 addr;
        char   *name;

}
tr_stab; /* typedef struct */





struct tr_prefs
{
        /* :) ... a developer need this */
        tr_u64 debug;
        
        
        /* N64 image or DMEM/IMEM binaries */

        tr_s32 load_what;



        /* Symbols */
        
        tr_s32  symbols;
        char    *symbol_file;
        tr_s32  n_symbols;      /* there are n_symbols symbols in table */
        tr_stab *symtab;
        
        

        /* Display */
        
        tr_s32 display;
        
        

        /* Controller/Cartridge */
        
        tr_s32 plugged_controller;
        tr_s32 eeprom;
        
        

        /* X11 */
        
        tr_s32 UseSHM;
        


        /* Joystick */
        
        tr_s32 joystick;
        

        
        /* Sound */
        
        tr_s32 sound;
        char   *sound_file;
        tr_s32 sound_bits;
        tr_s32 sound_freq;
        tr_s32 sound_stereo;



        /* this is for displays in which the depth can be specified */

#ifdef COMMAND_DEPTH
        tr_s32 depth;
#endif



        /* more should come */

        tr_u64 r22;
        
}; /* struct tr_prefs */





/*** extern section ***/

extern struct tr_prefs prefs;





/*** define section ***/

/* tr_prefs.load_what */

#define N64_NOTHING     0       /* These 4 defines are used to    */
#define N64_IMAGE       1       /* determinate                    */
#define N64_IMEM        2       /* if an image or dmem/imem       */
#define N64_DMEM        4       /* is/are to load.                */



/* tr_prefs.symbols */

#define NO_SYMBOLS              0
#define USE_SYMBOLS             1



/* tr_prefs.sound */

#define NO_SOUND                0
#define NORMAL_SOUND            1
#define FILE_SOUND              2



/* tr_prefs.joystick */

#define NO_JOYSTICK             0
#define USE_JOYSTICK            1



/* tr_prefs.display */

#define NO_DISPLAY              0
#define LOCAL_DISPLAY           1
#define LOCAL_DISPLAY_PROCESS   2
#define REMOTE_DISPLAY          3



/* tr_prefs.debug */

#ifndef TR_64BIT_LONG

#define DBG_CPU_INSTR           0x0000000000000001ULL
#define DBG_CPU_COUNT           0x0000000000000002ULL
#define DBG_CPU_REG             0x0000000000000004ULL
#define DBG_NATIVE              0x0000000000000008ULL
#define DBG_RSP_INSTR           0x0000000000000010ULL
#define DBG_RSP_COUNT           0x0000000000000020ULL
#define DBG_RSP_REG             0x0000000000000040ULL
#define DBG_RDP                 0x0000000000000080ULL

#define DBG_INTR_COMP           0x0000000000000100ULL
#define DBG_INTR_VI             0x0000000000000200ULL
#define DBG_INTR_SI             0x0000000000000400ULL
#define DBG_INTR_SP             0x0000000000000800ULL
#define DBG_INTR_PI             0x0000000000001000ULL
#define DBG_INTR_AI             0x0000000000002000ULL
#define DBG_INTR_DP             0x0000000000004000ULL

#define DBG_MEMORY              0x0000000000010000ULL
#define DBG_RD_RAM              0x0000000000020000ULL
#define DBG_RD_REG              0x0000000000040000ULL
#define DBG_SP_DMEM             0x0000000000080000ULL
#define DBG_SP_IMEM             0x0000000000100000ULL
#define DBG_SP_REG              0x0000000000200000ULL
#define DBG_SP_PC_REG           0x0000000000400000ULL
#define DBG_SP_IBIST_REG        0x0000000000800000ULL
#define DBG_DPC_REG             0x0000000001000000ULL
#define DBG_DPS_REG             0x0000000002000000ULL
#define DBG_MI_REG              0x0000000004000000ULL
#define DBG_VI_REG              0x0000000008000000ULL

#define DBG_AI_REG              0x0000000010000000ULL
#define DBG_PI_REG              0x0000000020000000ULL
#define DBG_RI_REG              0x0000000040000000ULL
#define DBG_SI_REG              0x0000000080000000ULL
#define DBG_PI_DOM2_ADDR1       0x0000000100000000ULL
#define DBG_PI_DOM1_ADDR1       0x0000000200000000ULL
#define DBG_PI_DOM2_ADDR2       0x0000000400000000ULL
#define DBG_PI_DOM1_ADDR2       0x0000000800000000ULL
#define DBG_PI_ROM              0x0000001000000000ULL
#define DBG_PI_RAM              0x0000002000000000ULL
#define DBG_PI_DOM1_ADDR3       0x0000004000000000ULL

#define DBG_DISP_REFRESH        0x0000100000000000ULL
#define DBG_DISP_MODE           0x0000200000000000ULL

#define DBG_CONTR_PIF           0x0001000000000000ULL
#define DBG_CONTR_BUTTONS       0x0002000000000000ULL

#define DBG_AUDIO               0x0008000000000000ULL

#define DBG_EXC_OV              0x0010000000000000ULL
#define DBG_EXC_FPE             0x0020000000000000ULL
#define DBG_EXC_TR              0x0040000000000000ULL
#define DBG_EXC_BP              0x0080000000000000ULL
#define DBG_EXC_SYS             0x0100000000000000ULL
#define DBG_EXC_RI              0x0200000000000000ULL
#define DBG_EXC_TLB             0x0400000000000000ULL

#else /* #ifndef TR_64BIT_LONG */

#define DBG_CPU_INSTR           0x0000000000000001UL
#define DBG_CPU_COUNT           0x0000000000000002UL
#define DBG_CPU_REG             0x0000000000000004UL
#define DBG_NATIVE              0x0000000000000008UL
#define DBG_RSP_INSTR           0x0000000000000010UL
#define DBG_RSP_COUNT           0x0000000000000020UL
#define DBG_RSP_REG             0x0000000000000040UL
#define DBG_RDP                 0x0000000000000080UL

#define DBG_INTR_COMP           0x0000000000000100UL
#define DBG_INTR_VI             0x0000000000000200UL
#define DBG_INTR_SI             0x0000000000000400UL
#define DBG_INTR_SP             0x0000000000000800UL
#define DBG_INTR_PI             0x0000000000001000UL
#define DBG_INTR_AI             0x0000000000002000UL
#define DBG_INTR_DP             0x0000000000004000UL

#define DBG_MEMORY              0x0000000000010000UL
#define DBG_RD_RAM              0x0000000000020000UL
#define DBG_RD_REG              0x0000000000040000UL
#define DBG_SP_DMEM             0x0000000000080000UL
#define DBG_SP_IMEM             0x0000000000100000UL
#define DBG_SP_REG              0x0000000000200000UL
#define DBG_SP_PC_REG           0x0000000000400000UL
#define DBG_SP_IBIST_REG        0x0000000000800000UL
#define DBG_DPC_REG             0x0000000001000000UL
#define DBG_DPS_REG             0x0000000002000000UL
#define DBG_MI_REG              0x0000000004000000UL
#define DBG_VI_REG              0x0000000008000000UL

#define DBG_AI_REG              0x0000000010000000UL
#define DBG_PI_REG              0x0000000020000000UL
#define DBG_RI_REG              0x0000000040000000UL
#define DBG_SI_REG              0x0000000080000000UL
#define DBG_PI_DOM2_ADDR1       0x0000000100000000UL
#define DBG_PI_DOM1_ADDR1       0x0000000200000000UL
#define DBG_PI_DOM2_ADDR2       0x0000000400000000UL
#define DBG_PI_DOM1_ADDR2       0x0000000800000000UL
#define DBG_PI_ROM              0x0000001000000000UL
#define DBG_PI_RAM              0x0000002000000000UL
#define DBG_PI_DOM1_ADDR3       0x0000004000000000UL

#define DBG_DISP_REFRESH        0x0000100000000000UL
#define DBG_DISP_MODE           0x0000200000000000UL

#define DBG_CONTR_PIF           0x0001000000000000UL
#define DBG_CONTR_BUTTONS       0x0002000000000000UL

#define DBG_AUDIO               0x0008000000000000UL

#define DBG_EXC_OV              0x0010000000000000UL
#define DBG_EXC_FPE             0x0020000000000000UL
#define DBG_EXC_TR              0x0040000000000000UL
#define DBG_EXC_BP              0x0080000000000000UL
#define DBG_EXC_SYS             0x0100000000000000UL
#define DBG_EXC_RI              0x0200000000000000UL
#define DBG_EXC_TLB             0x0400000000000000UL

#endif /* #else #ifndef TR_64BIT_LONG */



/* tr_prefs.plugged_controller */

#define MAX_CONTR       4      /* BUG: maybe 6 ???*/

#define NOCONTR         0x00
#define CONTR_1         0x01
#define CONTR_2         0x02
#define CONTR_3         0x04
#define CONTR_4         0x08
#define CONTR_5         0x10    /* BUG: ??? */
#define CONTR_6         0x20    /* BUG: ??? */





#endif /* #ifndef __PARSER_EXTERN_H */
