/**
 * TrueReality - include/type_sizes.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/*
 The `TR_64BIT_LONG' stuff works on my alphastation255(ev45)/egcs-2.91.66
 and alphastation255(ev45)/gcc-2.95.1.
 Please report if you have success on another 64bit machine - Niki W. Waibel!
*/





#ifndef __TR_TYPE_SIZES_H
#define __TR_TYPE_SIZES_H





    /* for printing with `printf' stuff */
#ifdef TR_64BIT_LONG
        #define PR64 "l"
        #define PR32 ""
        #define PR16 "h"
        #define PR8  ""
#else
        #define PR64 "ll"
        #define PR32 "l"
        #define PR16 "h"
        #define PR8  ""
#endif





/*
 *   unsigned types
 *
 */

#ifdef TR_64BIT_LONG
        typedef unsigned long      tr_u64;   /* real 64 bit stuff */
        typedef unsigned int       tr_u32;
#else
        typedef unsigned long long tr_u64;   /* simulated (non ansi) 64 bit */
        typedef unsigned long      tr_u32;
#endif
        typedef unsigned short     tr_u16;
#ifdef TR_MACOS
        typedef char               tr_u8;
#else
        typedef unsigned char      tr_u8;
#endif



/*
 *   signed types
 *
 */

#ifdef TR_64BIT_LONG
        typedef signed long      tr_s64;   /* real 64 bit stuff */
        typedef signed int       tr_s32;
#else
        typedef signed long long tr_s64;   /* simulated (non ansi) 64 bit */
        typedef signed long      tr_s32;
#endif
        typedef signed short     tr_s16;
        typedef signed char      tr_s8;





/*
 *   for the static compiler we need this union
 *
 */

typedef union   /* i think that a union is the best solution */
{
    /***\
    *
    *  f: stores a pointer to the emulated instruction
    * rs: source register
    * rt: target register
    * rd: destination register
    * sa: shift amount
    *  i: immediate data
    *  t: target address (28bit (26bit << 2))
    *
    \***/

        struct
        {
                void  (*f)();
                tr_u8 rs;
                tr_u8 rt;
                tr_u8 rd;
                tr_u8 sa;
        }
        r; /* register */

        struct
        {
                void   (*f)();
                tr_u8  rs;
                tr_u8  rt;
                tr_s16 i;
        }
        i; /* immediate */

        struct
        {
                void   (*f)();
                tr_u32 t;
        }
        j; /* jump */

        struct
        {
                void  (*f)();
                tr_u8 fmt;
                tr_u8 ft;
                tr_u8 fs;
                tr_u8 fd;
        }
        f; /* cop1 (fpu) */

        struct
        {
                void   (*f)();
                tr_u32 c;   /* stores the original opcode */
        }
        d; /* debug (for reserved and invalid instructions) */

}
tr_instruction;




#endif /* #ifndef __TR_TYPE_SIZES_H */

