/**
 * TrueReality - input/keyboard/msdos.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include "../../include/config.h"
#include "../../include/parser_extern.h"
#include "../../include/type_sizes.h"
#include "../../include/memory_extern.h"
#include "../../include/dispatch.h"

#include <allegro.h>

#include "keyboard.h"

#ifdef DEBUG
        #include "../../include/debug.h"
#endif





tr_u32 CheckVisual()
{

        static tr_u16    keystate = 0;
        static tr_s8    stick_x = 0;
        static tr_s8    stick_y = 0;


        /* KeyPress and KeyReleases */


        if (key[KEY_ESC])
            dispatch(SHUTDOWN, 0);

        if (key[KEY_F8])
            dispatch(DISPLAY_REFRESH, 0);

        if (key[KEY_Q])    /* Start Button */
            keystate |= 0x1000;

        if (key[KEY_W])    /* A-Button */
            keystate |= 0x8000;

        if (key[KEY_E])    /* B-Button */
            keystate |= 0x4000;

        if (key[KEY_HOME]) /* C-Button UP */
            keystate |= 0x0008;

        if (key[KEY_END])  /* C-Button DOWN */
            keystate |= 0x0004;

        if (key[KEY_DEL])  /* C-Button LEFT */
            keystate |= 0x0002;

        if (key[KEY_PGDN]) /* C-Button RIGHT */
            keystate |= 0x0001;

        if (key[KEY_UP])   /* D-Button UP */
            keystate |= 0x0800;

        if (key[KEY_DOWN]) /* D-Button DOWN */
            keystate |= 0x0400;

        if (key[KEY_LEFT])  /* D-Button LEFT */
            keystate |= 0x0200;

        if (key[KEY_RIGHT]) /* D-Button RIGHT */
            keystate |= 0x0100;

        if (key[KEY_A])     /* R-Button */
            keystate |= 0x0010;

        if (key[KEY_S])     /* L-Button */
            keystate |= 0x0020;

        if (key[KEY_D])     /* Z-Button */
            keystate |= 0x2000;

        if (key[KEY_K])     /* Stick Up */
            if(stick_y < 120)
                stick_y += 0x08;

        if (key[KEY_J])     /* Stick Down */
            if(stick_y > -120)
                stick_y -= 0x08;

        if (key[KEY_L])     /* Stick Right */
            if(stick_x < 120)
                stick_x += 0x08;

        if (key[KEY_H])     /* Stick Left */
            if(stick_x > -120)
                stick_x -= 0x08;


        if (!key[KEY_Q])       /* Start Button */
            keystate &= ~0x1000;

        if (!key[KEY_W])    /* A-Button */
            keystate &= ~0x8000;

        if (!key[KEY_E])    /* B-Button */
            keystate &= ~0x4000;

        if (!key[KEY_HOME]) /* C-Button UP */
            keystate &= ~0x0008;

        if (!key[KEY_END])  /* C-Button DOWN */
            keystate &= ~0x0004;

        if (!key[KEY_DEL])  /* C-Button LEFT */
            keystate &= ~0x0002;

        if (!key[KEY_PGDN]) /* C-Button RIGHT */
            keystate &= ~0x0001;

        if (!key[KEY_UP])   /* D-Button UP */
            keystate &= ~0x0800;

        if (!key[KEY_DOWN]) /* D-Button DOWN */
            keystate &= ~0x0400;

        if (!key[KEY_LEFT])  /* D-Button LEFT */
            keystate &= ~0x0200;

        if (!key[KEY_RIGHT]) /* D-Button RIGHT */
            keystate &= ~0x0100;

        if (!key[KEY_A])     /* R-Button */
            keystate &= ~0x0010;

        if (!key[KEY_S])     /* L-Button */
            keystate &= ~0x0020;

        if (!key[KEY_D])     /* Z-Button */
            keystate &= ~0x2000;


#ifdef DEBUG
        if(prefs.debug & DBG_CONTR_BUTTONS)
                printf
                (
                        "Controller: "
                        "0x%04hx %+4d %+4d (0x%08lx) (0xBBBBXXYY) "
                        "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",

                        keystate, stick_x, stick_y,
                        (tr_u32)(keystate << 16) | ((tr_u32)(tr_u16)(stick_x << 8)) | ((tr_u32)(tr_u8)stick_y),
                        (keystate & 0x8000) ? "A "      : "",
                        (keystate & 0x4000) ? "B "      : "",
                        (keystate & 0x2000) ? "Z "      : "",
                        (keystate & 0x1000) ? "START "  : "",
                        (keystate & 0x0800) ? "Dup "    : "",
                        (keystate & 0x0400) ? "Ddown "  : "",
                        (keystate & 0x0200) ? "Dleft "  : "",
                        (keystate & 0x0100) ? "Dright " : "",
                        (keystate & 0x0080) ? ""        : "",
                        (keystate & 0x0040) ? ""        : "",
                        (keystate & 0x0020) ? "L "      : "",
                        (keystate & 0x0010) ? "R "      : "",
                        (keystate & 0x0008) ? "Cup "    : "",
                        (keystate & 0x0004) ? "Cdown "  : "",
                        (keystate & 0x0002) ? "Cleft "  : "",
                        (keystate & 0x0001) ? "Cright " : ""
                );
#endif

        return(((tr_u32)(keystate << 16)) | ((tr_u32)(tr_u16)(stick_x << 8)) | ((tr_u32)(tr_u8)stick_y));


} /* tr_u32 CheckVisual() */

