/**
 * TrueReality - io.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>

#include "type_sizes.h"
#include "io.h"





static void io_open(tr_u32 where);
static void io_close(tr_u32 where);





static tr_u32 level_console;
static tr_u32 level_file;
static tr_u32 level_window;
static tr_u32 level_serial;
static tr_u8  open_console = 0;
static tr_u8  open_file    = 0;
static tr_u8  open_window  = 0;
static tr_u8  open_serial  = 0;





void tr_io_set(tr_u32 where, tr_u32 level)
{

    /* we have to do the same and the same thing in this
       routine - so i decided to create the following define */

#define TR_IO_SET(this_where, this_level, this_open)    \
        if(where & this_where)                          \
        {                                               \
                this_level = level;                     \
                                                        \
                if(level == TR_IO_NOTHING)              \
                        if(this_open)                   \
                                io_close(this_where);   \
                                                        \
                if(level > TR_IO_NOTHING)               \
                        if(!this_open)                  \
                                io_open(this_where);    \
        }

        TR_IO_SET(TR_IO_CONSOLE, level_console, open_console);
        TR_IO_SET(TR_IO_FILE,    level_file,    open_file);
        TR_IO_SET(TR_IO_WINDOW,  level_window,  open_window);
        TR_IO_SET(TR_IO_SERIAL,  level_serial,  open_serial);

#undef TR_IO_SET

} /* void tr_io_set(tr_u32 where, tr_u32 level) */



static void io_open(tr_u32 where)
{
        switch(where)
        {
            case TR_IO_CONSOLE:
                break;

            case TR_IO_FILE:
                break;

            case TR_IO_WINDOW:
                break;

            case TR_IO_SERIAL:
                break;

        } /* switch(where) */

} /* static void io_open(tr_u32 where); */



static void io_close(tr_u32 where)
{
        switch(where)
        {
            case TR_IO_CONSOLE:
                break;

            case TR_IO_FILE:
                break;

            case TR_IO_WINDOW:
                break;

            case TR_IO_SERIAL:
                break;

        } /* switch(where) */

} /* static void io_close(tr_u32 where); */



int tr_io_printf(tr_u32 level, tr_u32 module, const char *fmt, ...)
{
        va_list ap;
        int     r;



        switch(level)
        {
            case TR_IO_INFO:
                fputs("TR:INFO:", stdout);
                break;

            case TR_IO_WARNING:
                fputs("TR:WARNING:", stdout);
                break;

            case TR_IO_ERROR:
                fputs("TR:ERROR:", stdout);
                break;

            case TR_IO_FATAL:
                fputs("TR:ERROR:", stdout);
                break;

        } /* switch(level) */

        switch(module)
        {
            case TR_IO_CPU:
                fputs("CPU:", stdout);
                break;

            case TR_IO_RSP:
                fputs("RSP:", stdout);
                break;

            case TR_IO_RDP:
                fputs("RDP:", stdout);
                break;

            case TR_IO_HLE:
                fputs("HLE:", stdout);
                break;

            case TR_IO_COMPILE:
                fputs("COMPILE:", stdout);
                break;

            case TR_IO_QUIT:
                fputs("QUIT:", stdout);
                break;

        } /* switch(level) */

        va_start(ap, fmt);
        r = vprintf(fmt, ap);
        va_end(ap);

        return(r);

} /* int tr_io_printf(tr_u32 level, tr_u32 module, char *fmt, ...) */



int tr_io_askf(int how, const char *fmt, ...)
{
        va_list ap;
        int     r;
        int     c;

        fputs("TR:QUESTION:>", stdout);
        va_start(ap, fmt);
        r = vprintf(fmt, ap);
        va_end(ap);
        fputs("< ", stdout);

        switch(how)
        {
            case TR_IO_ASK_YESNO:
                fputs("(Y/n)", stdout);
                break;

            case TR_IO_ASK_NOYES:
                fputs("(N/y)", stdout);
                break;

        } /* switch(how) */
        fputs(" ? ", stdout);
        fflush(stdout);

        switch(how)
        {
            case TR_IO_ASK_YESNO:
            case TR_IO_ASK_NOYES:
                do
                {
                        c = getchar();
                        tolower(c);
                }
                while(c != 'y' && c != 'n' );
                if(c == 'y')
                        r = TR_IO_ASK_YES;
                else
                        r = TR_IO_ASK_NO;

        } /* switch(how) */

        return(r);

} /* int tr_io_askf(int how, const char *fmt, ...) */

