/**
 * TrueReality - native/ROM_scanner.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/










#include <stdio.h>
#include <stdlib.h>

#include <config.h>
#include <ansi_esc_codes.h>
#include <parser_extern.h>
#include <file_loader_extern.h>

#include "../N64/registers.h"

#include <native.h>

#ifdef DEBUG
        #include <debug.h>
#endif





#define NATIVE_SEARCHER(num, name)                                                      \
        fputs("        Searching for '" #name "' ...", stdout); fflush(stdout);         \
                                                                                        \
        native_jal_table[found] = 0xffffffff;                                           \
        native_fkt_table[found] = native_nop;                                           \
                                                                                        \
        length = sizeof(native_code_##name) / 4;                                        \
        for(i=0; (found<=num) && (i<1024*1024/4 - length); i++)                         \
        {                                                                               \
                for(j=0; j<length; j++)                                                 \
                {                                                                       \
                        if(img[i+j] != native_code_##name[j])                           \
                                break;                                                  \
                        else                                                            \
                        {                                                               \
                                if(j == length-1)                                       \
                                {                                                       \
                                        fputs(".", stdout);                             \
                                        native_jal_table[found] = 0x0c000000 | (((4*i - 0x1000 + rom.prgcode_base_orig) >> 2) & 0x03ffffff); \
                                        printf(" FOUND @ $%08lx (jalcode: $%08lx)\n", 4*i - 0x1000 + rom.prgcode_base_orig, native_jal_table[found]); fflush(stdout); \
                                        native_fkt_table[found++] = native_##name;      \
                                }                                                       \
                        }                                                               \
                }                                                                       \
        }




tr_u32    native_jal_table[NUM_OF_NATIVE_ROUTINES];
n_fkt   native_fkt_table[NUM_OF_NATIVE_ROUTINES];





tr_u32 native_code_set_sr[] =
{
        0x40846000,
        0x00000000,
        0x03e00008,
        0x00000000
};






tr_u32 native_code_get_sr[] =
{
        0x40026000,
        0x03e00008,
        0x00000000
};






tr_u32 native_code_disable_interrupts[] =
{
        0x40086000,
        0x2401fffe,
        0x01014824,
        0x40896000,
        0x31020001,
        0x00000000,
        0x03e00008,
        0x00000000
};





tr_u32 native_code_restore_interrupts[] =
{
        0x40086000,
        0x01044025,
        0x40886000,
        0x00000000,
        0x00000000,
        0x03e00008,
        0x00000000
};





tr_u32 native_code_set_vi_mode[] =
{
        0x27bdffd8,
        0xafbf001c,
        0xafa40028,
        0x0c080574,   /* jal     disable_interrupts */
        0xafb00018,
        0x3c0f8020,
        0x8def50ac,
        0x8fae0028,
        0x3c198020,
        0x24180001,
        0xadee0008,
        0x8f3950ac,
        0x3c088020,
        0x00408025,
        0xa7380000,
        0x8d0850ac,
        0x02002025,
        0x8d090008,
        0x8d2a0004,
        0x0c08057c,   /* jal     restore_interrupts */
        0xad0a000c,
        0x8fbf001c,
        0x8fb00018,
        0x27bd0028,
        0x03e00008,
        0x00000000
};







void Create_Native_Table()
{
        int     i, j, length, found = 0;
        tr_u32    *img = (tr_u32 *)rom.image;



        puts("Creating native table: (code by Niki W. Waibel)"); fflush(stdout);

        NATIVE_SEARCHER(0, set_sr)
        NATIVE_SEARCHER(1, get_sr)
        NATIVE_SEARCHER(2, disable_interrupts)
        NATIVE_SEARCHER(3, restore_interrupts)
        native_code_set_vi_mode[3]  = native_jal_table[2];
        native_code_set_vi_mode[19] = native_jal_table[3];
        NATIVE_SEARCHER(4, set_vi_mode)

        puts("-------------------------------------------------------------------------------");
        fflush(stdout);

}





void Destroy_Native_Table()
{

}





