/**
 * TrueReality - display/output/amigados/amigadosMESA.c
 * Copyright (C) 1998, 1999 Mathias "AmiDog" Roslund
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/**
 *
 * All of this stuff comes from AmiDog - great work (niki)
 *
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>

#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>

#include <cybergraphx/cybergraphics.h>
#include <powerup/ppcinline/cybergraphics.h>

#ifdef FPS
        #include <sys/time.h>
        #include <unistd.h>
#endif

#ifdef OPENGL
        #include <GL/Amigamesa.h> /* AmigaOS specific */
#endif

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "../display.h"
#include "../sw_renderer.h"
#include "amigadosMESA.h"
#include "../../../N64/registers.h"
#include "../../../N64/rdp_registers.h"
#include "../../../N64/draw2d_sw.h"

#ifdef OPENGL
        struct amigamesa_context *amesa=NULL;
#endif

/* Scaling */
static float stretch_x;
static float stretch_y;

/* Original Variables */
static tr_u32 screenwidth;
static tr_u32 screenheight;

static tr_u32 width = 0;
static tr_u32 height = 0;
static int WIDTH=640;
static int HEIGHT=480;

/* Local functions */
static int InitDisplay();
static int DestroyDisplay();
static void PrintText(char *text, int x, int y);
static void PutImage(int x, int y, int w, int h);

/* Amiga specific variables */
struct IntuitionBase *IntuitionBase;
struct GfxBase *GfxBase;
struct Library *CyberGfxBase;
struct Screen *DispScreen;
struct Window *DispWindow;

/* CGFX variables */
unsigned long pixfmt;
unsigned char *baseaddress;
int bpr;

static char *rdp_address;
static int rdp_format;
static int rdp_size;
static int rdp_width;

int open_libraries()
{
  IntuitionBase = (struct IntuitionBase *) OpenLibrary("intuition.library", 39);

  if(IntuitionBase==NULL)
    return -1;

  GfxBase = (struct GfxBase *) OpenLibrary("graphics.library", 39);

  if(GfxBase==NULL)
  {
    CloseLibrary(IntuitionBase);
    return -1;
  }

  CyberGfxBase = OpenLibrary("cybergraphics.library", 40);

  if(CyberGfxBase==NULL)
  {
    CloseLibrary(GfxBase);
    CloseLibrary(IntuitionBase);
    return -1;
  }

  return 0;
}

void close_libraries()
{
  CloseLibrary(CyberGfxBase);
  CloseLibrary(GfxBase);
  CloseLibrary(IntuitionBase);
}

void close_screen()
{
  /* Free MESA */
  if (amesa!=NULL)
    AmigaMesaDestroyContext(amesa);

  CloseWindow(DispWindow);
  CloseScreen(DispScreen);
}

int open_screen()
{
  unsigned long VideoMode;
  int width, height, depth;

  /* Ask the user for a screenmode */
  VideoMode = CModeRequestTags(NULL, TAG_DONE);

  if (VideoMode==0)
  {
    printf("You MUST choose a screenmode!\n");
    close_libraries();
    return -1;
  }

  /* Get VideoMode information */
  width=GetCyberIDAttr(CYBRIDATTR_WIDTH, VideoMode);
  height=GetCyberIDAttr(CYBRIDATTR_HEIGHT, VideoMode);
  depth=GetCyberIDAttr(CYBRIDATTR_DEPTH, VideoMode);

  /* Open screen */
  DispScreen = OpenScreenTags(NULL,
               SA_Width,width,
               SA_Height,height,
               SA_Depth,depth,
               SA_Quiet,TRUE,
               SA_ShowTitle,FALSE,
               SA_Type,CUSTOMSCREEN,
               SA_DisplayID,VideoMode,
               TAG_DONE);

  /* Could the screen be opened? */
  if (DispScreen == NULL)
  {
    close_libraries();
    printf("Could NOT open screen!\n");
    return -1;
  }

  /* Open window */
  DispWindow = OpenWindowTags(NULL,
               WA_CustomScreen,(int)DispScreen,
               WA_Width,DispScreen->Width,
               WA_Height,DispScreen->Height,
               WA_IDCMP,IDCMP_RAWKEY,
               WA_Backdrop,TRUE,
               WA_Borderless,TRUE,
               WA_Activate,TRUE,
               TAG_DONE);

  /* Could the window be opened? */
  if(DispWindow == NULL)
  {
    CloseScreen(DispScreen);
    close_libraries();
    printf("Could NOT open window!\n");
    return -1;
  }

  /* Get screenmode information */
  UnLockBitMap(LockBitMapTags(DispScreen->RastPort.BitMap,
                              LBMI_PIXFMT, (ULONG)&pixfmt,
                              LBMI_BASEADDRESS, (ULONG)&baseaddress,
                              LBMI_BYTESPERROW, (ULONG)&bpr,
                              TAG_END));

  /* Init MESA */
  amesa=AmigaMesaCreateContextTags(AMA_Left, 0,
                                   AMA_Bottom, 0,
                                   AMA_Width, DispScreen->Width,
                                   AMA_Height, DispScreen->Height,
                                   //AMA_Window, DispWindow,
                                   AMA_RastPort, &DispScreen->RastPort,
                                   AMA_Screen, DispScreen,
                                   AMA_DoubleBuf, TRUE,
                                   AMA_RGBMode, TRUE,
                                   AMA_AlphaFlag, FALSE,
                                   AMA_Forbid3DHW, FALSE,
                                   AMA_Fullscreen, TRUE,
                                   AMA_DirectRender, TRUE,
                                   AMA_TwoBuffers, FALSE,
                                   AMA_Visual, NULL,
                                   AMA_Fast, FALSE,
                                   AMA_VeryFast, FALSE,
                                   AMA_NoDepth, TRUE,
                                   AMA_NoStencil, TRUE,
                                   AMA_NoAccum, TRUE,
                                   TAG_END);

  AmigaMesaMakeCurrent(amesa, amesa->buffer);

  /* Init variables */
  screenwidth=width;
  screenheight=height;
  WIDTH  = screenwidth;
  HEIGHT = screenheight;
  stretch_x = (float)WIDTH/width;
  stretch_y = (float)HEIGHT/((width*3)>>2);

  /* Clear the screen */
  FillPixelArray(amesa->front_rp, 0, 0, screenwidth, screenheight, 0x00000000);
  FillPixelArray(amesa->back_rp, 0, 0, screenwidth, screenheight, 0x00000000);

  SetAPen(amesa->back_rp, 0);
  SetBPen(amesa->back_rp, 1);

  return 0;
}



#ifdef OPENGL
#include "../gl_renderer.h"   /* OpenGL stuff */
#endif



/*******************************************************************************
*                                                                              *
* int OpenVisual()                                                             *
* int CloseVisual()                                                            *
*                                                                              *
********************************************************************************
*                                                                              *
* These routines are used by the emulator to open and to close a visual.       *
*                                                                              *
* Both return a negative int on error and zero on success.                     *
*                                                                              *
*******************************************************************************/

int OpenVisual()
{
  if(InitDisplay() < 0)
    exit(20);

  prefs.display = LOCAL_DISPLAY;

  return(0);

} /* int OpenVisual() */

int CloseVisual()
{
  DestroyDisplay();

  prefs.display = NO_DISPLAY;

  return(0);

} /* int CloseVisual() */

/*******************************************************************************
*                                                                              *
* static int InitDisplay()                                                     *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to initialize the visual.                               *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int OpenVisual()'.  *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/

static int InitDisplay()
{
  printf("Initializing AmigaOS display: (by Mathias \"AmiDog\" Roslund)\n");

  printf("Opening libraries ... ");
  if (open_libraries()<0)
    return -1;
  else
    printf("OK\n");

  printf("Opening screen ... ");
  if (open_screen()<0)
    return -1;
  else
    printf("OK\n");

  printf("-------------------------------------------------------------------------------\n");

  return(0);

} /* static int InitDisplay() */

/*******************************************************************************
*                                                                              *
* static int DestroyDisplay()                                                  *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to destroy the visual.                                  *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int CloseVisual()'. *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/

static int DestroyDisplay()
{
  printf("-------------------------------------------------------------------------------\n");
  printf("Destroying AmigaOS display:\n");

  printf("Closing screen ... ");
  close_screen();
  printf("OK\n");

  printf("Closing libraries ... ");
  close_libraries();
  printf("OK\n");

  printf("-------------------------------------------------------------------------------\n");

  return(0);

} /* static int DestroyDisplay() */


static void PrintText(char *text, int x, int y)
{
  Move(amesa->back_rp, x, y);
  Text(amesa->back_rp, text, strlen(text));
}


void RefreshVisual()
{}


void GetVisualName(char *name, int *n)
{
  strncpy(name, DISPLAYNAME, (size_t) *n);
  *n = strlen(DISPLAYNAME);
}


static void PutImage(int X, int Y, int W, int H)
{}


void SetVisualTitle(char *title)
{}


void SetVisualSize(int w, int h)
{
        glViewport(0, 0, w, h);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glOrtho(0, width, height, 0, -1, 1);
}


void InitVisualRenderBuffer()
{
        width = ((tr_u32 *)mem.vi_reg)[2];
        height = (width*3)>>2;
        WIDTH  = screenwidth;
        HEIGHT = screenheight;
        stretch_x = (float)WIDTH/width;
        stretch_y = (float)HEIGHT/((width*3)>>2);

#ifndef JUST3D
        DrawScreenGL(width, height);
#endif
        /* from FiRES and Gil */
        glOrtho(0, width, height, 0, -1, 1);
        glViewport(0, 0, WIDTH, HEIGHT);

        /* set viewer */
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glFrustum(-4.0, 4.0, -2.704225, 2.74225, 4.0, 4096.0);

        /* position viewer */
        glMatrixMode(GL_MODELVIEW);

        glCullFace(GL_BACK);
        glFrontFace(GL_CCW);
        glDisable(GL_CULL_FACE);

        glDepthFunc(GL_LEQUAL);
        //glEnable(GL_DEPTH_TEST);

        //speed!
        //glShadeModel(GL_SMOOTH);
        //glShadeModel(GL_FLAT);
    /* (niki) next 2 functions are a speed problem on slow machines */
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    /* (niki)if we enable this here we cannot switch it off later. why the hell? */
        /*glEnable(GL_BLEND);*/

        glAlphaFunc(GL_GEQUAL, 0.07f);
        glDisable(GL_ALPHA_TEST);

        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    /* (niki) this caused a seg fault */
    /*
        glEnable(GL_TEXTURE_2D);
    */

        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

   /* from FiRES and Gil end */
}


void SetVisualRenderBuffer(int format, int size, int width, char *address)
{
    /* format and size cannot be supported in X (as far as i know)
       we store it here ... maybe we can simulate it in FlushVisualRenderBuffer */
        rdp_format  = format;
        rdp_size    = size;
        rdp_width   = width;
        rdp_address = address;
}


void FlushVisualRenderBuffer()
{
#ifdef JUST3D
        /* Swap buffers */
        AmigaMesaSwapBuffers(amesa);

        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glClearDepth(1.0);
        glClear(GL_COLOR_BUFFER_tr_s64 | GL_DEPTH_BUFFER_tr_s64);
#endif
}


void SetVisualClipRectangle(int ulx, int uly, int lrx, int lry)
{}

