/**
 * TrueReality - output/display/display.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#ifndef __TR_DISPLAY_H
#define __TR_DISPLAY_H





#include <type_sizes.h>





int  tr_display_open();      /* should open a window/screen     */
int  tr_display_close();     /* should close that window/screen */

void tr_display_refresh();   /* should redraw the window/screen
                                (this is called whenever
                                a redraw is needed)             */


void GetVisualName(char *, int *n);    /* guess! :)        */
void SetVisualTitle(char *);           /* guess! :)        */
void SetVisualSize(int w, int h);      /* guess! :)        */





/**
*
* for RDP
*
**/



/* The next 3 routines are for the render buffer.
   If you can draw directly to RDRAM then do it (do nothing in this routines)!
   In standard X you cannot (or maybe i just don't know how). */

void InitVisualRenderBuffer();
/*
 * Called when gfx HLE is started.
 */

void FlushVisualRenderBuffer();
/*
 * Called whenever the renderbuffer is set to another address.
 */

void SetVisualRenderBuffer(int format, int size, int width, char *address);
/*
 * format:  0 rgba, 1 yuv, 2 ci, 3 ia, 4 i   ... i think that just rgba is used.
 * size:    0 4bit, 1 8bit, 2 16bit, 3 32bit ... i think that just 16 and 32 bit are used.
 * width:   You might ask where is the height ... i don't know. i use height=width*3/4
 * address: This also could be a parameter for FlushVisualRenderBuffer().
 *          It says where to draw the buffer after rendering (when FlushVisualRenderBuffer() occurs).
 */


void Visual3D_FPS();



/* Sets the clipping rectangle.
   This routine is called from SETSCISSOR. */
void SetVisualClipRectangle(int ulx, int uly, int lrx, int lry);



/* Sets the foreground color for the point, line, tri and rect routines. */

void SetVisualColor(int color);

void ClearVisual();

void DrawVisualFillRect(int ulx, int uly, int lrx, int lry);

void DrawVisualLine(int ulx, int uly, int lrx, int lry);

void DrawVisualTriangle(int x1, int y1, int x2, int y2, int x3, int y3);

void DrawVisualRectangle(int ulx, int uly, int lrx, int lry);

/* obsolete - now done in DrawVisualTexRectangle()/-Triangle()
void LoadVisualBlock(int uls, int ult, int lrs, int dxt, int tile);
*/
/*
 * Loads a texture into tmem - later we should buffer the textures.
 *
 * uls, ult: always 0? this is used for offset
 * lrs: lrs - uls + 1 = pixel of tile
 * dxt: used with offset.
 *      look at the source in x11/x11.c how i have used it.
 * tile: 0..7
 */

/* obsolete - now done in DrawVisualTexRectangle()/-Triangle()
void LoadVisualTile(int uls, int ult, int lrs, int lrt, int tile);
*/
/*
 * Loads a texture into tmem - later we should buffer the textures.
 */

void DrawVisualTexRectangle(
        int ulx, int uly,
        int lrx, int lry,
        int tile,
        float s, float t, float dsdx, float dtdy
        );

void DrawVisualTexTriangle(int vn[3]);

#ifdef OPENGL
        void EnableVisualAlpha();
        void DisableVisualAlpha();
        void VisualFlat();
        void VisualSmooth();
        void DrawScreenGL();   /* from Gil */
#endif





#endif /* #ifndef __TR_DISPLAY_H */

