#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <gl.h>

#include "config.h"
#include "type_sizes.h"
#include "memory_extern.h"
#include "rdp_registers.h"

/* This will draw the current n64 frame buffer to the screen */
/* At the moment only 15 bit graphics at a resolution of 320*240 is supported. */
/* It probably only works right with big endian at the momment. */
/* This routine was coded by Gil Pedersen */
void DrawScreenGL(int width, int height)
{
	int i, j;
	static unsigned char	tbuffer[256*256*4];
	unsigned short			*src, *base_src_ptr;
	unsigned char			*dest;
	GLint					hadDepthTest, hadBlending;
	
	glGetIntegerv(GL_DEPTH_TEST, &hadDepthTest); /* Should be glGetBooleanv, but there's a problem with it on macos compiler */
	glDisable(GL_DEPTH_TEST);
	glGetIntegerv(GL_BLEND, &hadBlending); /* Should be glGetBooleanv, but there's a problem with it on macos compiler */
	glDisable(GL_BLEND);
	
    glMatrixMode(GL_PROJECTION);
	glPushMatrix();
    glLoadIdentity();
	glOrtho(0, width, height-1.2, 0, -1, 1);
    glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
    glLoadIdentity();

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glEnable(GL_TEXTURE_2D);
  
	base_src_ptr = (unsigned short *)((tr_u32  *)(mem.rd_ram + ((tr_u32 *)mem.vi_reg)[1]));
	
	/* Just to be sure */	
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
    glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

	if (width == 320)
	{
		// First a 256*240 rectangle
		for(i=0; i<240; i++)
		{
			for(j=0; j<256; j++)
			{
				src = base_src_ptr + i*320 + j;
				dest = (unsigned char*)tbuffer + (i*256+j)*4;
				*(dest)   = (((*src >> 11 ) & 0x001f)<<3);
				*(dest+1) = (((*src >> 6 ) & 0x001f)<<3);
				*(dest+2) = (((*src >> 1 ) & 0x001f)<<3);
			}
		}
		
	    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, tbuffer);
		
		glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(0.0, 1.0); glVertex2f(  0.0, 256.0);
		glTexCoord2f(0.0, 0.0); glVertex2f(  0.0,   0.0);
		glTexCoord2f(1.0, 1.0); glVertex2f(256.0, 256.0);
		glTexCoord2f(1.0, 0.0); glVertex2f(256.0,   0.0);
		glEnd();
		
		// Then the last part, which is a 64*240 rectangle
		base_src_ptr += 256;
		
		for(i=0; i<240; i++)
		{
			for(j=0; j<64; j++)
			{
				src = base_src_ptr + i*320 + j;
				dest = (unsigned char*)tbuffer + (i*64+j)*4;
				*(dest)   = (((*src >> 11 ) & 0x001f)<<3);
				*(dest+1) = (((*src >> 6 ) & 0x001f)<<3);
				*(dest+2) = (((*src >> 1 ) & 0x001f)<<3);
			}
		}
	    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 64, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, tbuffer);
		
		glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(0.0, 1.0); glVertex2f(256.0, 256.0);
		glTexCoord2f(0.0, 0.0); glVertex2f(256.0,   0.0);
		glTexCoord2f(1.0, 1.0); glVertex2f(320.0, 256.0);
		glTexCoord2f(1.0, 0.0); glVertex2f(320.0,   0.0);
		glEnd();
	}
	glDisable(GL_TEXTURE_2D);

	glPopMatrix();
    glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	if (hadBlending)
		glEnable(GL_BLEND);
	if (hadDepthTest)
		glEnable(GL_DEPTH_TEST);
}
