/**
 * TrueReality - Display/local/x11/x11.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

int cnt=0;



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <ncurses.h>
#include <vgakeyboard.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "../warp_resolution_and_depth.h"
#include "ncurses.h"

#ifdef DEBUG
        #include <debug.h>
#endif


// static char               title[] = "TrueReality from \"Niki W. Waibel\" and many others";
static tr_u32               width  =   0;   /* initial VI_WIDTH_REG (0x4400008) value */
static tr_u32               height =   0;   /* width*3/4 */
#undef DEPTH
static int                DEPTH=3;
 int HEIGHT,WIDTH;
 WINDOW *win;
static char *Buf;
 static tr_u32               flags  = 0;



static  int     InitDisplay();
static  int     DestroyDisplay();
void		RefreshVisual();

static  void    PrintText(char *text, int x, int y);

#ifdef DEBUG
        static  void PrintDisplayMode(tr_u32 flags);
#endif

static  void    PutImage(int x, int y, int w, int h);
        
static  void    AllocDisplayMem();
static  void    FreeDisplayMem();










/*******************************************************************************
*                                                                              *
* int OpenVisual()                                                             *
* int CloseVisual()                                                            *
*                                                                              *
********************************************************************************
*                                                                              *
* These routines are used by the emulator to open and to close a visual.       *
*                                                                              *
* Both return a negative int on error and zero on success.                     *
*                                                                              *
*******************************************************************************/



int OpenVisual()
{
        if(InitDisplay() < 0)
                return(-1);

        prefs.display = LOCAL_DISPLAY;

        return(0);

} /* int OpenVisual() */





int CloseVisual()
{
        DestroyDisplay();

        prefs.display = NO_DISPLAY;

        return(0);

} /* int CloseVisual() */










/*******************************************************************************
*                                                                              *
* static int InitDisplay()                                                     *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to initialize the visual.                               *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int OpenVisual()'.  *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/





static int InitDisplay()
{
        printf("Initializing ncurses display: (by \"Guenther Sohler\")\n");
   

        printf("Opening display ...\n"); fflush(stdout);

        puts("trying ncurses\n");fflush(stdout);
        win=initscr();
        getmaxyx(win,HEIGHT,WIDTH);
        puts("Allocating Display memory ...");
        Buf= malloc(WIDTH*HEIGHT);

        if (Buf == NULL)
        {
         puts("failed!\n");
         return -1;
        }         
        puts("done\n");

        if (has_colors())
        {
         start_color();
         init_pair(1,COLOR_BLACK,COLOR_BLACK);
         init_pair(2,COLOR_BLUE,COLOR_BLACK);
         init_pair(3,COLOR_GREEN,COLOR_BLACK);
         init_pair(4,COLOR_CYAN,COLOR_BLACK);
         init_pair(5,COLOR_RED,COLOR_BLACK);
         init_pair(6,COLOR_MAGENTA,COLOR_BLACK);
         init_pair(7,COLOR_YELLOW,COLOR_BLACK);
         init_pair(8,COLOR_WHITE,COLOR_BLACK);
        } 
        cbreak();
        noecho();
        nonl();
        DEPTH=3;
        clear();
        RefreshVisual();
        keyboard_init();
        return(0);
   
} /* static int InitDisplay() */




/*******************************************************************************
*                                                                              *
* static int DestroyDisplay()                                                  *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to destroy the visual.                                  *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int CloseVisual()'. *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/





static int DestroyDisplay()
{
 keyboard_clearstate();
 keyboard_close();
 endwin();
 free(Buf);
 printf(" OK\n");
 printf("-------------------------------------------------------------------------------\n"); fflush(stdout);
 return(0);
} /* static int DestroyDisplay() */






static void PrintText(char *text, int x, int y)
{
 if (width == 0 || height == 0)
 {
   attrset(A_BOLD | COLOR_PAIR(8));
   mvprintw(0,0,"%s",text);
   return;
 } 
 attrset(A_BOLD | COLOR_PAIR(8));
 mvprintw(x*WIDTH/width,y*HEIGHT/height,"%s",text);
} /* static void PrintText(char *text, int x, int y) */






#ifdef DEBUG

static void PrintDisplayMode(tr_u32 flags)
{
        printf("Displaymode:\n");
   

   
        switch(flags & 0x0003)
        {
            case 0:
                printf("   blank (no data, no sync)\n");
                break;

            case 1:
                printf("   reserved\n");
                break;

            case 2:
                printf("   5/5/5/3 (\"16\" bit)\n");
                break;
         
            case 3:
                printf("   8/8/8/8 (32 bit)\n");
                break;
         
        } /* switch(flags & 0x0003) */
   

   
        if(flags & 0x0004)
                printf("   gamma_dither_enable\n");
        if(flags & 0x0008)
                printf("   gamma_enable\n");
        if(flags & 0x0010)
                printf("   divot_enable\n");
        if(flags & 0x0040)
                printf("   serrate\n");
   

   
        printf("   anti-alias (aa) mode\n");
   
        switch((flags & 0x0300) >> 8)
        {
            case 0:
                printf("      aa & resamp (always fetch extra lines)\n");
                break;

            case 1:
                printf("      aa & resamp (fetch extra lines if needed)\n");
                break;

            case 2:
                printf("      resamp only (treat as all fully covered)\n");
                break;
         
            case 3:
                printf("      neither (replicate pixels, no interpolate)\n");
                break;
         
        } /* switch((flags & 0x0300) >> 8) */
   
} /* static void PrintDisplayMode(tr_u32 flags) */

#endif





void RefreshVisual()
{
 tr_u32    new_width;
 tr_u8    *Imsrc1, *Imdst1;
 tr_u16   *Imsrc2;
 tr_u32    *Imsrc4;
 flags = ((tr_u32 *)mem.vi_reg)[0];
 keyboard_update();
 if (keyboard_keypressed(SCANCODE_ESCAPE))
 {
  dispatch(SHUTDOWN,0);
 }
#ifdef DEBUG
        if(prefs.debug & DBG_DISP_REFRESH)
                puts("DBG_DISP_REFRESH: refresh");
        if(prefs.debug & DBG_DISP_MODE)
                PrintDisplayMode(flags);
#endif
   
   
        /* video width */
        /* if width in VI_WIDTH_REG (0x04400008) has changed set windows width */
   
        new_width = ((tr_u32 *)mem.vi_reg)[2];
        if(new_width != width)
        {
                if(new_width < 320 || new_width > 640)
                        width = 640;
                else
                        width = new_width;

                height = (width * 3) >> 2;

        }
        /* display it! */

        if( new_width >= 320 && new_width <= 640)
        {
                if(flags & 0x0003)
                {
         
                        /* source video origin */
                        Imsrc4 = (tr_u32  *) (mem.rd_ram + ((tr_u32 *)mem.vi_reg)[1]);
                        Imsrc2 = (tr_u16 *) Imsrc4;
                        Imsrc1 = (tr_u8  *) Imsrc2;
         
                        /* destination video origin */
                        Imdst1 = (char *) Buf;
/* everything in the following switch comes from:  */
/* Guenther Sohler <guenther.sohler@newlogic.at>   */
/* HEY, GREAT WORK! (niki.waibel@gmx.net)          */
                        switch(DEPTH)
                        {
                            case 3:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;
                                    
                                    case 2:   /* local: 3bit   n64: 5/5/5/3 (16bit) */
                                        if ( mem.vi_reg[1]+height*width*2 < 0x400000)
                                        {
                                         render_image_16_to_06(Imsrc2,Imdst1,width,height,WIDTH,HEIGHT);
                                        }        
                                        break;
                                    
                                    case 3:   /* local: 8bit   n64: 8/8/8/8 (32bit) */
                                        if (  mem.vi_reg[1]+height*width*4 < 0x400000)
                                        {
                                         render_image_32_to_06(Imsrc4,Imdst1,width,height,WIDTH,HEIGHT);
                                        }        
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;
            
                            default:
                                fprintf(stderr, "Display depth of %d bit is not supported! (yet ;) )\n", DEPTH);
                                
                        } /* switch(DEPTH) */
         
                        PutImage(0, 0, WIDTH, HEIGHT);
                }
                else
                {
                        PrintText("DISPLAY: blank (no data, no sync)", 20, 20);
                }
        }
        else
        {
                if(!(flags & 0x0003))
                        PrintText("DISPLAY: blank (no data, no sync)", 20, 20);
         
                PrintText("DISPLAY: wrong width (640 will be used!)", 20, 40);
        }
        refresh();
   
} /* void RefreshVisual() */



void GetVisualName(char *name, int *n)
{
   strncpy(name, DISPLAYNAME, (size_t) *n);
   
   *n = strlen(DISPLAYNAME);
   
} /* void GetVisualName() */





static void PutImage(int X, int Y, int W, int H)
{
 int i,j;
 char c;
 char col;
 if (width == 0 || height == 0)
 {
  return;
 }
 for (j=Y;j<Y+H;j++)
 {
  setsyx(j,X);
  for (i=X;i<X+W;i++)
  {
   c=Buf[j*WIDTH+i];
   col=((c&0x20)>>3) | ((c&0x08)>>2) | ((c&0x02)>>1) ;
   attrset(A_BOLD | COLOR_PAIR(col+1));
   
   addch(32+c);
  }
 }
} /* static void PutImage(int X, int Y, int W, int H) */





static void FreeDisplayMem()
{
} /* static void FreeDisplayMem() */





static void AllocDisplayMem()
{
} /* static void AllocDisplayMem() */


/* for RDP */

void SetVisualClipRect(int x, int y, int width, int height)
{}

void DrawVisualFillRect(int ulx, int uly, int lrx, int lry, int color)
{}
