/**
 * TrueReality - output/display/warp_resolution_and_depth.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/*
*
*  All ideas from: Guenther Sohler
*
*  The 15/24/32bit and all endian stuff were hacked out by myself (Niki W.
*  Waibel) using the 8/16bit stuff from Guenther Sohler as a bases.
*  HEY GUE, GREAT WORK!
*
*/





#include <stdio.h>

#include <config.h>
#include <type_sizes.h>
#ifdef X11_DISPLAY
        #include <x11_extern.h>
#endif
#include "warp_resolution_and_depth.h"





void
render_image_16_to_06(tr_u16 *imsrc2, tr_u8 *imdst1, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;
        tr_u8    c;

        for(j=0, i=0; i<height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for (l=0, k=0; k<width; k++)
                        {
#ifdef TR_ENDIAN_IS_LITTLE
                                ui = imsrc2[(i*width+k) ^ 0x01];
#endif
#ifdef TR_ENDIAN_IS_BIG
                                ui = imsrc2[(i*width+k)];
#endif
                                c  = ((ui & 0x30) >> 4) | ((ui & 0x600) >> 7) | ((ui & 0xc000) >> 10);

                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst1[j*WIDTH+l] = c;
                                }
                        }
                }
        }
}





void
render_image_16_to_15(tr_u16 *imsrc2, tr_u16 *imdst2, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;

        for(j=0, i=0; i<height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
#ifdef TR_ENDIAN_IS_BIG
                                ui = imsrc2[i*width+k];
#endif
#ifdef TR_ENDIAN_IS_LITTLE
                                ui = imsrc2[(i*width+k)^0x01];
#endif
                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst2[j*WIDTH+l] = ui >> 1;
                                }
                        }
                }
        }
}





/*
void
render_image_16_to_16(tr_u16 *imsrc2, tr_u16 *imdst2, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;

        for(j=0, i=0; i<height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
#ifdef TR_ENDIAN_IS_BIG
                                ui = imsrc2[i*width+k];
#endif
#ifdef TR_ENDIAN_IS_LITTLE
                                ui = imsrc2[(i*width+k)^0x01];
#endif
                                ui = (ui & 0xffc0) | ((ui >> 1) & 0x1f);

                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst2[j*WIDTH+l] = ui;
                                }
                        }
                }
        }
}
*/





void render_image_16_to_16(tr_u16 *imsrc2, tr_u16 *imdst2, int width, int height, int WIDTH, int HEIGHT)
{
/* this code is alpha optimized!!! */

        char * p_end = (char *)imdst2 + 153600;
        register tr_u64  ui1;
        register tr_u64  ui2;
        register tr_u64  ui3;
        register tr_u64  ui4;
        register tr_u64  ui5;
        register tr_u64  ui6;
        register tr_u64  ui7;
        register tr_u64  ui8;
/*
        tr_u32  start, stop;
*/
        register const tr_u64 mask1 = 0x0000ffff0000ffffUL;
        register const tr_u64 mask2 = 0x003e003e003e003eUL;
        register const tr_u64 mask3 = 0xffc0ffc0ffc0ffc0UL;

/*
        start = tr_perf_virtcc();
*/

        for( ; (char *)imdst2 < p_end; )
        {
#ifdef TR_ENDIAN_IS_BIG
                                /*ui = imsrc2[];*/
#endif
#ifdef TR_ENDIAN_IS_LITTLE
                    /* load */
                        ui1 = ((tr_u64 *)imsrc2)[0];
                        ui2 = ((tr_u64 *)imsrc2)[1];
                        ui3 = ((tr_u64 *)imsrc2)[2];
                        ui4 = ((tr_u64 *)imsrc2)[3];
                        ui5 = ((tr_u64 *)imsrc2)[4];
                        ui6 = ((tr_u64 *)imsrc2)[5];
                        ui7 = ((tr_u64 *)imsrc2)[6];
                        ui8 = ((tr_u64 *)imsrc2)[7];
                        imsrc2 = (tr_u16 *)(((char *)(imsrc2)) + 64);
                    /* swap */
                        ui1 = ( (ui1 & (mask1 <<  0)) << 16 ) |
                              ( (ui1 & (mask1 << 16)) >> 16 ) ;
                        ui1 = ( ui1 & mask3 ) | ( (ui1 & mask2) >>  1 ) ;
                        ui2 = ( (ui2 & (mask1 <<  0)) << 16 ) |
                              ( (ui2 & (mask1 << 16)) >> 16 ) ;
                        ui2 = ( ui2 & mask3 ) | ( (ui2 & mask2) >>  1 ) ;
                        ui3 = ( (ui3 & (mask1 <<  0)) << 16 ) |
                              ( (ui3 & (mask1 << 16)) >> 16 ) ;
                        ui3 = ( ui3 & mask3 ) | ( (ui3 & mask2) >>  1 ) ;
                        ui4 = ( (ui4 & (mask1 <<  0)) << 16 ) |
                              ( (ui4 & (mask1 << 16)) >> 16 ) ;
                        ui4 = ( ui4 & mask3 ) | ( (ui4 & mask2) >>  1 ) ;
                        ui5 = ( (ui5 & (mask1 <<  0)) << 16 ) |
                              ( (ui5 & (mask1 << 16)) >> 16 ) ;
                        ui5 = ( ui5 & mask3 ) | ( (ui5 & mask2) >>  1 ) ;
                        ui6 = ( (ui6 & (mask1 <<  0)) << 16 ) |
                              ( (ui6 & (mask1 << 16)) >> 16 ) ;
                        ui6 = ( ui6 & mask3 ) | ( (ui6 & mask2) >>  1 ) ;
                        ui7 = ( (ui7 & (mask1 <<  0)) << 16 ) |
                              ( (ui7 & (mask1 << 16)) >> 16 ) ;
                        ui7 = ( ui7 & mask3 ) | ( (ui7 & mask2) >>  1 ) ;
                        ui8 = ( (ui8 & (mask1 <<  0)) << 16 ) |
                              ( (ui8 & (mask1 << 16)) >> 16 ) ;
                        ui8 = ( ui8 & mask3 ) | ( (ui8 & mask2) >>  1 ) ;
                    /* store */
                        ((tr_u64 *)imdst2)[0] = ui1;
                        ((tr_u64 *)imdst2)[1] = ui2;
                        ((tr_u64 *)imdst2)[2] = ui3;
                        ((tr_u64 *)imdst2)[3] = ui4;
                        ((tr_u64 *)imdst2)[4] = ui5;
                        ((tr_u64 *)imdst2)[5] = ui6;
                        ((tr_u64 *)imdst2)[6] = ui7;
                        ((tr_u64 *)imdst2)[7] = ui8;
                        imdst2 = (tr_u16 *)(((char *)(imdst2)) + 64);
#endif
        }

/*
        stop = tr_perf_virtcc();
        printf("cycles: %u\n", stop - start);
*/

} /* 16 -> 16 */





void
render_image_16_to_24(tr_u16 *imsrc2, tr_u8 *imdst1, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;

        for(j=0, i=0; i<height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
#ifdef TR_ENDIAN_IS_LITTLE
                                ui = imsrc2[(i*width+k)^0x01];
#endif
#ifdef TR_ENDIAN_IS_BIG
                                ui = imsrc2[i*width+k];
#endif
                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst1[3*(j*WIDTH+l)+0] = (ui << 2) & 0xf8;
                                        imdst1[3*(j*WIDTH+l)+1] = (ui >> 3) & 0xf8;
                                        imdst1[3*(j*WIDTH+l)+2] = (ui >> 8) & 0xf8;
                                }
                        }
                }
        }

}





void
render_image_16_to_32(tr_u16 *imsrc2, tr_u32 *imdst4, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;

        for(j=0, i=0; i<height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
#ifdef TR_ENDIAN_IS_BIG
                                ui = imsrc2[i*width+k];
#endif
#ifdef TR_ENDIAN_IS_LITTLE
                                ui = imsrc2[(i*width+k)^0x01];
#endif
                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst4[j*WIDTH+l] =
                                                ((ui << 18) & 0xf80000) |
                                                ((ui <<  5) & 0x00f800) |
                                                ((ui <<  2) & 0x0000f8);
                                }
                        }
                }
        }
}










void
render_image_32_to_06(tr_u32 *imsrc4, tr_u8 *imdst1, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u32    ul;
        tr_u8    c;

        for(j=0, i=0; i<height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
                                ul = imsrc4[i*width+k];
                                c = ((ul & 0xc0) >> 6) | ((ul & 0xc000) >> 12) | ((ul & 0xc00000) >> 18);
                                
                                for (; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst1[j*WIDTH] = c;
                                }
                        }
                }
        }
}





void
render_image_32_to_15(tr_u32 *imsrc4, tr_u16 *imdst2, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;
        tr_u32    ul;

        for(j=0, i=0; i < height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
                                ul = imsrc4[i*width+k];
                                ui = ((ul & 0xf8) >> 3) | ((ul & 0xf800) >> 6) | ((ul & 0xf80000) >> 9);

                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst2[j*WIDTH+l] = ui;
                                }
                        }
                }
        }
}





void
render_image_32_to_16(tr_u32 *imsrc4, tr_u16 *imdst2, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;
        tr_u16   ui;
        tr_u32    ul;

        for(j=0, i=0; i < height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
                                ul = imsrc4[i*width+k];
                                ui = ((ul & 0xf8) >> 3) | ((ul & 0xfc00) >> 5) | ((ul & 0xf80000) >> 8);

                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst2[j*WIDTH+l] = ui;
                                }
                        }
                }
        }
}





void
render_image_32_to_24(tr_u32 *imsrc4, tr_u8 *imdst1, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;

        for(j=0, i=0; i < height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst1[3*(j*WIDTH+l)+0] = (imsrc4[i*width+k] & 0xf800)     >> 8;
                                        imdst1[3*(j*WIDTH+l)+1] = (imsrc4[i*width+k] & 0xf80000)   >> 16;
                                        imdst1[3*(j*WIDTH+l)+2] = (imsrc4[i*width+k] & 0xf8000000) >> 24;
                                }
                        }
                }
        }
}





void
render_image_32_to_32(tr_u32 *imsrc4, tr_u32 *imdst4, int width, int height, int WIDTH, int HEIGHT)
{
        int     i, j, k, l;

        for(j=0, i=0; i < height; i++)
        {
                for(; (j+1)*height <= (i+1)*HEIGHT; j++)
                {
                        for(l=0, k=0; k<width; k++)
                        {
                                for(; (l+1)*width <= (k+1)*WIDTH; l++)
                                {
                                        imdst4[j*WIDTH+l] = imsrc4[i*width+k] >> 8;
                                }
                        }
                }
        }
}

